/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTDialog.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTPrototype.jsm");

let EXPORTED_SYMBOLS = ["UnMHTConfirmDialogClass"];

/**
 * 汎用確認ダイアログ
 *
 * ウィンドウの引数
 * {
 *   result   : Object  結果を格納するオブジェクト (オプショナル),
 *
 *   title    : String  ダイアログのタイトル (必須),
 *   message1 : String  ダイアログのメッセージ (必須),
 *   message2 : String  ダイアログのメッセージ 2 行目 (オプショナル),
 *   message3 : String  ダイアログのメッセージ 3 行目 (オプショナル),
 *
 *   accept   : String  OK ボタン [yes] のラベル (必須),
 *   cancel   : String  キャンセルボタン [no] のラベル (オプショナル),
 *   extra1   : String  スキップボタン [skip] のラベル (オプショナル),
 *   extra2   : String  中断ボタン [abort] のラベル (オプショナル),
 *   toall    : String  全てに適用チェックのラベル (オプショナル),
 *
 *   rename   : String  リネームボタン [rename] のラベル (オプショナル),
 *   filename : String  既存のファイルのファイル名 (rename により必須),
 *   baseDir  : nsIFile  既存のファイルの親ディレクトリ (rename により必須),
 *   exist1   : String  ファイルが存在するメッセージの前半 (rename により必須),
 *   exist2   : String  ファイルが存在するメッセージの後半 (rename により必須)
 * }
 *
 * 結果
 * {
 *   result     : String   押したボタン (yes/no/abort/skip/rename),
 *   applyToAll : Boolean  全てに適用するか,
 *   filename   : Boolean  リネームするファイル名
 * }
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 * @inherits   UnMHTDialog
 * @implements IUnMHTDialogExtra
 */
function UnMHTConfirmDialogClass (window) {
  UnMHTDialog.call (this, window);
}
UnMHTConfirmDialogClass.prototype = {
  /**
   * 各種ボタンを押したイベント
   *
   * @param  String type
   *         ボタンの種類 (yes/no/abort/skip/rename)
   * @return Boolean
   *         ダイアログを閉じるか
   */
  onClick : function (type) {
    this.el ("message").value = "";
    this.el ("message").hidden = true;
    
    let filename = this.el ("filename").value;
    
    if (type == "rename") {
      let file = this._baseDir.clone ();
      if (!file) {
        this.el ("message").value = UnMHTStringBundle.get ("illegal_directory")
        this.el ("message").hidden = false;
        this.el ("filename").focus ();
        
        return false;
      }
      
      file.append (filename);
      
      if (file.exists ()) {
        this.el ("message").value
          = this._existsMessage1 + filename + this._existsMessage2;
        this.el ("message").hidden = false;
        this.el ("filename").focus ();
        
        return false;
      }
    }
    
    if (this._result) {
      this._result.result = type;
      this._result.applyToAll = this.el ("apply_to_all").checked;
      this._result.filename = filename;
    }
    
    if (type == "abort" || type == "skip" || type == "rename") {
      /* ダイアログのボタンではないので手動で閉じる */
      this._window.close ();
    }
    
    return true;
  },
  
  /* ==== IUnMHTDialogExtra インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct : function () {
    this._result = null;
    this._baseDir = null;
    UnMHTPrototype.super (this, "destruct").call (this);
  },
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    let arg = this._window.arguments [0];
    
    let dialog = this.el ("unmht_confirm");
    
    if ("result" in arg) {
      this._result = arg.result;
    }
    
    dialog.setAttribute ("title", arg.title);
    this.el ("title").setAttribute ("value", arg.title);
    
    this.el ("message1").value = arg.message1;
    if ("message2" in arg && arg.message2) {
      this.el ("message2").value = arg.message2;
      this.el ("message2").hidden = false;
    }
    if ("message3" in arg && arg.message3) {
      this.el ("message3").value = arg.message3;
      this.el ("message3").hidden = false;
    }
    
    dialog.getButton ("accept").setAttribute ("label", arg.accept);
    
    if ("cancel" in arg) {
      dialog.getButton ("cancel").setAttribute ("label", arg.cancel);
    }
    else {
      dialog.getButton ("cancel").hidden = true;
    }
    
    if ("extra1" in arg) {
      dialog.getButton ("extra1").setAttribute ("label", arg.extra1);
    }
    else {
      dialog.getButton ("extra1").hidden = true;
    }
    
    if ("extra2" in arg) {
      dialog.getButton ("extra2").setAttribute ("label", arg.extra2);
    }
    else {
      dialog.getButton ("extra2").hidden = true;
    }
    
    if ("toall" in arg && arg.toall) {
      this.el ("apply_to_all").checked = false;
      this.el ("apply_to_all").label = arg.toall;
      this.el ("apply_to_all").hidden = false;
    }
    
    if ("rename" in arg) {
      this.el ("rename").label = arg.rename;
      
      if ("filename" in arg) {
        this.el ("filename").value = arg.filename;
      }
      
      this.el ("renamebox").hidden = false;
    }
    
    if ("baseDir" in arg) {
      this._baseDir = arg.baseDir;
    }
    
    if ("exist1" in arg && arg.exist1) {
      this._existsMessage1 = arg.exist1 + " ";
    }
    if ("exist2" in arg && arg.exist2) {
      this._existsMessage2 = " " + arg.exist2;
    }
  },
  
  /**
   * ダイアログを OK したイベント
   *
   * @return Boolean
   *         ダイアログを閉じるか
   */
  onDialogAccept : function () {
    return this.onClick ("yes");
  },
  
  /**
   * ダイアログをキャンセルしたイベント
   *
   * @return Boolean
   *         ダイアログを閉じるか
   */
  onDialogCancel : function () {
    return this.onClick ("no");
  },
  
  /**
   * ダイアログの extra1 を押したイベント
   */
  onDialogExtra1 : function () {
    this.onClick ("skip");
  },
  
  /**
   * ダイアログの extra2 を押したイベント
   */
  onDialogExtra2 : function () {
    this.onClick ("abort");
  },
  
  /* ==== private ==== */

  _result : null,       /* Object  結果オブジェクト */
  _baseDir : null,      /* nsIFile  ファイルのベースディレクトリ */
  _existsMessage1 : "", /* String  ファイルが存在した場合のメッセージ */
  _existsMessage2 : ""  /* String  ファイルが存在した場合のメッセージ */
};
UnMHTPrototype.createInheritance (UnMHTConfirmDialogClass, UnMHTDialog);
