/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTAppInfo.jsm");
XPCOMUtils.importRelative (this, "UnMHTHTMLUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTMailPrivacy.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTDOMUtil"];

/**
 * DOM 操作ユーティリティ
 */
let UnMHTDOMUtil = {
  /**
   * ノードが選択範囲内かどうかチェックする
   *
   * @param  nsIDOMNode node
   *         対象のノード
   * @param  nsISelection selection
   *         対象の選択範囲
   * @return Boolean
   *         範囲内かどうか
   */
  isNodeInSelection : function (node, selection) {
    for (let i = 0; i < selection.rangeCount; i ++) {
      let range = selection.getRangeAt (i);
      
      let nodeRange = node.ownerDocument.createRange ();
      let w = node.ownerDocument.defaultView;
      nodeRange.selectNode (node);
      
      let compareSS = 0;
      let compareSE = 0;
      let compareES = 0;
      let compareEE = 0;
      try {
        compareSS
          = range.compareBoundaryPoints (w.Range.START_TO_START, nodeRange);
        compareSE
          = range.compareBoundaryPoints (w.Range.START_TO_END, nodeRange);
        compareES
          = range.compareBoundaryPoints (w.Range.END_TO_START, nodeRange);
        compareEE
          = range.compareBoundaryPoints (w.Range.END_TO_END, nodeRange);
      }
      catch (e) {
        /* 選択範囲が別のドキュメント */
        return false;
      }
        
      if (!((compareSS == -1
             && compareSE == -1
             && compareES == -1
             && compareEE == -1)
            || (compareSS == 1
                && compareSE == 1
                && compareES == 1
                && compareEE == 1))) {
        return true;
      }
    }
    
    return false;
  },
  
  /**
   * HTML ツリーを文字列に変換する
   *
   * @param  String location
   *         ドキュメントの URI
   * @param  nsIDOMNode node
   *         対象の要素
   * @param  nsISelection targetSelection
   *         対象の選択範囲もしくは null
   * @param  Object params
   *           {
   *             noScript   : Boolean スクリプトを削除するか,
   *             isMail     : Boolean メールか,
   *             saveCanvas : Boolean キャンバスを保存するか
   *           }
   * @return String
   *         要素全体の HTML
   */
  convertToPlainText : function (location, node, targetSelection, params) {
    if (!node) {
      return "";
    }
    if (!node.nodeName) {
      return "";
    }
    
    if (targetSelection) {
      if (node instanceof Ci.nsIDOMHTMLHeadElement
          || node instanceof Ci.nsIDOMHTMLStyleElement
          || node instanceof Ci.nsIDOMHTMLTitleElement) {
        targetSelection = null;
      }
    }
    
    if (targetSelection) {
      if (!UnMHTDOMUtil.isNodeInSelection (node, targetSelection)) {
        return "";
      }
    }
    
    let text = "";
    
    if (node instanceof Ci.nsIDOMText) {
      text = UnMHTHTMLUtil.escapeEntity (node.nodeValue);
      
      if (targetSelection) {
        let textList = [];
        for (let i = 0; i < targetSelection.rangeCount; i ++) {
          let t = text;
          let isPartial = false;
          
          let range = targetSelection.getRangeAt (i);
          
          if (range.endContainer == node) {
            try {
              t = t.substr (0, range.endOffset);
              isPartial = true;
            }
            catch (e) {
            }
          }
          
          if (range.startContainer == node) {
            try {
              t = t.substr (range.startOffset);
              isPartial = true;
            }
            catch (e) {
            }
          }
          
          if (isPartial) {
            textList.push (t);
          }
        }
        
        if (textList.length > 0) {
          text = textList.join ("");
        }
      }
    }
    else if (node instanceof Ci.nsIDOMComment) {
      /* コメントをコメントとして展開する */
      text = "<!--" + node.nodeValue + "-->";
    }
    else if (node instanceof Ci.nsIDOMHTMLBaseElement
             && "isMail" in params
             && params.isMail) {
    }
    else if (node instanceof Ci.nsIDOMHTMLScriptElement
             && "noScript" in params
             && params.noScript) {
    }
    else {
      let attributes = "";
      
      let addStylePre = "";
      let addStylePost = "";
      
      if ("saveCanvas" in params
          && params.saveCanvas
          && node instanceof Ci.nsIDOMHTMLCanvasElement) {
        addStylePre
          += "background-image: url(" + node.toDataURL ("image/png") + ");";
      }
      
      if (node.attributes) {
        Array.forEach
          (node.attributes,
           function (attr) {
            let name = attr.name.toLowerCase ();
            let value = attr.value;
          
            if (name.match (/^__unmht_/)) {
              return;
            }
                        
            if (node instanceof Ci.nsIDOMHTMLOptionElement) {
              if (name == "selected") {
                return;
              }
            }
            
            if ("noScript" in params
                && params.noScript
                && name.match (/^on/)) {
              return;
            }
            
            if (node instanceof Ci.nsIDOMHTMLInputElement) {
              if (node.type == "checkbox"
                  || node.type == "radio") {
                if (name == "checked") {
                  return;
                }
              }
              else {
                if (name == "value") {
                  return;
                }
              }
            }
            
            if ((node instanceof Ci.nsIDOMHTMLFrameElement
                 || node instanceof Ci.nsIDOMHTMLIFrameElement)
                && name == "src") {
              value = node.contentDocument.location.href;
            }
            
            if (name.match (/^(src|href|background|action|data)$/)
                && "isMail" in params
                && params.isMail) {
              value = UnMHTMailPrivacy.hideMailURL (value, location);
            }
            
            if (name == "style") {
              if (addStylePre) {
                value = addStylePre + value;
                addStylePre = "";
              }
              if (addStylePost) {
                value = value + ";" + addStylePost;
                addStylePost = "";
              }
            }
            
            value = UnMHTHTMLUtil.escapeEntity (value);
            
            attributes += " " + name +  "=" + "\"" + value + "\"";
          });
      }
      
      if (addStylePre || addStylePost) {
        let name = "style";
        let value = "";
        if (addStylePre) {
          value = addStylePre + value;
          addStylePre = "";
        }
        if (addStylePost) {
          value = value + ";" + addStylePost;
          addStylePost = "";
        }
        attributes += " " + name +  "=" + "\"" + value + "\"";
      }
      
      if (node instanceof Ci.nsIDOMHTMLOptionElement) {
        if (node.selected) {
          let name = "selected";
          let value = "selected";
          attributes += " " + name +  "=" + "\"" + value + "\"";
        }
      }

      if (node instanceof Ci.nsIDOMHTMLInputElement) {
        if (node.type == "checkbox"
            || node.type == "radio") {
          if (node.checked) {
            let name = "checked";
            let value = "checked";
            attributes += " " + name +  "=" + "\"" + value + "\"";
          }
        }
        else if (node.type == "file") {
        }
        else {
          let name = "value";
          let value = UnMHTHTMLUtil.escapeEntity (node.value);
          attributes += " " + name +  "=" + "\"" + value + "\"";
        }
      }

      let nodeName = node.nodeName.toLowerCase ();
      
      text += "<" + nodeName + attributes + ">";
      
      /* NOSCRIPT, NOEMBED 要素は専用のインターフェース定義を持たない */
      if (nodeName == "noscript"
          && node.firstChild
          && node.firstChild instanceof Ci.nsIDOMText
          && node.firstChild.nextSibling == null) {
        /* NOSCRIPT 要素が展開されていない場合 */
        text += node.firstChild.nodeValue;
      }
      else if (nodeName == "noembed"
          && node.firstChild
          && node.firstChild instanceof Ci.nsIDOMText
          && node.firstChild.nextSibling == null) {
        /* NOEMBED 要素が展開されていない場合 */
        text += node.firstChild.nodeValue;
      }
      else if (node instanceof Ci.nsIDOMHTMLStyleElement
          || node instanceof Ci.nsIDOMHTMLScriptElement) {
        /* noScript の場合には SCRIPT 要素はここまで来ない */
        let node2 = node.firstChild;
        while (node2) {
          text
            += UnMHTHTMLUtil.unescapeEntity
            (this.convertToPlainText
             (location, node2, targetSelection, params));
                        
          node2 = node2.nextSibling;
        }
      }
      else if (node instanceof Ci.nsIDOMHTMLTextAreaElement) {
        text += UnMHTHTMLUtil.escapeEntity (node.value);
      }
      else {
        let node2 = node.firstChild;
        while (node2) {
          text
            += this.convertToPlainText
            (location, node2, targetSelection, params);
                        
          node2 = node2.nextSibling;
        }
      }
      
      if (!(node instanceof Ci.nsIDOMHTMLBRElement
            || node instanceof Ci.nsIDOMHTMLBaseElement
            || node instanceof Ci.nsIDOMHTMLLinkElement
            || node instanceof Ci.nsIDOMHTMLImageElement
            || node instanceof Ci.nsIDOMHTMLInputElement
            || node instanceof Ci.nsIDOMHTMLHRElement
            || node instanceof Ci.nsIDOMHTMLMetaElement)) {
        text += "</" + nodeName + ">";
      }
    }
    
    return text;
  },
  
  /**
   * ノードを親方向に探す
   * 対象のノードも含む
   *
   * @param  nsIDOMNode node
   *         対象のノード
   * @param  String nodeName
   *         探すノード名
   * @return nsIDOMNode
   *         見付かったノード
   *         見付からなかった場合 null
   */
  findParentNode : function (node, nodeName) {
    while (node) {
      if (node.nodeName.toLowerCase () == nodeName) {
        return node;
      }
      node = node.parentNode;
    }
    
    return null;
  },
  
  /**
   * コンテンツが選択されているかどうかを返す
   *
   * @param  nsIDOMWindow window
   *         動作中のウィンドウ
   * @param  Boolean usePopupNode
   *         コンテキストメニューから情報を取得するか
   * @return Boolean
   *         コンテンツが選択されているかどうか
   */
  isContentSelected : function (window, usePopupNode) {
    if (usePopupNode
        && "gContextMenu" in window
        && window.gContextMenu) {
      return window.gContextMenu.isContentSelected;
    }
    else {
      let targetDocument = null;
      if (usePopupNode && window.document.popupNode) {
        targetDocument = window.document.popupNode.ownerDocument;
      }
      else {
        targetDocument = UnMHTWindowUtil.getCurrentDocument (window);
      }
      
      if (!targetDocument.defaultView) {
        return false;
      }
      
      try {
        return !targetDocument.defaultView.getSelection ().isCollapsed;
      }
      catch (e) {
        return false;
      }
    }
  },
  
  /**
   * リンク先の URI を取得する
   * 
   * @param  nsIDOMWindow window
   *         動作中のウィンドウ
   * @param  nsIDOMNode node
   *         対象のノード
   * @return String
   *         リンク先の URI
   */
  getLinkURL : function (window) {
    if ("gContextMenu" in window
        && window.gContextMenu) {
      return window.gContextMenu.linkURL;
    }
    return "";
  },
  
  /**
   * 保存可能なリンクの上に乗っているかどうかを返す
   * コンテキストメニューを表示してからでないと正しい値を返さない
   *
   * @param  nsIDOMWindow window
   *         動作中のウィンドウ
   * @return Boolean
   *         リンクの上に乗っているかどうか
   */
  onSavableLink : function (window) {
    if ("gContextMenu" in window
        && window.gContextMenu) {
      return window.gContextMenu.onSaveableLink;
    }
    return false;
  },
  
  /**
   * ページ上の何でもない所に乗っているかどうかを返す
   *   (リンクや画像、入力フィールド等でない事)
   * コンテキストメニューを表示してからでないと正しい値を返さない
   *
   * @param  nsIDOMWindow window
   *         動作中のウィンドウ
   * @return Boolean
   *         ページ上の何でもない所に乗っているかどうか
   */
  onPage : function (window) {
    if ("gContextMenu" in window
        && window.gContextMenu) {
      return !(window.gContextMenu.onImage
               || window.gContextMenu.onCanvas
               || window.gContextMenu.onVideo
               || window.gContextMenu.onAudio
               || window.gContextMenu.onLink
               || window.gContextMenu.onTextInput);
    }
    
    return true;
  }
};
