/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

/**
 * Defines a Document Loader Factory to handle the MHT document type.
 *
 * Currently, this implementation just fires the standard MHT loading process,
 *  and refuses to create a content viewer and a stream listener.
 *
 * For more information on the document loading process, see
 *  <http://www.mozilla.org/newlayout/doc/webwidget.html> and
 *  <https://developer.mozilla.org/En/The_life_of_an_HTML_HTTP_request>
 *  (retrieved 2008-10-07).
 */

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTAppInfo.jsm");
XPCOMUtils.importRelative (this, "UnMHTExtractor.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileLoader.jsm");
XPCOMUtils.importRelative (this, "UnMHTScheme.jsm");

let EXPORTED_SYMBOLS = ["UnMHTDocumentLoaderFactory"];

/**
 * ファイルを展開するリスナ
 *
 * @param  String originalURISpec
 *         元のチャネルの URI
 * @param  nsIWebNavigation nav
 *         ドキュメントを開いているウィンドウの WebNavigation
 * @implements nsIRequestObserver
 * @implements nsIStreamListener
 */
function UnMHTDLFExtractListener (originalURISpec, nav) {
  this._originalURISpec = originalURISpec;
  this._nav = nav;
}
UnMHTDLFExtractListener.prototype = {
  QueryInterface: XPCOMUtils.generateQI ([Ci.nsIRequestObserver,
                                          Ci.nsIStreamListener]),
  
  /**
   * 破棄
   * 自分で呼ぶ
   */
  destruct : function () {
    this._originalURISpec = "";
    this._nav = null;
    this._data = "";
  },
  
  /* ==== nsIRequestObserver インターフェース ==== */
  
  /**
   * リクエスト開始のイベント
   *
   * @param  nsIRequest request
   *         対象のリクエスト
   * @param  nsISupports context
   *         ユーザ定義
   */
  onStartRequest : function (request, context) {
  },
      
  /**
   * リクエスト終了のイベント
   *
   * @param  nsIRequest request
   *         対象のリクエスト
   * @param  nsISupports context
   *         ユーザ定義
   * @param  Number statusCode
   *         終了コード
   */
  onStopRequest : function (request, context, statusCode) {
    let eFileInfo
    = UnMHTExtractor.extractMHT (this._originalURISpec, this._data, true);
      
    if (eFileInfo) {
      try {
        this._nav.loadURI
          (eFileInfo.baseURI,
           Ci.nsIWebNavigation.LOAD_FLAGS_REPLACE_HISTORY,
           null, null, null);
      }
      catch (e) {
      }
    }
    
    this.destruct ();
  },
      
  /* ==== nsIStreamListener インターフェース ==== */
  
  /**
   * データ到着のイベント
   *
   * @param  nsIRequest request
   *         対象のリクエスト
   * @param  nsISupports context
   *         ユーザ定義
   * @param  nsIInputStream inputStream
   *         データを取得するストリーム
   * @param  PRUint32 offset
   *         データの位置
   * @param  PRUint32 count 
   *         データの長さ
   */
  onDataAvailable : function (request, context, inputStream, offset,
                              count) {
    let bstream
    = Cc ["@mozilla.org/binaryinputstream;1"]
    .createInstance (Ci.nsIBinaryInputStream);
    bstream.setInputStream (inputStream);
    this._data += bstream.readBytes (count);
  },
  
  /* ==== private ==== */
  
  _originalURISpec : "", /* String  元のチャネルの URI */
  _nav : null,           /* nsIWebNavigation  ドキュメントを開いている
                          *   ウィンドウの WebNavigation */
  
  _data : ""             /* String  取得したデータ */
};

/**
 * Construct the UnMHTDocumentLoaderFactory object.
 *
 * Document loader factories are XPCOM services, so they should be constructed
 *  as singletons, for example as explained in
 *  <http://weblogs.mozillazine.org/weirdal/archives/019620.html> (retrieved
 *  2008-10-07). However, since this object doesn't store any state information,
 *  we might as well create it as a normal XPCOM object.
 *
 * Document loader factories must be registered with the
 *  "Gecko-Content-Viewers" category. For this extension, this is done
 *  dynamically on startup, because the list of MIME types that this document
 *  loader factory will handle is not known in advance.
 *
 * @implements nsIDocumentLoaderFactory
 */
function UnMHTDocumentLoaderFactory () {
}
UnMHTDocumentLoaderFactory.prototype = {
  classID:          Components.ID ("{fa22337c-baf4-4c05-ab92-5e1fb4a97aa9}"),
  contractID:       "@unmht.org/unmht-document-loader-factory;1",
  classDescription: "UnMHT Document Loader Factory JS Component",
  QueryInterface:   XPCOMUtils.generateQI ([Ci.nsIDocumentLoaderFactory]),
  
  /* ==== nsIDocumentLoaderFactory インターフェース ==== */
  
  /**
   * 空のページのコンテントビューアを作成する
   *
   * @param  String aCommand
   *         実行するコマンド
   * @param  nsIChannel aChannel
   *         入力チャネル
   * @param  nsILoadGroup aLoadGroup
   *         不明
   * @param  String aContentType
   *         表示するコンテンツの MIME-Type
   * @param  nsISupport aContainer
   *         不明
   * @param  nsISupport aExtraInfo
   *         不明
   * @param  Object aDocListenerResult
   *         入力チャネルのストリームのリスナ (出力)
   *         {
   *           value: nsIStreamListener
   *         }
   * @return nsIContentViewer
   *         空のページのコンテントビューア
   */
  createInstance: function (aCommand, aChannel, aLoadGroup, aContentType,
                            aContainer, aExtraInfo, aDocListenerResult) {
    if (aCommand != "view") {
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    }
    
    let originalURISpec = aChannel.URI.spec;
    
    if (originalURISpec.match (/^(imap|mailbox|news|nntp|pop|snews):\/\//)) {
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    }
    
    let emptyChannel
    = Services.io.newChannelFromURI
    (Services.io.newURI (UnMHTScheme.createEmptyURI (), "utf-8", null));
    
    let listener = {};
    
    let viewer
    = Cc ["@mozilla.org/content/document-loader-factory;1"]
    .getService (Ci.nsIDocumentLoaderFactory)
    .createInstance ("view", emptyChannel, aLoadGroup, "text/html",
                     aContainer, aExtraInfo, listener);
    
    emptyChannel.asyncOpen
    (listener.value.QueryInterface (Ci.nsIStreamListener), null);
    
    let nav = null;
    try {
      nav = aContainer.QueryInterface (Ci.nsIInterfaceRequestor)
        .getInterface (Ci.nsIWebNavigation);
    }
    catch (e) {
    }
    
    UnMHTFileLoader.registerHandled (originalURISpec);
    
    aDocListenerResult.value
    = new UnMHTDLFExtractListener (originalURISpec, nav);
    
    return viewer;
  },
  
  /**
   * ドキュメントに対応するコンテントビューアの作成
   *
   * @param  nsISupport aContainer
   *         不明
   * @param  nsIDocument aDocment
   *         対象のドキュメント
   * @param  String aCommand
   *         実行するコマンド
   */
  createInstanceForDocument: function (aContainer, aDocument, aCommand) {
    /* This function should never have been called. As of 2008-10-07, it may
     *  be called only from nsDocShell::CreateAboutBlankContentViewer, that
     *  doesn't handle exceptions, so we return null instead of throwing
     *  NS_ERROR_NOT_IMPLEMENTED. */
    return null;
  },
  
  /**
   * 空のドキュメントの作成
   *
   * @param  nsILoadGroup aLoadGroup
   *         不明
   * @param  nsIPrincipal aPrincipal
   *         不明 (オプショナル)
   */
  createBlankDocument: function (aLoadGroup, aPrincipal) {
    /* See the comment inside createInstanceForDocument. */
    return null;
  }
};
