/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTArrayUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTHTMLUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTMIMEUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTMaxRunTime.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative (this, "UnMHTShutdown.jsm");
XPCOMUtils.importRelative (this, "UnMHTUnicodeConverter.jsm");

let EXPORTED_SYMBOLS = ["UnMHTExtractor"];

/**
 * パートの展開情報
 */
function UnMHTExtractParam () {
  let randomHex = function () {
    return Math.floor (Math.random () * 15).toString (16).toUpperCase ();
  };
  
  this.cid = [randomHex() for (i in UnMHTArrayUtil.range (1, 32))].join ("");
}
UnMHTExtractParam.prototype = {
  charset : "",         /* String  キャラクタセット */
  mimetype : "",        /* String  MIME-Type */
  
  isHTML : false,       /* Boolean  HTML かどうか */
  isCSS : false,        /* Boolean  CSS かどうか */
  
  isMixed : false,      /* Boolean  multipart/mixed かどうか */
  
  cid : "",             /* String  Content-ID 代わり */
  
  location : "",        /* String  ファイルのアドレス */
  normalLocation : "",  /* String  正規化したファイルのアドレス
                         *         フラグメント抜き */
  
  contentBase : "",     /* String  Content-Base の値 */
  
  baseDir : "",         /* String  親ディレクトリ */
  referredBaseDir : "", /* String  参照元のファイルの親ディレクトリ */
  /*
   * referredBaseDir は
   * IE で保存された MHT ファイルにおける以下の問題を回避するためのもの
   *
   * [状況]
   *   ファイル
   *     http://unmht.org/x.html
   *     http://unmht.org/style/a.css
   *     http://unmht.org/style/b.css
   *   x.html 中の記述
   *     <link href="./style/a.css" ... />
   *   a.css 中の記述
   *     @import url("./b.css");
   *
   * [期待される結果]
   *   相対パス "./b.css" を a.css の URI から解決した URL が
   *   b.css の Content-Location に記述される
   *
   *   Content-Location: http://unmht.org/style/b.css
   *
   *   これにより a.css から b.css が参照できる
   *
   * [実際の結果]
   *   相対パス "./b.css" を x.html の URI から解決した URL が
   *   b.css の Content-Location に記述される
   *
   *   Content-Location: http://unmht.org/b.css
   *
   *   これにより a.css から b.css が参照できない
   *
   * [対策]
   *   1. すべての X から参照されているすべての Y について
   *      Y.referredBaseDir を X.baseDir とする
   *   2. Z.referredBaseDir が空でない Z について
   *      Z 内の相対パス R による参照について
   *      Z.baseDir から R を解決した URI を A とする
   *      Z.referredBaseDir から R を解決した URI を B とする
   *   3. F.contentLocation が A に一致するような F が存在しない、かつ
   *      G.contentLocation が B に一致するような G が存在する場合
   *      R による参照先を G とする
   *      この時、G.referredBaseDir は F.baseDir となり 2 から再度適用する
   *
   * [例]
   *   0. {x.html}.baseDir = "http://unmht.org/"
   *      {a.css}.baseDir = "http://unmht.org/style/"
   *      {b.css}.contentLocation = "http://unmht.org/b.css"
   *   1. {a.css}.referredBaseDir = {x.html}.baseDir
   *                                "http://unmht.org/"
   *   2. Z = {a.css}
   *      R = "./b.css"
   *      A = {a.css}.baseDir + R
   *        = "http://unmht.org/style/b.css"
   *      B = {a.css}.referredBaseDir + R
   *        = "http://unmht.org/b.css"
   *   3. A != {b.css}.contentLocation
   *      B == {b.css}.contentLocation
   *      # a.css から b.css が参照できる
   *      {b.css}.referredBaseDir = {a.css}.baseDir
   *                                "http://unmht.org/style/"
   *   2. Z = {b.css}
   *      A = {b.css}.baseDir + R
   *        = "http://unmht.org/" + R
   *      B = {b.css}.referredBaseDir + R
   *        = "http://unmht.org/style/" + R
   *      # b.css から "./i.png" で参照されたファイル i.png が
   *      #   {i.png}.contentLocation = "http://unmht.org/i.png"
   *      # となっていても
   *      #   {i.png}.contentLocation = "http://unmht.org/style/i.png"
   *      # となっていても参照できる
   *
   * [問題]
   *   誤って生成された Content-Location に一致するファイルが
   *   同 MHT 内に存在した場合これらを区別する事はできない
   */
  leafName : "",        /* String  ファイル名 */
  normalLeafName : "",  /* String  正規化したファイル名 */
  
  fragment : "",        /* String  フラグメント */
  normalFragment : "",  /* String  正規化したフラグメント */
  
  content : "",         /* String  修正済みのファイルの中身 */
  
  refName : ""          /* String  参照用のファイル名 */
};

/**
 * ファイル全体の展開情報
 */
function UnMHTExtractFileInfo () {
}
UnMHTExtractFileInfo.prototype = {
  original : "",         /* String  展開したファイル名の URI */
  baseURI : "",          /* String  開始ページの URI */
  
  startPart : null,      /* UnMHTMIMEPart 開始ページのパート */
  
  topPart : null,        /* UnMHTMIMEPart  トップレベルのパート */
  subject : "",          /* String  Subject の値 */
  date : "",             /* String  Date の値 */
  
  parts : null,          /* Array  mht ファイルの中にあるパート
                          *   [UnMHTMIMEPart パート, ...] */
  refParts : null,       /* Object  参照できるパート
                          *   <String 参照用のファイル名,
                          *    UnMHTMIMEPart パート> */
  
  extractTime : 0,       /* Number  展開時刻 */
  lastAccessTime : 0,    /* Number  最終アクセス時刻 */
  
  size : 0,              /* Number  全キャッシュのサイズ */
  
  /**
   * データを開放する
   */
  destruct : function () {
    this.parts
    .forEach (function (part) {
        part.eParam.content = null;
        part.eParam = null;
      });
    this.parts = null;
    
    if (this.topPart) {
      UnMHTMIMEUtil.destructPart (this.topPart);
      this.topPart = null;
    }
    
    this.startPart = null;
    this.refParts = null;
  }
};

let PERM_IMAGE_ALLOW_ALL = 1;
let PERM_IMAGE_BLOCK_ALL = 2;
let PERM_IMAGE_BLOCK_THIRD_PIRTY = 3;

/*
 * ファイルの展開と展開情報の管理
 */
let UnMHTExtractor = {
  /**
   * 一意性を確保するためにファイル名を正規化する
   *
   * @param  String original
   *         ファイル名
   * @return String
   *         正規化したファイル名
   */
  normalizeName : function (original) {
    if (!original) {
      return "";
    }
    
    for (let i = 0; i < 10; i ++) {
      let last = original;
      original = UnMHTHTMLUtil.unescapeEntity (unescape (original));
      if (last == original) {
        break;
      }
    }
        
    return original;
  },
  
  /**
   * キャッシュから展開情報を返す
   *
   * @param  String originalURISpec
   *         展開したファイル名の URI 表記
   * @return UnMHTExtractFileInfo
   *         展開情報
   *         見付からなければ null
   */
  getFileInfoFromOriginalURISpec : function (originalURISpec) {
    return this._getFileInfo (UnMHTScheme.enUnMHTURI (originalURISpec));
  },
  
  /**
   * mht ファイルを展開する
   *
   * @param  String originalURISpec
   *         展開したファイル名の URI 表記
   * @param  String text
   *         mht ファイルの内容
   * @param  Boolean extract
   *         展開するかどうか
   *           false の場合、text で与えられた内容のファイルを作る
   * @return UnMHTExtractFileInfo
   *         展開情報
   */
  extractMHT : function (originalURISpec, text, extract) {
    let eFileInfo = new UnMHTExtractFileInfo ();
    
    eFileInfo.original
    = originalURISpec
    .replace
    (/[\(\)\'\"\[\] ]/g, 
     function (matched) {
      let hexString = "0" + matched.charCodeAt (0).toString (16).toUpperCase ();
      return "%" + hexString.slice (-2);
    });
    
    eFileInfo.baseURI = UnMHTScheme.enUnMHTURI (eFileInfo.original);
    
    eFileInfo.extractTime = (new Date ()).getTime ();
    
    UnMHTMaxRunTime.setExtractTime ();
    
    eFileInfo.parts = [];
    eFileInfo.refParts = {};
    
    if (extract) {
      eFileInfo.topPart
        = UnMHTMIMEUtil.decodeMessage (text.replace (/\r\n|\r|\n/g, "\r\n"));
    }
    
    if (!eFileInfo.topPart) {
      let createDummyMessage = function (mimetype, body) {
        let data
        = "From: <Created by UnMHT>" + "\r\n"
        + "Content-Type: " + mimetype + "\r\n"
        + "\r\n"
        + body;
        
        return data;
      };
      
      let isHTML= false;
      if (!originalURISpec.match (/\.(eml|mbs)$/i) && text.match (/^\s*</)) {
        isHTML = true;
      }
      
      if (extract && isHTML) {
        eFileInfo.topPart = UnMHTMIMEUtil.decodeMessage
          (createDummyMessage ("text/html; charset=\"UTF-8\"", text));
      }
      else {
        eFileInfo.topPart = UnMHTMIMEUtil.decodeMessage
          (createDummyMessage ("text/plain; charset=\"UTF-8\"", text));
      }
    }
    
    if ("subject" in eFileInfo.topPart.fields) {
      let value = eFileInfo.topPart.fields ["subject"];
      eFileInfo.subject = UnMHTMIMEUtil.decodeWord (value);
    }
    if ("date" in eFileInfo.topPart.fields) {
      eFileInfo.date = eFileInfo.topPart.fields ["date"];
    }
    
    eFileInfo.startPart = UnMHTMIMEUtil.findStartPart (eFileInfo.topPart);
    
    this._createExtractParam (eFileInfo, eFileInfo.topPart, null);
    
    this._setRefName (eFileInfo);
    
    this._modifyAttribute (eFileInfo);
    
    this._modifyCSS (eFileInfo);
    
    this._createMixedDocument (eFileInfo);
    
    this._getPartsInfo (eFileInfo);
    
    UnMHTMaxRunTime.restoreTime ();
    
    eFileInfo.lastAccessTime = (new Date ()).getTime ();
    
    this._addFileInfo (eFileInfo);
    
    if (UnMHTPref.getPref ("open.cache.expire")) {
      this.removeOldFileInfo
        (UnMHTPref.getPref ("open.cache.expire.time") * 60000);
    }
    
    return eFileInfo;
  },
  
  /**
   * CSS 中の unmht: 形式の URI を解除する
   *
   * @param  UnMHTExtractFileInfo eFileInfo
   *         展開情報
   * @param  String content
   *         対象のドキュメント
   * @return String
   *         変換したドキュメント
   */
  unmodifyCSS : function (eFileInfo, content) {
    return content
    .replace (this._cssRe,
              function (matched, prev,
                        dquote, dqpath, quote, qpath, path) {
                return this._unCssFunc
                (eFileInfo, matched, prev,
                 dquote, dqpath, quote, qpath, path);
              }.bind (this));
  },
  
  /**
   * HTML 中の unmht: 形式の URI を解除する
   * この際 BASE 要素の内容を baseDir にする
   *
   * @param  UnMHTExtractFileInfo eFileInfo
   *         展開情報
   * @param  String content
   *         対象のドキュメント
   * @return String
   *         変換したドキュメント
   */
  unmodifyAttribute : function (eFileInfo, content) {
    return content
    .replace (this._elementRe,
              function (matched, prev, name, attrs, next) {
                return this._unElementFunc
                (eFileInfo, matched, prev, name, attrs, next);
              }.bind (this));
  },
  
  /**
   * Content-Location に対応するパスからパートを探す
   *
   * @param  UnMHTExtractFileInfo eFileInfo
   *         展開情報
   * @param  UnMHTMIMEPart originPart
   *         呼び出し元のパート
   * @param  String path
   *         対象のパス
   * @param  Boolean referred
   *         referredBaseDir をチェックするか
   * @param  Boolean useLeafNameCheck
   *         ファイル名のみのチェックを行うか
   * @return Array
   *           [UnMHTMIMEPart 対象のパート, String フラグメント]
   *         見付からなければパートが null
   */
  findPart : function (eFileInfo, originPart, path, referred,
                       useLeafNameCheck) {
    let fragment = "";
    let re = path.match (/^([^#]*)(#.+)$/);
    if (re) {
      [, path, fragment] = re;
    }
    if (fragment && path == "") {
      return [originPart, fragment];
    }
    
    let normalFragment = this.normalizeName (fragment);
    
    let normalPath = this.normalizeName (path);
    
    let part;
    
    if (normalFragment) {
      part
        = UnMHTArrayUtil.find
        (eFileInfo.parts,
         function (part) (normalPath == part.eParam.normalLocation
                          && normalFragment == part.eParam.normalFragment),
         null);
      if (part) { return [part, ""]; }
    }
    
    part
    = UnMHTArrayUtil.find
    (eFileInfo.parts,
     function (part) normalPath == part.eParam.normalLocation,
     null);
    if (part) { return [part, fragment]; }
    
    if (originPart.eParam.baseDir) {
      /* 相対パスでの参照をチェック */
      let resolvedPath
      = this.normalizeName
      (UnMHTFileUtil.jointPath (originPart.eParam.baseDir, normalPath));
      
      part
      = UnMHTArrayUtil.find
      (eFileInfo.parts,
       function (part) resolvedPath == part.eParam.normalLocation,
       null);
      if (part) { return [part, fragment]; }
    }
    
    re = normalPath.match (/^cid:(.+)$/);
    if (re) {
      let cid = re [1];
      
      part
      = UnMHTArrayUtil.find
      (eFileInfo.parts,
       function (part) cid == part.contentID,
       null);
      if (part) { return [part, fragment]; }
    }
    
    
    if (UnMHTScheme.isUnMHTURI (normalPath)) {
      let param = UnMHTScheme.getUnMHTURIParam (normalPath);
      let refName = param.refName;
      
      if (param.original == eFileInfo.original) {
        if (!refName) {
          return [eFileInfo.startPart, fragment];
        }
        
        if (refName in eFileInfo.refParts) {
          return [eFileInfo.refParts [refName], fragment];
        }
        
        return [null, ""]
      }
      
      /* JavaScript 等で生成された参照元からの絶対アドレスとしてチェック */
      /* **** FIXME: 相対アドレスは未対応 **** */
      
      let modPath
      = this.normalizeName
      (UnMHTFileUtil.jointPath (originPart.eParam.baseDir, "/" + param.path));
      
      part
      = UnMHTArrayUtil.find
      (eFileInfo.parts,
       function (part) modPath == part.eParam.normalLocation,
       null);
      if (part) { return [part, fragment]; }
    }
    
    if (referred && originPart.eParam.referredBaseDir != "") {
      let referredPath
      = this.normalizeName
      (UnMHTFileUtil.jointPath (originPart.eParam.referredBaseDir, normalPath));
      
      part
      = UnMHTArrayUtil.find
      (eFileInfo.parts,
       function (part) referredPath == part.eParam.normalLocation,
       null);
      if (part) { return [part, fragment]; }
    }
    
    if (useLeafNameCheck) {
      /* ファイル名だけでも同じものがあれば使う
       * あまり良い方法ではないので最小限にする */
      let noQuery = normalPath.replace (/\?.+/, "");
      part
      = UnMHTArrayUtil.find
      (eFileInfo.parts, function (part) {
        let leafName = part.eParam.normalLeafName;
        let length = leafName.length;
        return (leafName
                && noQuery.length >= length + 1
                && noQuery.substr (noQuery.length - length) == leafName
                && noQuery.substr (noQuery.length - length - 1, 1)
                .match (/[\/\?\\:]/));
      }, null);
      if (part) { return [part, fragment]; }
    }
    
    /* 含まれない */
    return [null, ""];
  },
  
  /**
   * unmht 形式の URI から展開情報を取得する
   *
   * @param  String uri
   *         unmht: 形式のファイルの URI
   * @return Array
   *         [UnMHTExtractFileInfo, UnMHTMIMEPart]
   *         展開情報
   *         見付からなかった場合はそれぞれ null
   */
  getFileInfoAndPart : function (uri) {
    if (!UnMHTScheme.isUnMHTURI (uri)) {
      return [null, null];
    }
    
    let part = null;
    
    let param = UnMHTScheme.getUnMHTURIParam (uri);
    let baseURI = param.baseURI;
    let refName = param.refName;
    
    let eFileInfo = this._getFileInfo (baseURI);
    if (eFileInfo) {
      eFileInfo.lastAccessTime = (new Date ()).getTime ();
      
      if (refName) {
        if (refName in eFileInfo.refParts) {
          part = eFileInfo.refParts [refName];
        }
      }
      else {
        part = eFileInfo.startPart;
      }
    }      
    
    return [eFileInfo, part];
  },

  /**
   * unmht 形式の URI からの相対パスを展開する
   *
   * @param  String uri
   *         unmht: 形式のファイルの URI
   * @param  String spec
   *         相対パス
   * @return String
   *         ファイルの URI
   *         見付からなかった場合は null
   */
  resolveURI : function (uri, spec) {
    if (!spec) {
      return uri;
    }
    
    let [eFileInfo, part] = this.getFileInfoAndPart (uri);
    if (eFileInfo && part) {
      eFileInfo.lastAccessTime = (new Date ()).getTime ();
      
      let [targetPart, fragment]
        = this.findPart (eFileInfo, part, spec, false, true);
      if (targetPart) {
        return eFileInfo.baseURI + targetPart.eParam.refName + fragment;
      }
      
      if (!part.eParam.location.match (/^file/)) {
        try {
          return Services.io.newURI (part.eParam.location, null, null)
            .resolve (spec);
        }
        catch (e) {
        }
      }
    }
  
    return null;
  },

  /**
   * mht に含まれるファイルの URI を取得する
   * 他のアドオンに提供する API
   *
   * @param  nsIURI contentLocation
   *         対象の URI
   * @param  nsIURI requestOrigin
   *         呼び出し元の URI
   *         unmht 形式
   * @return nsIURI
   *         ファイルの URI
   *         存在しなかった場合は null
   */
  getMHTFileURI : function (contentLocation, requestOrigin) {
    let [eFileInfo, part] = this.getFileInfoAndPart (requestOrigin.spec);
    
    if (eFileInfo && part) {
      eFileInfo.lastAccessTime = (new Date ()).getTime ();
      
      let [targetPart, fragment]
        = this.findPart (eFileInfo, part, contentLocation.spec, false, false);
      if (targetPart) {
        return Services.io.newURI
          (param.baseURI + targetPart.eParam.refName + fragment,
           null, null);
      }
    }
    
    return null;
  },
  
  /**
   * キャッシュから展開情報の配列を返す
   *
   * @return Array
   *         展開情報
   *         [UnMHTExtractFileInfo, ...]
   */
  getFileInfoList : function () {
    return [eFileInfo for ([uri, eFileInfo] in Iterator (this._eFileInfoList))]
    .sort (function (x, y) {
        return y.lastAccessTime - x.lastAccessTime;
      });
  },
  
  /**
   * キャッシュから古い展開情報を削除する
   *
   * @param  Number expire
   *         有効期限 [ms]
   */
  removeOldFileInfo : function (expire) {
    let now = (new Date ()).getTime ();
    for (let [uri, eFileInfo] in Iterator (this._eFileInfoList)) {
      if (now < eFileInfo.lastAccessTime + expire) {
        continue;
      }
      
      eFileInfo.destruct ();
      
      delete this._eFileInfoList [uri];
    }
  },
  
  /**
   * mht への参照のチェックおよび
   * mht 内の URI を置き換える
   *
   * @param  Number contentType
   *         コンテントの種類
   * @param  nsIURI contentLocation
   *         対象の URI
   * @param  nsIURI requestOrigin
   *         参照元の URI
   * @param  nsIDOMElement context
   *         ロード先
   * @return Boolean
   *         許可するかどうか
   */
  shouldLoad : function (contentType, contentLocation, requestOrigin, context) {
    if (!requestOrigin) {
      /* 呼び出し元が無い場合チェック不可能なので許可する */
      return true;
    }
    
    if (contentLocation.scheme == "chrome"
        || contentLocation.scheme == "data") {
      /* chrome, data への参照は対象外 */
      return true;
    }
    
    if (contentLocation.scheme == "unmht"
        && requestOrigin.scheme != "file"
        && requestOrigin.scheme != "unmht"
        && requestOrigin.scheme != "chrome"
        && requestOrigin.scheme != "moz-nullprincipal") {
      let param = UnMHTScheme.getUnMHTURIParam (contentLocation.spec);
      if (param.scheme == "file") {
        /* file, unmht, chrome 以外からの
         * unmht を経由した file への参照は拒否する */
        return false;
      }
    }
    
    if (requestOrigin.scheme != "unmht") {
      /* unmht 以外からの参照は対象外 */
      return true;
    }
    /* ---- ここから unmht からの参照 ---- */
    
    if (!context) {
      /* ロード先が無い場合 URI の置換えができないので許可する */
      return true;
    }
    
    if (context instanceof Ci.nsIDOMXULElement) {
      /* タブを開く場合対象外 */
      return true;
    }
    /* ---- ここからドキュメント内へのロード ---- */
    
    if (this._loadImageType == PERM_IMAGE_BLOCK_ALL
        && contentType == Ci.nsIContentPolicy.TYPE_IMAGE) {
      /* 画像の読み込みの拒否は Firefox の設定に従う */
      return false;
    }
    
    if (context instanceof Ci.nsIDOMHTMLDocument
        || context instanceof Ci.nsIDOMHTMLStyleElement) {
      /* CSS によるロードの場合 */
      return this._checkExternal (contentLocation);
    }
    /* ---- ここから HTML 要素へのロード ---- */
    
    let [eFileInfo, originPart]
    = this.getFileInfoAndPart (requestOrigin.spec);
    if (!(eFileInfo && originPart)) {
      /* 参照元のファイルが未展開、もしくは存在しないので拒否 */
      return false;
    }
    
    if (UnMHTScheme.isUnMHTURI (contentLocation.spec)) {
      let param = UnMHTScheme.getUnMHTURIParam (contentLocation.spec);
      if (param.original == eFileInfo.original) {
        /* 既に同じ mht ファイル内への参照になっているので許可 */
        return true;
      }
    }
    
    /**
     * 属性を変更する関数
     *
     * @param  nsIDOMElement node
     *         対象の要素
     * @param  String attr
     *         属性名
     * @param  String value
     *         属性値
     */
    let setAttr = function (node, attr, value) {
      if (node instanceof Ci.nsIDOMHTMLEmbedElement
          || node instanceof Ci.nsIDOMHTMLObjectElement) {
        /* EMBED, OBJECT は属性を変えても効果が無いので
         * 要素を置き換える */
        node.ownerDocument.defaultView.setTimeout
        (function () {
          try {
            let newNode = node.cloneNode (true);
            newNode.setAttribute ("__unmht_" + attr,
                                  contentLocation.spec);
            newNode.setAttribute (attr, value);
            node.parentNode.replaceChild (newNode, node);
          }
          catch (e) {
          }
        }, 100);
      }
      else {
        node.setAttribute ("__unmht_" + attr, contentLocation.spec);
        node.setAttribute (attr, value);
          
        node.ownerDocument.defaultView.setTimeout
        (function () {
          try {
            node.setAttribute (attr, value);
          }
          catch (e) {
          }
        }, 100);
      }
    }.bind (this);
    
    /**
     * 属性を削除する関数
     *
     * @param  nsIDOMElement node
     *         対象の要素
     * @param  String attr
     *         属性名
     */
    let removeAttr = function (node, attr) {
      node.removeAttribute (attr);
      node.ownerDocument.defaultView.setTimeout
      (function () {
        try {
          node.removeAttribute (attr);
        }
        catch (e) {
        }
      }, 100);
    };
    
    let attr = "src";
    if (context instanceof Ci.nsIDOMHTMLObjectElement) {
      attr = "data";
    }
    
    let [targetPart, fragment]
    = this.findPart (eFileInfo, originPart, contentLocation.spec, false, false);
    if (targetPart) {
      setAttr (context, attr,
               eFileInfo.baseURI + targetPart.eParam.refName + fragment); 
      return false;
    }
    
    if (contentLocation.scheme == "file") {
      /* file への読み込みは削除 */
      removeAttr (context, attr);
      return false;
    }
    
    if (contentLocation.spec.indexOf (requestOrigin.spec) == 0) {
      /* JavaScript 等で生成された参照元からの相対アドレス */
      /* **** FIXME: パスを遡るものはチェック不可 **** */
      let path = contentLocation.spec.substr (requestOrigin.spec.length);
      
      let origSpec;
      if (path.match (/^\?/)) {
        /* クエリのみ */
        origSpec
          = originPart.eParam.location.replace (/\?.*$/, "") + path;
      }
      else {
        origSpec
          = UnMHTFileUtil.jointPath (originPart.eParam.baseDir, path);
      }
      
      let origContentLocation;
      try {
        origContentLocation = Services.io.newURI (origSpec, null, null)
      }
      catch (e) {
        return this._checkExternal (contentLocation);
      }
      
      if (origContentLocation.spec != contentLocation.spec) {
        /* 本来のアドレスでチェックし直す */
        return this.shouldLoad (contentType, origContentLocation,
                                requestOrigin, context);
      }
    }
    
    return this._checkExternal (contentLocation);
  },

  /* ==== private ==== */
  
  _eFileInfoList : new Object (), /* Obejct  展開情報を保持しておく
                                   *   <String 正規化した baseURI,
                                   *    UnMHTExtractFileInfo 展開情報> */
  
  _loadImageType : PERM_IMAGE_ALLOW_ALL, /* Number  Firefox の
                                          *         画像読み込み設定 */
  
  _enableOffline : false, /* Boolean  オフラインモード */
  
  /**
   * 初期化処理
   */
  init : function () {
    let f;
    
    f = function () {
      this._enableOffline = UnMHTPref.getPref ("open.offline");
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "open.offline", f, this);
    f ();
    
    f = function () {
      this._loadImageType
      = UnMHTPref.prefBranch.getIntPref ("permissions.default.image");
    }.bind (this);
    UnMHTPref.addListener ("permissions.default.image", f, this);
    f ();
  },
  
  /**
   * 終了処理
   */
  term : function () {
    UnMHTPref.removeAllListeners (this);
    
    this.removeOldFileInfo (0);
  },
  
  /**
   * パートの情報を作成する
   *
   * @param  UnMHTExtractFileInfo eFileInfo
   *         展開情報
   * @param  UnMHTMIMEPart part
   *         対象のパート
   * @param  UnMHTMIMEPart parentPart
   *         親のパート
   */
  _createExtractParam : function (eFileInfo, part, parentPart) {
    let isMixed = (part.isMultipart && part.contentSubType == "mixed");
    
    if (part.isMultipart) {
      part.parts.forEach
        (function (p) {
          this._createExtractParam (eFileInfo, p, part);
        }.bind (this));
      
      if (!isMixed) {
        return;
      }
    }
    
    let eParam = new UnMHTExtractParam ();
    if (part.contentID) {
      eParam.cid = part.contentID;
    }
    if (isMixed) {
      eParam.isMixed = true;
      eParam.content = "";
      
      eParam.mimetype = "text/html";
      eParam.charset = "utf-8";
      
      eParam.isHTML = true;
      
      eParam.location = "mixed-" + (new Date ()).getTime () + ".html";
    }
    else {
      eParam.content = part.body;
    
      eParam.mimetype = part.mimetype;
      eParam.charset = part.charset;
      if (part.contentType == "text") {
        eParam.isHTML = true;
        if (part.contentSubType == "css") {
          eParam.isCSS = true;
        }
      }
    
      if ("content-base" in part.fields) {
        let value = part.fields ["content-base"];
        try {
          value = UnMHTUnicodeConverter.convertToUnicode (value, "utf-8");
        }
        catch (e) {
          try {
            value = UnMHTUnicodeConverter.convertToUnicode (value, "shift_jis");
          }
          catch (e) {
          }
        }
        eParam.contentBase = value;
      }
    
      if (eParam.contentBase) {
        if (part.contentLocation) {
          eParam.location
          = UnMHTFileUtil.jointPath (eParam.contentBase, part.contentLocation);
        }
        else {
          eParam.location = eParam.contentBase;
        }
      }
      else {
        eParam.location = part.contentLocation;
      }
      
      if (parentPart) {
        eParam.location
        = UnMHTFileUtil.jointPath (parentPart.contentLocation, eParam.location);
      }
      
      if (!eParam.location
          && eParam.mimetype == "text/html") {
        /* 開始ページの Content-Location が
         * 指定されない場合があるので index.html とする */
        eParam.location = "index.html";
      }
    }
    part.eParam = eParam;
    
    this._getPathInfo (part);
    
    eFileInfo.parts.push (part);
    
    if (eParam.mimetype == "application/octet-stream"
        && part.body.substr (0, 32).indexOf ("<") != -1) {
      /* CGI 生成のページ等が application/octet-stream として
       * 保存されている場合がある */
      if (part.body.search (/<html/i) != -1) {
        eParam.mimetype = "text/html";
        eParam.isHTML = true;
      }
    }
  },
  
  /**
   * パスの情報を取得する
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   */
  _getPathInfo : function (part) {
    let leafNameRe = null;
    let eParam = part.eParam;
    
    let location = eParam.location;
    let fragment = "";
    
    let re = location.match (/^([^#]*)(#.+)$/);
    if (re) {
      [, location, fragment] = re;
    }
    
    let locationWithoutQuery = location;
    let query = "";
    re = locationWithoutQuery.match (/^([^\?]*)(\?.+)$/);
    if (re) {
      [, locationWithoutQuery, query] = re;
    }
    
    if (locationWithoutQuery.indexOf ("mhtml:") == 0) {
      /* mhtml で始まる場合 (IE のプロトコル ?)
       * ! の前後に分割する */
      leafNameRe = /^(.*)\!([^\!]+)$/;
    }
    else if (locationWithoutQuery.indexOf ("\\") != -1) {
      leafNameRe = /^(.*)\\([^\\]*)$/;
    }
    else if (locationWithoutQuery.indexOf ("/") != -1) {
      leafNameRe = /^(.*)\/([^\/]*)$/;
    }
    else if (locationWithoutQuery.indexOf (":") != -1) {
      leafNameRe = /^(.*):([^:]*)$/;
    }
    
    if (leafNameRe != null) {
      re = locationWithoutQuery.match (leafNameRe);
      if (re) {
        [, eParam.baseDir, eParam.leafName] = re;
      }
      else {
        eParam.baseDir = locationWithoutQuery;
        eParam.leafName = "";
      }
    }
    else {
      eParam.baseDir = "";
      eParam.leafName = locationWithoutQuery;
    }
    
    if (part.contentDispositionFilename) {
      eParam.leafName = part.contentDispositionFilename;
    }
    
    eParam.referredBaseDir = "";
    
    eParam.fragment = fragment;
    
    eParam.normalLocation = this.normalizeName (location);
    eParam.normalLeafName = this.normalizeName (eParam.leafName);
    eParam.normalFragment = this.normalizeName (eParam.fragment);
  },
  
  /**
   * 参照名を重複しないように設定する
   *
   * @param  UnMHTExtractFileInfo eFileInfo
   *         展開情報
   */
  _setRefName : function (eFileInfo) {
    let indices = {};
    
    eFileInfo.parts.forEach (function (part, i) {
        let refBase;
        let extension = "";
        
        let re = part.eParam.leafName.match (/^(.+)\.([A-Za-z0-9]+)$/);
        if (re) {
          [, refBase, extension] = re;
        }
        else if (part.eParam.leafName) {
          refBase = part.eParam.leafName;
        }
        else if (part.contentID) {
          refBase = part.contentID;
        }
        else if (part.eParam.mimetype == "text/html") {
          refBase = "index";
          extension = "html";
        }
        else {
          refBase = i;
        }
        if (extension) {
          extension = "." + extension;
        }
        
        refBase = encodeURIComponent (refBase).replace (/\//g, "_");
        
        let refName = refBase + extension;
        
        if (refName in indices) {
          part.eParam.refName = refBase + "." + indices [refName] + extension;
          indices [refName] ++;
        }
        else {
          part.eParam.refName = refName;
          indices [refName] = 1;
        }
      }.bind (this));
  },
  
  _cssRe : /(@import\s*(?:url\s*\(\s*)?|[^a-zA-Z0-9_]url\s*\(\s*)(?:(\"|&quot;|&#x22;|&#34;)((?:\\\"|[^\"])*)\2|(\'|&apos;|&#x27;|&#39;)((?:\\\'|[^\'])*)\4|([^\"\';\(\)]+))/ig,
  
  /**
   * CSS 中の URI を unmht: 形式に変換する
   * replace に渡す関数
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   * @param  String matched
   *         マッチした全体
   * @param  String prev
   * @param  String dquote
   * @param  String dqpath
   * @param  String quote
   * @param  String qpath
   * @param  String path
   *         以下参照
   *
   *         dquote
   *         -
   * @import "12345.jpg";
   * -------- ---------
   * prev     dqpath
   *
   *      quote
   *      -
   * url ('12345.jpg');
   * ----- ---------
   * prev  qpath
   *
   * @return String
   *         変換した結果
   */
  _cssFunc : function (eFileInfo, part,
                       matched, prev, dquote, dqpath, quote, qpath, path) {
    path = dqpath || qpath || path;
    quote = quote || dquote || "";
    if (!prev.match (/\(/)) {
      quote = quote || "\'";
    }
    
    if (!path.match (/^[A-Za-z0-9\-]+:/)) {
      /* 相対パスは UnMHTProtocolHandler.newURI の方で変える */
      return matched;
    }
    
    let [targetPart, fragment]
    = this.findPart (eFileInfo, part, path, part.eParam.isCSS, true);
    if (targetPart) {
      targetPart.eParam.referredBaseDir = part.eParam.baseDir;
      return prev + quote
        + eFileInfo.baseURI + targetPart.eParam.refName + quote;
    }
    
    if (path.match (/^file:/)) {
      return prev + quote + quote;
    }
    
    return matched;
  },
  
  /**
   * CSS 中の unmht: 形式の URI を解除する
   * replace に渡す関数
   *
   * 引数と返り値は part が無い以外は _cssFunc と同じ
   */
  _unCssFunc : function (eFileInfo,
                         matched, prev, dquote, dqpath, quote, qpath, path) {
    path = dqpath || qpath || path;
    quote = quote || dquote || "";
    if (!prev.match (/\(/)) {
      quote = quote || "\'";
    }
    
    if (!path.match (/^unmht:/)) {
      return matched;
    }
    
    let param = UnMHTScheme.getUnMHTURIParam (path);
    let baseURI = param.baseURI;
    let refName = param.refName;
    let part = null;
    if (refName) {
      if (refName in eFileInfo.refParts) {
        part = eFileInfo.refParts [refName];
      }
    }
    else {
      part = eFileInfo.startPart;
    }
    if (part) {
      let location = part.contentLocation;
      if (!location) {
        location = "cid:" + part.eParam.cid;
      }
      
      return prev + quote + location + quote;
    }
    
    return matched;
  },
  
  /**
   * CSS 中の URI を unmht: 形式に変換する
   *
   * @param  UnMHTExtractFileInfo eFileInfo
   *         展開情報
   */
  _modifyCSS : function (eFileInfo) {
    eFileInfo.parts.forEach (function (part) {
        if (part.eParam.isHTML
            || part.eParam.isCSS) {
          part.eParam.content
          = part.eParam.content
          .replace (this._cssRe,
                    function (matched, prev,
                              dquote, dqpath, quote, qpath, path) {
                      return this._cssFunc
                      (eFileInfo, part, matched, prev,
                       dquote, dqpath, quote, qpath, path);
                    }.bind (this));
        }
      }.bind (this));
  },
  
  _elementRe : /(<([\?A-Za-z0-9_:\-]+))(\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)(>)/g,
  _attributeRe : /(\s(class|src|href|background|action|data)\s*=\s*)(?:(\")((?:\\\"|[^\"])*)\"|(\')((?:\\\'|[^\'])*)\'|([^\"\'\\ >]+))/ig,
  /**
   * HTML 中の URI を unmht: 形式に変換する
   * replace に渡す関数
   *
   * @param  UnMHTExtractFileInfo eFileInfo
   *         展開情報
   * @param  UnMHTMIMEPart part
   *         対象のパート
   * @param  String matched
   *         マッチした全体
   * @param  String prev
   * @param  String attrs
   * @param  String next
   * @return String
   *         変換した結果
   */
  _elementFunc : function (eFileInfo, part,
                           matched, prev, name, attrs, next) {
    if (name.match (/^base$/i)) {
      return matched;
    }
    let isObject = false;
    if (name.match (/^object$/i) || name.match (/^embed$/i)) {
      isObject = true;
    }
    let isAnchor = false;
    if (name.match (/^a$/i)) {
      isAnchor = true;
    }
    let isLink = false;
    if (name.match (/^link$/i)) {
      isLink = true;
    }
    let isOriginalLink = false;
    
    attrs
    = attrs.replace
    (this._attributeRe,
     function (matched, prev, attr,
                dquote, dqpath, quote, qpath, path) {
      path = dqpath || qpath || path;
      quote = quote || dquote || "\"";
      
      if (attr == "class") {
        if (isAnchor && path == "unmht_link_to_original") {
          isOriginalLink = true;
        }
        
        return matched;
      }
      
      if (isOriginalLink && attr == "href") {
        return matched;
      }
      
      if (path.match (/^mailto:/)) {
        return matched;
      }
       
      if (!isObject && attr.match (/data/i)) {
        return matched;
      }
      
      let isRelative = false;
      if (!path.match (/^[A-Za-z0-9\-]+:/)) {
        isRelative = true;
      }
      if (isRelative && !isAnchor) {
        /* 相対パスは UnMHTProtocolHandler.newURI の方で変える */
        return matched;
      }
      
      let [targetPart, fragment]
      = this.findPart (eFileInfo, part, path, false, !isAnchor);
      if (targetPart) {
        if (isLink) {
          targetPart.eParam.referredBaseDir = part.eParam.baseDir;
        }
        
        return prev + quote
          + eFileInfo.baseURI + targetPart.eParam.refName + fragment
          + quote;
      }
    
      if (path.match (/^file:/)
          || path.match (/^cid:/)) {
        return prev + quote + quote;
      }
      
      if (isRelative && isAnchor) {
        /* リンク先はあらかじめ展開しておく */
        try {
          return prev + quote
          + Services.io.newURI (part.eParam.location, null, null)
          .resolve (path) + quote;
        }
        catch (e) {
        }
      }
    
      return matched;
    }.bind (this));
    
    return prev + attrs + next;
  },
  /**
   * HTML 中の unmht: 形式の URI を解除する
   * この際 BASE 要素の内容を baseDir にする
   * replace に渡す関数
   *
   * 引数と返り値は part が無い以外は _elementFunc と同じ
   */
  _unElementFunc : function (eFileInfo, matched, prev, name, attrs, next) {
    let isBase = false;
    if (name.match (/^base$/i)) {
      isBase = true;
    }
    
    attrs
    = attrs.replace
    (this._attributeRe,
     function (matched, prev, attr,
                dquote, dqpath, quote, qpath, path) {
      path = dqpath || qpath || path;
      quote = quote || dquote || "\"";
      
      if (!path.match (/^unmht:/)) {
        return matched;
      }
      
      let param = UnMHTScheme.getUnMHTURIParam (path);
      let refName = param.refName;
      let part = null;
      if (refName) {
        if (refName in eFileInfo.refParts) {
          part = eFileInfo.refParts [refName];
        }
      }
      else {
        part = eFileInfo.startPart;
      }
      if (part) {
        let location = part.contentLocation;
        if (isBase) {
          location = part.eParam.baseDir;
        }
        if (!location && !isBase) {
          location = "cid:" + part.eParam.cid;
        }
         
        return prev + quote + location + quote;
      }
       
      return matched;
    }.bind (this));
    
    return prev + attrs + next;
  },
  
  _baseRe : /<base\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*?href\s*=\s*(?:\"((?:\\\"|[^\"])*)\"|\'((?:\\\'|[^\'])*)\'|([^\"\'\\ >]+))(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*>/ig,
  
  /**
   * HTML 中の BASE の値を収集して削除する
   * replace に渡す関数
   *
   * 引数と返り値は _elementFunc と同じ
   */
  _baseFunc : function (eFileInfo, part,
                       matched, dqpath, qpath, path) {
    path = dqpath || qpath || path;
    
    part.eParam.baseDir
    = UnMHTFileUtil.jointPath
    (part.eParam.baseDir,
     UnMHTUnicodeConverter.convertToUnicode (path, part.eParam.charset)
     /* BASE 要素にディレクトリではなくファイルのパスが入っている場合が
      * あるので、末尾を削除する */
     .replace (/\/[^\/]+$/, ""))
    
    return "";
  },
  
  /**
   * HTML 中の URI を unmht: 形式に変換する
   * また、BASE 要素と Office 用の要素を削除し
   * 設定されていれば PowerPoint の警告ページをスキップする
   *
   * @param  UnMHTExtractFileInfo eFileInfo
   *         展開情報
   */
  _modifyAttribute : function (eFileInfo) {
    let isPPT = false;
    let framePart = null;
    
    eFileInfo.parts.forEach (function (part) {
        if (part.eParam.refName == "frame.htm") {
          framePart = part;
        }
          
        if (!part.eParam.isCSS && part.eParam.isHTML) {
          part.eParam.content
            = part.eParam.content
            .replace (this._baseRe,
                      function (matched, dqpath, qpath, path) {
                        return this._baseFunc
                        (eFileInfo, part, matched, dqpath, qpath, path);
                      }.bind (this))
            .replace (this._elementRe,
                      function (matched, prev, name, attrs, next) {
                        return this._elementFunc
                        (eFileInfo, part, matched, prev, name, attrs, next);
                      }.bind (this));
            
          if (part.eParam.content.indexOf
              ("xmlns:v=\"urn:schemas-microsoft-com:vml\"") != -1) {
            part.eParam.content
              = part.eParam.content
              .replace (/<\/?(o|p|oa|v):(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*?>/g, "");
            isPPT = true;
          }
        }
      }.bind (this));
    
    if (isPPT && framePart
        && UnMHTPref.getPref ("open.ppt.skip_warning")) {
      eFileInfo.startPart = framePart;
    }
  },
  
  /**
   * multipart/mixed なドキュメントを作成する
   *
   * @param  UnMHTExtractFileInfo eFileInfo
   *         展開情報
   */
  _createMixedDocument : function (eFileInfo) {
    let createBody = function (part) {
      if (!part) {
        return "";
      }
      let toUTF8 = function (text) {
        return UnMHTUnicodeConverter.convertFromUnicode
        (text, "utf-8");
      };
        
      let getUTF8Content = function (part) {
        let content = part.eParam.content;
          
        if (!part.eParam.charset.match (/^utf-8$/i)) {
          content
            = UnMHTUnicodeConverter.convertFromUnicode
            (UnMHTUnicodeConverter.convertToUnicode
             (content, part.eParam.charset),
             "utf-8");
        }
          
        return content;
      };
      
      let content;
      if (part.eParam.mimetype == "text/html") {
        content
          = getUTF8Content (part)
          .replace (/<\/?(html|body)(?:\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)?>/ig, "")
          .replace (/<![\?A-Za-z0-9_:\-]+(?:\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)?>/ig, "");
      }
      else if (part.eParam.mimetype == "text/plain") {
        content
        = "<pre>"
        + UnMHTHTMLUtil.escapeEntity (getUTF8Content (part))
        .replace (/[ \t]/g, "&nbsp;")
        + "</pre>";
      }
      else {
        let href
        = UnMHTHTMLUtil.escapeEntity
        (eFileInfo.baseURI + part.eParam.refName);
        let filename
        = UnMHTHTMLUtil.escapeEntity
        (part.eParam.leafName || part.contentDispositionFilename || "???");
        
        if (part.contentType == "image") {
          content
            = "<div style=\"text-align: center;\">"
            + "<img src=\"" + href + "\" alt=\"" + filename + "\" />"
            + "</div>"
        }
        else {
          content = "<a href=\"" + href + "\">" + filename + "</a>";
        }
        
        content = toUTF8 (content);
      }
      
      if (part.contentDispositionFilename) {
        content
        = "<fieldset style=\"border: 1px solid graytext;\">"
        + "<legend style=\"color: graytext; font-size: 80%;\">"
        + toUTF8 (part.contentDispositionFilename)
        + "</legend>"
        + content
        + "</fieldset>";
      }
      return content;
    }.bind (this);
    
    eFileInfo.parts.forEach (function (part) {
        if (part.eParam.isMixed) {
          part.eParam.content
            = part.parts
            .map (function (p) createBody (UnMHTMIMEUtil.findStartPart (p)))
            .join ("");
        }
      }.bind (this));
  },
  
  /**
   * パートの情報を収集する
   *
   * @param  UnMHTExtractFileInfo eFileInfo
   *         展開情報
   */
  _getPartsInfo : function (eFileInfo) {
    eFileInfo.parts.forEach (function (part) {
        eFileInfo.refParts [part.eParam.refName] = part;
          
        eFileInfo.size += part.eParam.content.length;
      }.bind (this));
  },
  
  /**
   * キャッシュに展開情報を格納する
   *
   * @param UnMHTExtractFileInfo eFileInfo
   *         展開情報
   */
  _addFileInfo : function (eFileInfo) {
    this._eFileInfoList [this.normalizeName (eFileInfo.baseURI)] = eFileInfo;
  },
  
  /**
   * キャッシュから展開情報を返す
   *
   * @param  String baseURI
   *         展開情報の baseURI
   * @return UnMHTExtractFileInfo
   *         展開情報
   *         見付からなければ null
   */
  _getFileInfo : function (baseURI) {
    baseURI = this.normalizeName (baseURI)
    
    if (baseURI in this._eFileInfoList) {
      return this._eFileInfoList [baseURI];
    }
    
    return null;
  },
  
  /**
   * 外部読み込みをチェックする
   *
   * @param  nsIURI contentLocation
   *         対象の URI
   * @return Boolean
   *         許可するかどうか
   */
  _checkExternal : function (contentLocation) {
    if (contentLocation.scheme == "unmht") {
      return true;
    }
    
    return !this._enableOffline;
  }
};
UnMHTExtractor.init ();
UnMHTShutdown.addListener (function (data, reason) {
    UnMHTExtractor.term ();
  });
