/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTMIMEUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTSave.jsm");
XPCOMUtils.importRelative (this, "UnMHTTimer.jsm");
XPCOMUtils.importRelative (this, "UnMHTVersion.jsm");

let EXPORTED_SYMBOLS = ["UnMHTFileConverter"];

/**
 * 変換コンテキスト
 *
 * @param  Array files
 *         対象の HTML ファイル
 *         [nsIFile, ...]
 * @param  String suffix
 *         ディレクトリ名の末尾
 * @param  UnMHTFileConverterListener listener
 *         イベントリスナ
 */
function UnMHTFileConvertContext (files, suffix, listener) {
  this._files = files;
  this._suffix = suffix;
  this._listener = listener;
}
UnMHTFileConvertContext.prototype = {
  progress : 0,           /* Number  進捗状況 (0-1) */
  
  /**
   * 破棄
   * start 後は自分で呼ぶ
   */
  destruct : function () {
    this._files = null,
    this._listener = null;
  },
  
  /**
   * 変換開始
   */
  start : function () {
    this._currentIndex = -1;
    
    this._setNextTimer ();
  },
  
  /* ==== private ==== */
  
  _currentIndex : 0, /* Number  files 中の現在の位置 */
  
  _files : null,    /* Array  対象の HTML ファイル
                    *   [nsIFile, ...] */
  _suffix : "",     /* String  ディレクトリ名の末尾 */
  _listener : null, /* UnMHTFileConverterListener  イベントリスナ */
  
  _count : 0,       /* Number  変換したファイル数 */
  
  /**
   * 次のタイマーをセットする
   */
  _setNextTimer : function () {
    UnMHTTimer.setTimeout
    (function () {
      this._convertNext ();
    }.bind (this), 100);
  },
  
  /**
   * 次のファイルを変換する
   */
  _convertNext : function () {
    this._currentIndex ++;
    
    if (this._currentIndex >= this._files.length) {
      this.progress = 1;
      this._listener.onComplete (this, this._count);
      this.destruct ();
      return;
    }
    
    let file = this._files [this._currentIndex];
    
    if (file.path.match (/\.mht(ml)?$/i)) {
      this._setNextTimer ();
      return;
    }
    
    this.progress = this._currentIndex / this._files.length;
    this._listener.onStartFile (this, file);
    
    let baseDir = file.parent.clone ();
    let filenameBase = file.leafName.replace (/\.[^\.]+$/, "");
    
    let findDir = function (suffix) {
      let d = baseDir.clone ();
      d.append (filenameBase + suffix);
      if (d && d.exists () && d.isDirectory ()) {
        return d;
      }
      return null;
    };
    
    let dir
    = findDir (this._suffix) || findDir ("_files") || findDir (".files");
    if (!dir) {
      this._setNextTimer ();
      return;
    }
    
    let resourceFiles
    = UnMHTFileUtil.getDirectoryEntries (dir).filter (function (f) f.isFile ());
    
    let filename
    = UnMHTSave.getUniqueFilename (baseDir, filenameBase);
    
    let mhtFile = baseDir.clone ();
    mhtFile.append (filename);
    
    if (UnMHTFileConverter.convertFile
        (mhtFile, file, dir.leafName, resourceFiles, file.leafName)) {
      this._count ++;
    }
    
    this._setNextTimer ();
  }
};

/**
 * HTML ファイルを MHT ファイルに変換
 */
let UnMHTFileConverter = {
  /**
   * 変換コンテキストを作成する
   *
   * @param  Array files
   *         対象の HTML ファイル
   *         [nsIFile, ...]
   * @param  String suffix
   *         ディレクトリ名の末尾
   * @param  UnMHTFileConverterListener listener
   *         イベントリスナ
   * @return UnMHTFileConvertContext
   */
  createContext : function (files, suffix, listener) {
    return new UnMHTFileConvertContext (files, suffix, listener);
  },
  
  /**
   * HTML ファイルを MHT に変換
   *
   * @param  nsIFile mhtFile
   *         保存先のファイル
   * @param  nsIFile htmlFile
   *         対象の HTML ファイル
   * @param  String dirname
   *         他のリソースファイルを格納したディレクトリ名
   * @param  Array resourceFiles
   *         他のリソースファイル
   *         [nsIFile, ...]
   * @param  String subject
   *         Subject の値
   * @return Boolean
   *         変換に成功したか
   */
  convertFile : function (mhtFile, htmlFile, dirname, resourceFiles, subject) {
    let fstream = UnMHTFileUtil.openOutputStream (mhtFile);
    
    let topPart = new UnMHTMIMEPart ();
    topPart.isMultipart = true;
    topPart.boundary = UnMHTMIMEUtil.createUniqueBoundary ();
    topPart.startMimetype = "text/html";
    
    UnMHTMIMEUtil.addField (topPart, "From",
                            "<Saved by UnMHT Ver." + UnMHTVersion + ">");
    UnMHTMIMEUtil.addField (topPart, "Subject", subject);
    UnMHTMIMEUtil.addField (topPart, "Date",
                            UnMHTMIMEUtil.formatDate (new Date ()));
    UnMHTMIMEUtil.addField (topPart, "MIME-Version", "1.0");
    UnMHTMIMEUtil.addField (topPart, "Content-Type",
                            "multipart/related"
                            + ";boundary=\"" + topPart.boundary + "\""
                            + ";type=\"" + topPart.startMimetype + "\"");
    topPart.parts = [];
    
    [htmlFile].concat (resourceFiles).forEach (function (file) {
        let ext = "";
        let re = file.leafName.match (/\.([^\.]+)(\?.+)?(#.+)?$/i);
        if (re) {
          ext = re [1];
        }
        let mimetype = UnMHTMIMEUtil.getTypeFromExtension (ext);
        
        let location;
        if (file == htmlFile) {
          location = file.leafName;
        }
        else {
          location = dirname + "/" + file.leafName;
        }
        
        let part = new UnMHTMIMEPart ();
        part.contentTransferEncoding = "base64";
        UnMHTMIMEUtil.addField (part, "Content-Type",
                                mimetype);
        UnMHTMIMEUtil.addField (part, "Content-Transfer-Encoding",
                                part.contentTransferEncoding);
        UnMHTMIMEUtil.addField (part, "Content-Location", location);
        
        part.body = UnMHTFileUtil.readData (file);
      
        topPart.parts.push (part);
      });
    
    UnMHTMIMEUtil.encodeMessage (topPart, fstream);
    
    UnMHTMIMEUtil.destructPart (topPart);
    
    try {
      mhtFile.lastModifiedTime = htmlFile.lastModifiedTime;
    }
    catch (e) {
    }
        
    return true;
  }
};
