/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTAppInfo.jsm");
XPCOMUtils.importRelative (this, "UnMHTExtractor.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTNetFileLoader.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTPrototype.jsm");
XPCOMUtils.importRelative (this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative (this, "UnMHTShutdown.jsm");
XPCOMUtils.importRelative (this, "UnMHTUnicodeConverter.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowHook.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowUtil.jsm");

XPCOMUtils.importRelative (this, "UnMHTTimer.jsm");

let EXPORTED_SYMBOLS = ["UnMHTFileLoader",
                        "UnMHTFileLoadHook"];

/**
 * ファイルロード用フック
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 * @inherits   UnMHTWindowHook
 * @implements IUnMHTWindowHook
 */
function UnMHTFileLoadHook (window) {
  UnMHTWindowHook.call (this, window);
  
  this.onDOMContentLoaded = this.onDOMContentLoaded.bind (this);
}
UnMHTFileLoadHook.prototype = {
  /* ---- UI イベント ---- */
  
  /**
   * コンテンツがロードされたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onDOMContentLoaded : function (event) {
    UnMHTFileLoader.onDOMContentLoaded (this._window, this._tabbrowser, event);
  },
  
  /* ==== IUnMHTWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct : function () {
    this._tabbrowser = null;
    
    UnMHTPrototype.super (this, "destruct").call (this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param  Boolean isStartup
   *         スタートアップイベントか
   * @param  Number reason
   *         スタートアップの理由
   */
  load : function (isStartup, reason) {
    if (!UnMHTWindowUtil.isMailWindow (this._window)) {
      this._tabbrowser = UnMHTWindowUtil.getBrowser (this._window);
      if (!this._tabbrowser) {
        return;
      }
      
      let appcontent = this._window.document.getElementById ("appcontent");
      appcontent.addEventListener
      ("DOMContentLoaded", this.onDOMContentLoaded, false);
    }
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param  Boolean isShutdown
   *         シャットダウンイベントか
   * @param  Number reason
   *         シャットダウンの理由
   */
  unload : function (isShutdown, reason) {
    if (!UnMHTWindowUtil.isMailWindow (this._window)) {
      let appcontent = this._window.document.getElementById ("appcontent");
      appcontent.removeEventListener
      ("DOMContentLoaded", this.onDOMContentLoaded);
    }
  },
  
  /* ==== private ==== */
  
  _tabbrowser : null /* nsIDOMXULElement  対象のウィンドウのタブブラウザ */
};
UnMHTPrototype.createInheritance (UnMHTFileLoadHook, UnMHTWindowHook);

/**
 * MHT ファイルのロード
 */
let UnMHTFileLoader = {
  /**
   * ハンドル済みの mht ファイルを登録する
   *
   * @param  String location
   *         アドレス
   */
  registerHandled : function (location) {
    this._handled [location] = true;
  },
  
  /* ---- UI イベント ----- */
  
  /**
   * コンテンツがロードされたイベント
   * mht ファイルの展開、アドレスの変換を行う
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMXULElement tabbrowser
   *         対象のタブブラウザ
   * @param  Event event
   *         対象のイベント
   */
  onDOMContentLoaded : function (window, tabbrowser, event) {
    let targetWindow = event.target.defaultView;
    let targetDocument = targetWindow.document;
    let href = targetDocument.location.href;
    
    if (href in this._handled) {
      /* ハンドル済みの mht ファイル */
      delete this._handled [href];
    }
    else if (href.match (/[^\/]\.(mht(ml)?|eml|mbs)$/i)) {
      /* 未ハンドルの mht ファイル */
      
      if (href.match (/^chrome:\/\/ietab\//)) {
        /* IETab がハンドル済み */
        return;
      }
        
      this.loadMHT
        (window, Services.io.newURI (href, null, null), targetWindow,
         true,  /* 開いているドキュメントを使用する */
         true,  /* 履歴を置き換える */
         true,  /* 展開する */
         "",    /* 開始ページを開く */
         true); /* キャッシュを使用する */
    }
    else if (UnMHTScheme.isUnMHTURI (href)) {
      let href = targetDocument.location.href;
      
      let param = UnMHTScheme.getUnMHTURIParam (href);
      let original = param.original;
      let scheme = param.scheme;
      
      if (scheme != "cache"
          && scheme != "cid"
          && scheme != "dummy"
          && scheme != "empty"
          && scheme != "error") {
        /* 通常のファイル */
        let [eFileInfo, part] = UnMHTExtractor.getFileInfoAndPart (href);
        if (eFileInfo) {
          if (part) {
            if (part == eFileInfo.startPart) {
              this._hookReload (window, targetWindow);
            }
          }
        }
        else {
          /* 未展開のファイルだった場合 */
          let uri;
          try {
            uri = Services.io.newURI (original, null, null);
          }
          catch (e) {
            return;
          }
        
          this.loadMHT
            (window, uri, targetWindow,
             false,    /* ドキュメント無し */
             true,     /* 履歴を置き換える */
             true,     /* 展開する */
             original, /* 指定されたファイルを開く */
             false);   /* キャッシュは使用しない */
        }
      }
      
      if (this._openMultiple == "tabs"
          || this._openMultiple == "first") {
        let [eFileInfo, part] = UnMHTExtractor.getFileInfoAndPart (href);
        if (eFileInfo && part && part == eFileInfo.startPart) {
          if ("x-unmht-tabs" in eFileInfo.topPart.fields) {
            /* 複数のファイルを含む MHT の
             * インデックスを開いた */
            this._replaceTabs (window, tabbrowser, targetDocument);
          }
        }
      }
    }
    else {
      let re = href.match (/^unmht:\/\/([^\/]*)\/(.*)$/);
      if (re) {
        /* Ver.5.7.10 以前の URI の可能性がある */
        let [, host, path] = re;
        let newURI = "unmht://" + host + ".unmht" + "/" + path;
        
        if (UnMHTScheme.isUnMHTURI (newURI)) {
          targetDocument.location = newURI;
        }
      }
    }
  },
  
  /**
   * mht ファイルを展開して開く
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIURI uri
   *         開く URI
   * @param  nsIDOMWindow targetWindow
   *         対象のウィンドウ
   * @param  Boolean hasDocument
   *         ドキュメントに内容を持っているか
   * @param  Boolean replace
   *         履歴を置き換えるかどうか
   * @param  Boolean extract
   *         展開するか
   * @param  String refName
   *         対象のファイル
   * @param  Boolean useCache
   *         キャッシュを使用するか
   */
  loadMHT : function (window, uri, targetWindow,
                      hasDocument, replace, extract, refName,
                      useCache) {
    if (uri.scheme == "file") {
      /* file プロトコルの場合、そのファイルから開く */
      this.loadMHTFromFile
      (window, uri, targetWindow,
       hasDocument, replace, extract, refName,
       UnMHTFileUtil.getFileFromURLSpec (uri.spec));
    }
    else {
      /* それ以外の場合、ウェブから開く */
      this.loadMHTFromWeb
      (window, uri, targetWindow,
       hasDocument, replace, extract, refName,
       useCache);
    }
  },
  
  /**
   * ファイルから mht ファイルを展開して開く
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIURI uri
   *         開く URI (本来の URI)
   * @param  nsIDOMWindow targetWindow
   *         対象のウィンドウ
   * @param  Boolean hasDocument
   *         ドキュメントに内容を持っているか
   * @param  Boolean replace
   *         履歴を置き換えるかどうか
   * @param  Boolean extract
   *         展開するか
   * @param  String refName
   *         対象のファイル
   * @param  nsIFile targetFile
   *         開くファイル
   * @param  String refName
   *         対象のファイル
   */
  loadMHTFromFile : function (window, uri, targetWindow,
                              hasDocument, replace, extract, refName,
                              targetFile) {
    let data = UnMHTFileUtil.readData (targetFile);
    if (data == "") {
      if (hasDocument) {
        this.loadMHTFromDocument
          (window, uri, targetWindow,
           true, replace, extract, refName);
      }
      else {
        this.loadErrorPage (window, targetWindow, uri.spec);
      }
      return;
    }
    this.loadMHTFromData
    (window, uri, targetWindow,
     hasDocument, replace, extract, refName,
     data);
  },
  
  /**
   * データから mht ファイルを展開して開く
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIURI uri
   *         開く URI (本来の URI)
   * @param  nsIDOMWindow targetWindow
   *         対象のウィンドウ
   * @param  Boolean hasDocument
   *         ドキュメントに内容を持っているか
   * @param  Boolean replace
   *         履歴を置き換えるかどうか
   * @param  Boolean extract
   *         展開するか
   * @param  String refName
   *         対象のファイル
   * @param  String data
   *         対象のデータ
   */
  loadMHTFromData : function (window, uri, targetWindow,
                              hasDocument, replace, extract, refName,
                              data) {
    let eFileInfo
    = UnMHTExtractor.extractMHT (uri.spec, data, extract);
    
    if (eFileInfo) {
      let nav
        = targetWindow.QueryInterface (Ci.nsIInterfaceRequestor)
        .getInterface (Ci.nsIWebNavigation);
            
      let flag = 0;
      if (replace) {
        flag = Ci.nsIWebNavigation.LOAD_FLAGS_REPLACE_HISTORY;
      }
      
      let urlspec = eFileInfo.baseURI;
      if (refName
          && refName in eFileInfo.refParts) {
        urlspec += refName;
      }
      
      nav.loadURI (urlspec, flag, null, null, null);
    }
  },
  
  /**
   * ドキュメントから mht ファイルを展開して開く
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIURI uri
   *         開く URI
   * @param  nsIDOMWindow targetWindow
   *         対象のウィンドウ
   * @param  Boolean hasDocument
   *         ドキュメントに内容を持っているか
   * @param  Boolean replace
   *         履歴を置き換えるかどうか
   * @param  Boolean extract
   *         展開するか
   * @param  String refName
   *         対象のファイル
   */
  loadMHTFromDocument : function (window, uri, targetWindow,
                                  hasDocument, replace, extract, refName) {
    let data = "";
    Array.forEach
    (targetWindow.document.getElementsByTagName ("pre"),
     function (node) {
      node = node.firstChild;
      while (node) {
        data += node.nodeValue;
        node = node.nextSibling;
      }
    });
    
    if (data == "") {
      this.loadErrorPage (window, targetWindow, uri.spec);
      return;
    }
    
    data = UnMHTUnicodeConverter.convertFromUnicode (data, "utf-8");
    
    this.loadMHTFromData
    (window, uri, targetWindow,
     hasDocument, replace, extract, refName,
     data);
  },
    
  /**
   * ウェブのファイルから mht ファイルを展開して開く
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIURI uri
   *         開く URI
   * @param  nsIDOMWindow targetWindow
   *         対象のウィンドウ
   * @param  Boolean hasDocument
   *         ドキュメントに内容を持っているか
   * @param  Boolean replace
   *         履歴を置き換えるかどうか
   * @param  Boolean extract
   *         展開するか
   * @param  String refName
   *         対象のファイル
   * @param  Boolean useCache
   *         キャッシュを使用するか
   */
  loadMHTFromWeb : function (window, uri, targetWindow,
                             hasDocument, replace, extract, refName,
                             useCache) {
    let referrer = null;
    if (hasDocument) {
      try {
        referrer
          = Services.io.newURI
          (targetWindow.document.location.href, null, null);
      }
      catch (e) {
      }
    }
    
    UnMHTNetFileLoader.loadData
    (uri, referrer, useCache,
     function (status, data) {
      if (status == UnMHTNetFileLoader.STATUS_NG) {
        if (hasDocument) {
          UnMHTFileLoader.loadMHTFromDocument
          (window, uri, targetWindow,
           true, replace, extract, refName);
        }
        else {
          UnMHTFileLoader.loadErrorPage (window, targetWindow, uri.spec);
        }
      }
      else {
        UnMHTFileLoader.loadMHTFromData
        (window, uri, targetWindow,
         hasDocument, replace, extract, refName,
         data);
      }
    });
  },
  
  /**
   * フォーカスのあるウィンドウから mht ファイルを開く
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  Boolean extract
   *         展開するか
   */
  loadMHTFromFocusedWindow : function (window, extract) {
    let targetWindow = window.document.commandDispatcher.focusedWindow;
    if (targetWindow == window) {
      let tabbrowser = UnMHTWindowUtil.getBrowser (this._window);
      if (!tabbrowser) {
        return;
      }
      targetWindow = tabbrowser.contentWindow;
    }
    
    let uri
    = Services.io.newURI (targetWindow.document.location.href, null, null);
    
    if (UnMHTScheme.isUnMHTURI (uri.spec)) {
      uri.spec = UnMHTScheme.deUnMHTURI (uri.spec);
    }
        
    this.loadMHT
    (window, uri, targetWindow,
     true,    /* 開いているドキュメントを使用する */
     false,   /* 履歴はそのまま */
     extract, /* 展開は引数次第 */
     "",      /* 開始ページを開く */
     true);   /* キャッシュを使用する */
  },
  
  /**
   * エラーメッセージを表示する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMWindow targetWindow
   *         対象のウィンドウ
   * @param  String location
   *         ファイルの URI
   */
  loadErrorPage : function (window, targetWindow, location) {
    let nav
    = targetWindow.QueryInterface (Ci.nsIInterfaceRequestor)
    .getInterface (Ci.nsIWebNavigation);
    
    let flag = Ci.nsIWebNavigation.LOAD_FLAGS_REPLACE_HISTORY;
    nav.loadURI (UnMHTScheme.enErrorURI (location), flag, null, null, null);
  },
  
  /* ==== private ==== */
  
  _handled : {},      /* Object  DLF 等によってハンドル済みの mht ファイル
                       *   <String URI, Boolean> */
  
  _openMultiple : "", /* String  複数のタブを含む MHT を開く時
                       *   tabs: 全てのページをタブとして開く
                       *   index: インデックスページを表示する
                       *   first: 最初のタブを開く */
  
  /**
   * 初期化処理
   */
  init : function () {
    let f;
    
    f = function () {
      this._openMultiple = UnMHTPref.getPref ("open.multiple.action");
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "open.multiple.action", f, this);
    f ();
  },
  
  /**
   * 終了処理
   */
  term : function () {
    UnMHTPref.removeAllListeners (this);
  },
  
  /**
   * インデックスページをタブに置き換える
   * 
   * @param  nsIDOMWindow window
   *         対象のウィンドウ
   * @param  nsIDOMXULElement tabbrowser
   *         対象のタブブラウザ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   */
  _replaceTabs : function (window, tabbrowser, targetDocument) {
    let nodes = targetDocument.getElementsByTagName ("a");
    
    let tabsCount = 0;
    Array.forEach
    ((nodes.length && this._openMultiple == "first") ? [nodes [0]] : nodes,
     function (node) {
      if ("className" in node
          && node.className == "unmht-tabs") {
        tabbrowser.addTab (node.href);
        tabsCount ++;
      }
    });
    
    if (tabsCount == 0) {
      /* X-UnMHT-Tabs: true を持っているがタブが無い場合
       * 5.7.7 以前のバグで生成されたファイルの可能性がある */
      return;
    }
    
    let tab = UnMHTWindowUtil.getTabForDocument (window, targetDocument);
    if (tab) {
      tabbrowser.removeTab (tab);
    }
  },
  
  /**
   * リロードをフックする
   * 
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMWindow targetWindow
   *         対象のウィンドウ
   */
  _hookReload : function (window, targetWindow) {
    let nav = null;
    
    try {
      nav
        = targetWindow.QueryInterface (Ci.nsIInterfaceRequestor)
        .getInterface (Ci.nsIWebNavigation);
    }
    catch (e) {
      return;
    }
    
    let listener = {
      QueryInterface : XPCOMUtils.generateQI ([Ci.nsISupportsWeakReference,
                                               Ci.nsISHistoryListener]),
      
      OnHistoryReload : function (reloadURI, reloadFlags) {
        if (UnMHTScheme.isUnMHTURI (reloadURI.spec)) {
          let param = UnMHTScheme.getUnMHTURIParam (reloadURI.spec);
          let uri = Services.io.newURI (param.original, null, null);
      
          UnMHTFileLoader.loadMHT
          (window, uri, nav,
           false,         /* ドキュメント無し */
           true,          /* 履歴は置き換える */
           true,          /* 展開する */
           param.refName, /* 現在開いているファイルを開く */
           false);        /* キャッシュは使用しない */
          return false;
        }
        
        return true;
      },
  
      OnHistoryGoBack : function () { return true; },
      OnHistoryGoForward : function () { return true; },
      OnHistoryGotoIndex : function () { return true; },
      OnHistoryNewEntry : function () { return true; },
      OnHistoryPurge : function () { return true; }
    };
    
    try {
      nav.sessionHistory.addSHistoryListener (listener);
    }
    catch (e) {
      return;
    }
    
    targetWindow.addEventListener
    ("unload",
     function () {
      try {
        nav.sessionHistory.removeSHistoryListener (listener);
      }
      catch (e) {
      }
    }, true);
  }
};
UnMHTFileLoader.init ();
UnMHTShutdown.addListener (function (data, reason) {
    UnMHTFileLoader.term ();
  });
