/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/DownloadLastDir.jsm");
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");

let EXPORTED_SYMBOLS = ["UnMHTFilePicker"];

/**
 * ファイル選択ダイアログ
 */
let UnMHTFilePicker = {
  /**
   * ファイル選択ダイアログを開く
   *
   * @param  nsIDOMWindow window
   *         親ウィンドウ
   * @param  String title
   *         ダイアログのタイトル
   * @param  Number mode
   *         nsIFilePicker のモード
   * @param  Boolean useLastDir
   *         最後に使用したディレクトリを使用/保存する
   * @param  nsIFile displayDirectory
   *         useLastDir が false の際に表示するディレクトリ
   *         ファイルを指している場合は親ディレクトリになる
   * @param  Function modFunc
   *         ダイアログに変更を加える関数
   *           function (filePicker) {}
   *         不要ならば null
   * @param  String uri
   *         保存対象の URI
   *         サイトごとに記憶する際に使用する
   *         不要ならば null
   * @return nsIFile/Array
   *         選択されたファイル
   *         mode が Ci.nsIFilePicker.modeOpenMultiple の場合 [nsIFile, ...]
   *         中断した場合は null
   *         それ以外の場合は nsIFile
   */
  open : function (window, title, mode, useLastDir, displayDirectory,
                   modFunc, uri) {
    let filePicker
    = Cc ["@mozilla.org/filepicker;1"].createInstance (Ci.nsIFilePicker);
    
    if (uri) {
      try {
        uri = Services.io.newURI (uri, null, null);
      }
      catch (e) {
        uri = null;
      }
    }
    
    filePicker.init (window, title, mode);
    
    let dir = null;
    if (useLastDir) {
      if (mode == Ci.nsIFilePicker.modeSave) {
        dir = this._getLastSaveDir (uri);
      }
      else {
        dir = this._getLastOpenDir ();
      }
    }
    else if (displayDirectory) {
      let dir = displayDirectory.clone ();
      if (dir && dir.exists ()) {
        if (!file.isDirectory ()) {
          dir = dir.parent;
        }
      }
      else {
        dir = null;
      }
    }
    if (dir) {
      filePicker.displayDirectory = dir;
    }
    
    if (modFunc) {
      modFunc (filePicker);
    }
    
    let ret = filePicker.show ();
    if (!(ret == Ci.nsIFilePicker.returnOK
          || ret == Ci.nsIFilePicker.returnReplace)) {
      return null;
    }
    
    switch (mode) {
      case Ci.nsIFilePicker.modeOpenMultiple: {
        let files
        = [file for (file in XPCOMUtils.IterSimpleEnumerator
                       (filePicker.files, Ci.nsIFile))];
        if (files.length) {
          this._setLastOpenDir (files [0].parent);
        }
        
        return files;
      }
      case Ci.nsIFilePicker.modeGetFolder: {
        this._setLastOpenDir (filePicker.file);
        return filePicker.file;
      }
      case Ci.nsIFilePicker.modeSave: {
        this._setLastSaveDir (uri, filePicker.file.parent);
        return filePicker.file;
      }
      default: {
        this._setLastOpenDir (filePicker.file.parent);
        return filePicker.file;
      }
    }
  },
  
  /* ==== private ==== */
  
  /**
   * 開くのに最後に使用したディレクトリを取得する
   *
   * @return nsIFile
   *         最後に使用したディレクトリ
   *         無い場合は null
   */
  _getLastOpenDir : function () {
    if (UnMHTPref.getPref ("use_global_lastdir")) {
      let name = "browser.open.lastDir";
      if (UnMHTPref.prefBranch.getPrefType (name) != 0) {
        return UnMHTPref.prefBranch.getComplexValue (name, Ci.nsILocalFile);
      }
    }
    else {
      let path = UnMHTPref.getPref ("open.lastdir");
      if (path) {
        return UnMHTFileUtil.initFile (path);
      }
    }
    
    return null;
  },
  
  /**
   * 開くのに最後に使用したディレクトリを設定する
   *
   * @param  nsIFile dir
   *         最後に使用したディレクトリ
   */
  _setLastOpenDir : function (dir) {
    if (UnMHTPref.getPref ("use_global_lastdir")) {
      UnMHTPref.prefBranch.setComplexValue
      ("browser.open.lastDir", Ci.nsILocalFile, dir);
    }
    else {
      UnMHTPref.setPref ("open.lastdir", dir.path);
    }
  },
  
  /**
   * 保存するのに最後に使用したディレクトリを取得する
   *
   * @param  String uri
   *         保存対象の URI
   *         不要ならば null
   * @return nsIFile
   *         最後に使用したディレクトリ
   *         無い場合は null
   */
  _getLastSaveDir : function (uri) {
    if (UnMHTPref.getPref ("use_global_lastdir")) {
      return gDownloadLastDir.getFile (uri);
    }
    else {
      let path = UnMHTPref.getPref ("save.lastdir");
      if (path) {
        return UnMHTFileUtil.initFile (path);
      }
    }
    
    return null;
  },
  
  /**
   * 保存するのに最後に使用したディレクトリを設定する
   *
   * @param  String uri
   *         保存対象の URI
   *         不要ならば null
   * @param  nsIFile dir
   *         最後に使用したディレクトリ
   */
  _setLastSaveDir : function (uri, dir) {
    if (UnMHTPref.getPref ("use_global_lastdir")) {
      gDownloadLastDir.setFile (uri, dir);
    }
    else {
      UnMHTPref.setPref ("save.lastdir", dir.path);
    }
  }
};
