/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTAppInfo.jsm");
XPCOMUtils.importRelative (this, "UnMHTDOMUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTFilePicker.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTMail.jsm");
XPCOMUtils.importRelative (this, "UnMHTMIMEUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTNetFileLoader.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTSave.jsm");
XPCOMUtils.importRelative (this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative (this, "UnMHTStatusbarProgress.jsm");
XPCOMUtils.importRelative (this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative (this, "UnMHTUnicodeConverter.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTFileSaver"];

/**
 * MHT で保存のラッパー
 * UI からの要求に応じてデータを取得して MHT での保存を開始する
 */
let UnMHTFileSaver = {
  /**
   * 現在のドキュメントを MHT で保存
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  save : function (window, callback) {
    this.saveDocument
    (window, UnMHTWindowUtil.getCurrentDocument (window), callback);
  },
  
  /**
   * 対象のドキュメントを MHT で保存
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  Function callback
   *         保存終了時のコールバック
   */
  saveDocument : function (window, targetDocument, callback) {
    this.saveDocumentWithSelection (window, targetDocument, null, callback);
  },
  
  /**
   * 現在のドキュメントの選択範囲を保存
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  saveSelection : function (window, callback) {
    this.saveSelectionInDocument
    (window, UnMHTWindowUtil.getCurrentDocument (window), callback);
  },
  
  /**
   * 対象のドキュメントの現在の選択範囲を MHT で保存
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  Function callback
   *         保存終了時のコールバック
   */
  saveSelectionInDocument : function (window, targetDocument, callback) {
    let targetSelection = targetDocument.defaultView.getSelection ();
    if (targetSelection.rangeCount == 0) {
      window.alert (UnMHTStringBundle.get ("no_selection"));
      if (callback) {
        callback (true);
      }
      return;
    }
    
    this.saveDocumentWithSelection
    (window, targetSelection.getRangeAt (0).startContainer.ownerDocument,
     targetSelection, callback);
  },
  
  /**
   * 対象のドキュメントの対象の選択範囲を MHT で保存
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  nsISelection targetSelection
   *         対象の選択範囲
   * @param  Function callback
   *         保存終了時のコールバック
   */
  saveDocumentWithSelection : function (window, targetDocument, targetSelection,
                                        callback) {
    this._enterSaveProcess
    (window, targetDocument, targetSelection,
     null, false,
     callback);
  },
  
  /**
   * リンク先を MHT で保存
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  saveLink : function (window, callback) {
    let href = UnMHTDOMUtil.getLinkURL (window);
    if (!href) {
      window.alert (UnMHTStringBundle.get ("not_link"));
      if (callback) {
        callback (true);
      }
      return;
    }
    
    let tabbrowser = UnMHTWindowUtil.getBrowser (window);
    if (!tabbrowser) {
      /* ==== FIXME: 警告を表示するべき? ==== */
      if (callback) {
        callback (true);
      }
      return;
    }
    
    let iframe = window.document.createElement ("iframe");
    let wrappedCallback = function (isStopped) {
      iframe.parentNode.removeChild (iframe);
      if (callback) {
        callback (isStopped);
      }
    };
    
    iframe.hidden = true;
    iframe.setAttribute ("src", href);
    iframe.setAttribute ("type", "content");
    iframe.addEventListener
    ("load",
     function () {
      this.saveDocument (window, iframe.contentDocument, wrappedCallback);
    }.bind (this), true);

    tabbrowser.parentNode.insertBefore (iframe, tabbrowser);
  },
  
  /**
   * 複数のタブを MHT で保存
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   */
  savemulti : function (window) {
    let pageURL = "chrome://unmht/content/savemulti.xul";
    let features = "chrome,titlebar,centerscreen,resizable";
    let arg = {
      targetWindow : window
    };
    window.openDialog (pageURL, "", features, arg);
  },
  
  /**
   * QuickSave
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  quicksave : function (window, callback) {
    let targetDocument = UnMHTWindowUtil.getCurrentDocument (window);
    
    if (!UnMHTSave.isSavable (window, targetDocument)) {
      window.alert (UnMHTStringBundle.get ("not_savable"));
      if (callback) {
        callback (true);
      }
      return;
    }
    
    let baseDir = null;
    let localFile = {value : null};
    if (UnMHTPref.getPref ("save.local.same_dir")
        && UnMHTSave.isLocalFile (targetDocument.location.href, localFile)) {
      baseDir = localFile.value.parent;
    }
    if (!baseDir) {
      let baseDirPath = UnMHTPref.getPref ("save.quick.dir");
      baseDir = baseDirPath
      ? UnMHTFileUtil.initFile (baseDirPath)
      : UnMHTFileUtil.getDesktopDirectory ().clone ();
    }
    if (!baseDir) {
      window.alert (UnMHTStringBundle.get ("illegal_directory"));
      if (callback) {
        callback (true);
      }
      return;
    }
    if (!baseDir.exists ()) {
      try {
        UnMHTFileUtil.createDirectory (baseDir);
      }
      catch (e) {
        window.alert (UnMHTStringBundle.get ("illegal_directory"));
        if (callback) {
          callback (true);
        }
        return;
      }
    }
    
    let date = new Date ();
    let filenameBase = UnMHTSave.getFilename (targetDocument, null, null, date);
    let filename = filenameBase + ".mht";
    
    let targetFile = baseDir.clone ();
    targetFile.append (filename);
    if (targetFile.exists ()) {
      let dupAction = UnMHTPref.getPref ("save.quick.dup.action");
      
      let result;
      if (dupAction == "ask") {
        result
          = UnMHTSave.showOverwriteConfirm
          (window, targetFile, filenameBase, false);
        
        if (result.result == "abort" || result.result == "no") {
          if (callback) {
            callback (true);
          }
          return;
        }
        
        if (result.result == "rename") {
          targetFile = baseDir.clone ();
          targetFile.append (result.filename);
        }
      }
      else {
        result = {
          result: dupAction,
          applyToAll: false,
          filename : ""
        };
        if (dupAction == "rename") {
          result.filename = UnMHTSave.getUniqueFilename (baseDir, filenameBase);
        }
      }
      
      if (result.result == "rename") {
        targetFile = baseDir.clone ();
        targetFile.append (result.filename);
      }
    }
    
    this._startSave
    (window, targetDocument,
     (UnMHTPref.getPref ("save.quick.use_selection")
      && UnMHTDOMUtil.isContentSelected (window, false))
     ? targetDocument.defaultView.getSelection (): null,
     null,
     UnMHTSave.stringToSaveState (UnMHTPref.getPref ("save.quick.state")), null,
     targetFile, callback);
  },
  
  /**
   * 現在のドキュメントを MHT としてメールで送信
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  sendmail : function (window, callback) {
    this.sendmailDocument
    (window, UnMHTWindowUtil.getCurrentDocument (window), callback);
  },

  /**
   * 対象のドキュメントを MHT としてメールで送信
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  Function callback
   *         保存終了時のコールバック
   */
  sendmailDocument : function (window, targetDocument, callback) {
    this._enterSaveProcess
    (window, targetDocument, null,
     null, true,
     callback);
  },
  
  /**
   * 対象のドキュメントを対象のファイルに保存する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  Number saveState
   *         保存状態
   * @param  Boolean useAddURIs
   *         追加 URI を使用するか
   * @param  nsIFile targetFile
   *         対象のファイル
   * @param  Function callback
   *         終了時コールバック
   */
  saveDocumentToFile : function (window, targetDocument,
                                 saveState, useAddURIs,
                                 targetFile, callback) {
    if (useAddURIs) {
      let saveParam = {
        saveState: saveState
      };
      this._showAdditionalDialog
      (window, targetDocument, null,
       null, saveParam,
       targetFile, callback);
      return;
    }
    
    this._startSave
    (window, targetDocument, null,
     null,
     saveState, null,
     targetFile, callback);
  },
  
  /**
   * ScrapBook を MHT で保存する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Function callback
   *         終了時コールバック
   */
  saveScrapBook : function (window, scrapBookData, callback) {
    this._enterSaveProcess
    (window, null, null,
     scrapBookData, false,
     callback);
  },
  
  /* ==== private ==== */
  
  /**
   * メール用のファイルとコールバックを作成する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  String targetFilename
   *         対象のファイル名
   * @param  Function callback
   *         保存終了時のコールバック
   * @return Array
   *         [nsIFile 対象のファイル, Function コールバック]
   */
  _generateMailData : function (window, targetDocument,
                                targetFilename, callback) {
    let targetFile = UnMHTFileUtil.getTempDirectory ().clone ();
    /* ==== FIXME: サブディレクトリを作った方が安全 ==== */
    targetFile.append (targetFilename);
    
    let title = targetDocument.title;
    
    let wrappedCallback = function (isStopped) {
      if (isStopped) {
        if (callback) {
          callback (isStopped);
        }
        return;
      }
      
      if (!UnMHTMail.send (window, targetFile, title)) {
        window.alert (UnMHTStringBundle.get ("sendmail_fail") + "\n"
                      + UnMHTStringBundle.get ("sendmail_fail2"));
      }
      
      if (callback) {
        callback (isStopped);
      }
    }.bind (this);
    
    return [targetFile, wrappedCallback];
  },
  
  /**
   * 保存プロセスを開始する
   *
   * 必要であればオプションダイアログを表示してから保存する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  nsISelection targetSelection
   *         対象の選択範囲
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Boolean isMail
   *         メールで送信するかどうか
   * @param  Function callback
   *         保存終了時のコールバック
   */
  _enterSaveProcess : function (window, targetDocument, targetSelection,
                                scrapBookData, isMail,
                                callback) {
    if (UnMHTPref.getPref ("save.filename.custom.mail_date")
        && window
        && UnMHTWindowUtil.isMailWindow (window)
        && "gFolderDisplay" in window
        && "selectedMessageUris" in window.gFolderDisplay
        && window.gFolderDisplay.selectedMessageUris.length == 1) {
      let mailSession
      = Cc ["@mozilla.org/messenger/services/session;1"]
      .getService (Ci.nsIMsgMailSession);
          
      let messageURL
      = mailSession.ConvertMsgURIToMsgURL
      (window.gFolderDisplay.selectedMessageUris [0], window.msgWindow);

      let uri;
      try {
        uri = Services.io.newURI (messageURL, null, null);
      }
      catch (e) {
      }
      
      if (uri) {
        UnMHTNetFileLoader.loadData
        (uri, null, true, function (status, data) {
          let date = new Date ();
          if (status != UnMHTNetFileLoader.STATUS_NG) {
            let mailPart
              = UnMHTMIMEUtil.decodeMessage
              (data.replace (/\r\n|\r|\n/g, "\r\n"), true);
            if (mailPart
                && "date" in mailPart.fields) {
              let tmpDate = new Date (mailPart.fields ["date"]);
              if (!isNaN (date.getFullYear ())) {
                date = tmpDate;
              }
            }
          }
          
          this._enterSaveProcessWithDate
          (window, targetDocument, targetSelection,
           scrapBookData, isMail,
           callback, date);
        }.bind (this));
        return;
      }
    }
    
    this._enterSaveProcessWithDate
    (window, targetDocument, targetSelection,
     scrapBookData, isMail,
     callback, new Date ());
  },
  
  /**
   * 保存プロセスを開始する
   *
   * 必要であればオプションダイアログを表示してから保存する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  nsISelection targetSelection
   *         対象の選択範囲
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Boolean isMail
   *         メールで送信するかどうか
   * @param  Function callback
   *         保存終了時のコールバック
   * @param  Date date
   *         日時
   */
  _enterSaveProcessWithDate : function (window, targetDocument, targetSelection,
                                scrapBookData, isMail,
                                callback, date) {
    let dialogCallback;
    if (isMail) {
      dialogCallback = function (saveParam) {
        let [targetFile, wrappedCallback]
        = this._generateMailData
        (window, targetDocument,
         saveParam.targetFilename,
         callback);
          
        this._startSave
        (window, targetDocument, targetSelection,
         scrapBookData,
         saveParam.saveState, saveParam.addURIs,
         targetFile, wrappedCallback);
      }.bind (this);
    }
    else {
      dialogCallback = function (saveParam) {
        this._showFilePicker
        (window, targetDocument, targetSelection,
         scrapBookData, saveParam,
         callback);
      }.bind (this)
    }
    
    let targetFilename
    = UnMHTSave.getFilename (targetDocument, scrapBookData, null, date)
    + ".mht";
    
    if (UnMHTPref.getPref ("save.show_option_dialog")) {
      this._showOptionDialog (window, targetFilename, targetSelection,
                              scrapBookData, isMail,
                              dialogCallback);
    }
    else {
      let saveParam = {
        targetFilename : targetFilename
      }
      if (isMail) {
        saveParam.saveState
        = UnMHTSave.stringToSaveState (UnMHTPref.getPref ("save.mail.state"));
        saveParam.addURIs =  null;
      }
      dialogCallback (saveParam);
    }
  },
  
  /**
   * オプションダイアログを表示してから保存する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  String targetFilename
   *         対象のファイル名
   * @param  nsISelection targetSelection
   *         対象の選択範囲
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Boolean isMail
   *         メールで送信するかどうか
   * @param  Function dialogCallback
   *         ダイアログのコールバック
   */
  _showOptionDialog : function (window, targetFilename, targetSelection,
                                scrapBookData, isMail,
                                dialogCallback) {
    let saveStatePref = isMail ? "save.mail.state" : "save.single.state";
    let saveState
    = UnMHTSave.stringToSaveState (UnMHTPref.getPref (saveStatePref));
    let saveStateMask = UnMHTSave.SAVESTATE_MASK_ALL;
    
    if (scrapBookData) {
      saveStateMask = UnMHTSave.SAVESTATE_MASK_ORIGINAL;
      if (!(saveState & saveStateMask)) {
        saveState = UnMHTSave.SAVESTATE_ORIGINAL;
      }
    }
    else if (targetSelection) {
      saveStateMask = UnMHTSave.SAVESTATE_MASK_CURRENT;
      if (!(saveState & saveStateMask)) {
        saveState = UnMHTSave.SAVESTATE_CURRENT;
      }
    }
    
    let pageURL = "chrome://unmht/content/saveoptions.xul";
    let features = "chrome,titlebar,centerscreen,resizable";
    let arg = {
      callback : function (saveParam) {
        saveParam.showAdditionalDialog = false; 
        dialogCallback (saveParam);
      },
      filename : targetFilename,
      showFilename : isMail,
      saveState : saveState,
      saveStateMask : saveStateMask,
      saveStatePref : saveStatePref
    };
    window.openDialog (pageURL, "", features, arg);
  },
  
  /**
   * ファイル選択ダイアログを表示してから保存する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  nsISelection targetSelection
   *         対象の選択範囲
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Object saveParam
   *         追加パラメータ
   * @param  Function callback
   *         保存終了時のコールバック
   */
  _showFilePicker : function (window, targetDocument, targetSelection,
                              scrapBookData, saveParam,
                              callback) {
    let showAdditionalDialog = false;
    
    let currentFilePicker = null;
    
    let types = [];
    
    let title = UnMHTStringBundle.get ("save_title");
    let mode = Ci.nsIFilePicker.modeSave;
    let modFunc = function (filePicker) {
      currentFilePicker = filePicker;
      
      let localFile = {value : null};
      if (UnMHTPref.getPref ("save.local.same_dir")
          && UnMHTSave.isLocalFile (targetDocument.location.href, localFile)) {
        filePicker.displayDirectory = localFile.value.parent;
      }
      
      if ("saveState" in saveParam
          && "addURIs" in saveParam
          && saveParam.addURIs) {
        types
        = [[saveParam.saveState, saveParam.addURIs.length > 0]];
      }
      else {
        let saveStateMask = UnMHTSave.SAVESTATE_MASK_ALL;
        
        if (scrapBookData) {
          saveStateMask = UnMHTSave.SAVESTATE_MASK_ORIGINAL;
        }
        if (targetSelection) {
          saveStateMask = UnMHTSave.SAVESTATE_MASK_CURRENT;
        }
        
        let stateList = UnMHTSave.getSaveStateList (saveStateMask);
        
        types
        = stateList.map (function (state) [state, false])
        .concat (stateList.map (function (state) [state, true]));
      }
      
      let lastStateStr = UnMHTPref.getPref ("save.single.state");
      let lastAdd = UnMHTPref.getPref ("save.single.add");
      
      types.forEach (function ([saveState, useAdd], i) {
          filePicker.appendFilter
            (UnMHTSave.makeFiletype (saveState, useAdd), "*.mht");
          
          if (UnMHTSave.saveStateToString (saveState) == lastStateStr
              && useAdd == lastAdd) {
            filePicker.filterIndex = i;
          }
        });
      
      filePicker.defaultString = saveParam.targetFilename;
    }.bind (this);
    
    let targetFile
    = UnMHTFilePicker.open
    (window, title, mode, true, null,
     modFunc, targetDocument ? targetDocument.location.href : null);
    if (!targetFile) {
      if (callback) {
        callback (false);
      }
      return;
    }
    
    if (!("saveState" in saveParam
          && "addURIs" in saveParam)) {
      let [saveState, useAdd] = types [currentFilePicker.filterIndex];
      UnMHTPref.setPref ("save.single.state", 
                         UnMHTSave.saveStateToString (saveState));
      UnMHTPref.setPref ("save.single.add", useAdd);
      
      saveParam.saveState = saveState;
      
      if (useAdd) {
        showAdditionalDialog = true;
      }
      else {
        saveParam.addURIs = null;
      }
    }
    
    targetFile.leafName = UnMHTFileUtil.validateFileName (targetFile.leafName);
    
    if (!targetFile.leafName.match (/\.mht(ml)?$/i)) {
      /* 拡張子 .mht が無いと UnMHT でハンドルできなくなるので追加する */
      
      let filenameBase = targetFile.leafName;
      
      targetFile = targetFile.parent.clone ();
      targetFile.append (filenameBase + ".mht");
      
      if (targetFile.exists ()) {
        /* .mht を追加した事でファイル名が被った */
        let result
          = UnMHTSave.showOverwriteConfirm
          (window, targetFile, filenameBase, false);
          
        if (result.result == "abort" || result.result == "no") {
          if (callback) {
            callback (false);
          }
          return;
        }
          
        if (result.result == "rename") {
          targetFile = targetFile.parent.clone ();
          targetFile.append (result.filename);
        }
      }
    }
    
    if (showAdditionalDialog) {
      this._showAdditionalDialog
      (window,
       targetDocument, targetSelection,
       scrapBookData, saveParam,
       targetFile, callback);
    }
    else {
      this._startSave
      (window, targetDocument, targetSelection,
       scrapBookData,
       saveParam.saveState, saveParam.addURIs,
       targetFile, callback);
    }
  },
  
  /**
   * 追加ファイルのダイアログを表示してから保存する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  nsISelection targetSelection
   *         対象の選択範囲
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Object saveParam
   *         追加パラメータ
   * @param  nsIFile targetFile
   *         対象のファイル
   * @param  Function callback
   *         終了時コールバック
   */
  _showAdditionalDialog : function (window, targetDocument, targetSelection,
                                    scrapBookData, saveParam,
                                    targetFile, callback) {
    let pageURL = "chrome://unmht/content/additional.xul";
    let features = "chrome,titlebar,centerscreen,resizable";
    let arg = {
      callback : function (addURIs) {
        this._startSave
        (window, targetDocument, targetSelection,
         scrapBookData,
         saveParam.saveState, addURIs,
         targetFile, callback);
      }.bind (this)
    };
    window.openDialog (pageURL, "", features, arg);
  },
  
  /**
   * ファイルを保存する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  nsISelection targetSelection
   *         対象の選択範囲
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Number saveState
   *         保存状態
   * @param  Array addURIs
   *         追加ファイル
   * @param  nsIFile targetFile
   *         対象のファイル
   * @param  Function callback
   *         終了時コールバック
   */
  _startSave : function (window, targetDocument, targetSelection,
                         scrapBookData,
                         saveState, addURIs,
                         targetFile, callback) {
    if (UnMHTPref.getPref ("save.progress.statusbar")) {
      let prog = UnMHTStatusbarProgressMap.get (window);
      let [location, subject]
      = UnMHTSave.getLocationAndTitle (targetDocument, scrapBookData);
      
      let saveContext
      = UnMHTSave.createContext (window, targetFile, location, subject, prog);
      
      let arg = {
        targetDocument : targetDocument,
        targetSelection : targetSelection,
        scrapBookData : scrapBookData,
        saveState : saveState,
        location : location,
        isHTML : true,
        isStartPart : false,
        window : window
      };
      if (!saveContext.addDocument (arg)) {
        saveContext.destruct ();
        
        if (callback) {
          callback (false);
        }
        return;
      }
      
      if (addURIs) {
        addURIs.forEach (function (file) {
            saveContext.addFile (null, file, false);
          });
      }
      
      prog.addContext (saveContext, callback);
      
      saveContext.start ();
    }
    else {
      let pageURL = "chrome://unmht/content/save.xul";
      let features = "chrome,titlebar,centerscreen,resizable";
      let arg = {
        targetFile : targetFile,
        targetDocument : targetDocument,
        saveState : saveState,
        scrapBookData : scrapBookData,
        targetSelection : targetSelection,
        addURIs : addURIs,
        callback : callback,
        window : window
      };
      window.openDialog (pageURL, "", features, arg);
    }
  }
};
