/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTShutdown.jsm");

let EXPORTED_SYMBOLS = ["UnMHTFileUtil"];

let PR_RDONLY = 0x01;
let PR_WRONLY = 0x02;
let PR_CREATE_FILE = 0x08;
let PR_TRUNCATE = 0x20;

/**
 * ファイル操作ユーティリティ
 */
let UnMHTFileUtil = {
  /**
   * ファイルを初期化する
   *
   * @param  String path
   *         ファイルのパス
   * @return nsIFile
   *         指定されたパスのファイル
   *         失敗した場合は null
   */
  initFile : function (path) {
    let file
    = Cc ["@mozilla.org/file/local;1"].createInstance (Ci.nsIFile);
    
    try {
      file.initWithPath (path);
    }
    catch (e) {
      return null;
    }
    
    return file;
  },
  
  /**
   * ファイルの入力ストリームを開く
   *
   * @param  nsIFile file
   *         対象のファイル
   * @return nsIFileInputStream
   *         ファイルの入力ストリーム
   *         失敗したら null
   */
  openInputStream : function (file) {
    let fstream
    = Cc ["@mozilla.org/network/file-input-stream;1"]
    .createInstance (Ci.nsIFileInputStream);
    
    try {
      fstream.init (file, PR_RDONLY, parseInt ("0444", 8), 0);
    }
    catch (e) {
      return null;
    }
    
    return fstream;
  },
  
  /**
   * ファイルの出力ストリームを開く
   *
   * @param  nsIFile file
   *         対象のファイル
   * @param  String permission
   *         パーミッション (オプショナル)
   *         undefined の場合は 0644
   * @return nsIFileOutputStream
   *         ファイルの出力ストリーム
   *         失敗したら null
   */
  openOutputStream : function (file, permission) {
    let fstream
    = Cc ["@mozilla.org/network/file-output-stream;1"]
    .createInstance (Ci.nsIFileOutputStream);
    
    try {
      fstream.init (file, PR_WRONLY | PR_CREATE_FILE | PR_TRUNCATE,
                    parseInt (permission || "0644", 8), 0);
    }
    catch (e) {
      return null;
    }
    
    return fstream;
  },
  
  /**
   * Zip 書き込みを作成する
   *
   * @param  nsIFile file
   *         対象のファイル
   * @return nsIZipWriter
   *         ZIP 書き込み
   *         失敗した場合は null
   */
  openZipWriter : function (file) {
    let zip = Cc ["@mozilla.org/zipwriter;1"].createInstance (Ci.nsIZipWriter);
    
    try {
      zip.open (file, PR_WRONLY | PR_CREATE_FILE | PR_TRUNCATE);
    }
    catch (e) {
      return null;
    }
    
    return zip;
  },
  
  /**
   * ファイルからデータを読み込む
   *
   * @param  nsIFile file
   *         対象のファイル
   * @return String
   *         ファイルのデータ
   */
  readData : function (file) {
    let fstream = this.openInputStream (file);
    if (!fstream) {
      return "";
    }
    
    let bstream
    = Cc ["@mozilla.org/binaryinputstream;1"]
    .createInstance (Ci.nsIBinaryInputStream);
    bstream.setInputStream (fstream);
    
    let data = bstream.readBytes (file.fileSize);
    
    bstream.close ();
    fstream.close ();
    
    return data;
  },
  
  /**
   * ディレクトリの内容を配列で返す
   *
   * @param  nsIFile dir
   *         対象のディレクトリ
   * @return Array
   *         ディレクトリの内容のファイル
   *         [nsIFile, ...]
   */
  getDirectoryEntries : function (dir) {
    let files = [];
    
    return [file for (file in XPCOMUtils.IterSimpleEnumerator
                        (dir.directoryEntries, Ci.nsIFile))];
  },
  
  /**
   * ファイル名に使えない文字を変換する
   *
   * @param  String filename
   *         元のファイル名
   * @return String
   *         変換したファイル名
   */
  validateFileName : function (filename) {
    let runtime
    = Cc ["@mozilla.org/xre/app-info;1"].getService (Ci.nsIXULRuntime);
    
    let sepRe = /\//g;
    
    if (runtime.OS == "WINNT") {
      sepRe = /[\\\/]/g;
      
      /* ファイル名に使えない文字
       *   |
       *   *
       *   :
       *   ?
       *   "
       *   <
       *   >
       * ファイル名に使用すると IE が MHT ファイルを開けなくなる文字
       *   #
       *   !
       */
      filename = filename.replace (/\"/g, "'")
        .replace (/[\*\:\?#\!\|]/g, "_")
        .replace (/</g, "(")
        .replace (/>/g, ")");
    }
    else if (runtime.OS == "Darwin") {
      sepRe = /[\:\/]/g;
    }
    
    return filename.replace (sepRe, "_");
  },
  
  /**
   * パスを展開、連結する
   *
   * @param  String baseDir
   *         親ディレクトリのパス
   *         末尾にセパレータが無くてもディレクトリ名として扱われる事に注意
   * @param  String sub
   *         相対パス、もしくは絶対パス
   * @return String
   *         連結したパス
   */
  jointPath : function (baseDir, sub) {
    if (!baseDir) {
      return sub;
    }
    
    /* セパレータを検出 */
    let sep;
    let pathRe = null;
    if (baseDir.indexOf ("\\") != -1) {
      /* Windows のパスの場合 */
      sep = "\\";
      pathRe = /[^\\]+/g;
    }
    else if (baseDir.indexOf ("/") != -1) {
      /* Mac OS X、Linux のパス、URI の場合 */
      sep = "/";
      pathRe = /[^\/]+/g;
    }
    else if (baseDir.indexOf (":") != -1) {
      /* Mac OS 9 以前のパスの場合 */
      sep = ":";
      pathRe = /[^:]+/g;
    }
    else {
      /* 区切りが見付からない場合連結できないので
       * サブのみ返す */
      return sub;
    }
    
    let scheme = "";
    let host = "";
    let re
    = baseDir.match (/^([A-Za-z0-9\-]+:\/\/\/?)([^\/]+)(.*)$/)
    || baseDir.match (/^([A-Za-z]+:)()(.*)$/)
    || baseDir.match (/^(\\)()(\\.*)$/);
    if (re) {
      [, scheme, host, baseDir] = re;
    }
    
    re = sub.match (/^\/\/\/?(.*)$/);
    if (re) {
      /* プロトコルのみ継承する */
      return scheme + re [1];
    }
    
    if (sub.match (/^([A-Za-z0-9\-]+:\/\/\/?)([^\/]+)(.*)$/)
        || sub.match (/^([A-Za-z]+:)(.*)$/)
        || sub.match (/^(\\)(\\.*)$/)) {
      /* sub が URI、 Windows のネイティブな絶対パス、
       * ネットワークドライブの場合 */
      return sub;
    }
        
    if (sub.substr (0, 1) == sep) {
      /* sub が絶対パスの場合 */
      baseDir = "";
    }
    
    if (sub.length > 5 && sub.substr (0, 5) == "data:") {
      /* sub が data スキームの場合 */
      return sub;
    }
    
    let path = baseDir + sep + sub;
    
    let isDir = false;
    if (path.substr (path.length - 1, 1) == sep) {
      isDir = true;
    }
    
    let newNames = [];
    path.replace
    (pathRe,
     function (matched) {
      if (matched == ".") {
      }
      else if (matched == "..") {
        newNames.pop ();
      }
      else if (matched == "") {
      }
      else {
        newNames.push (matched);
      }
      return "";
    });
    
    path = sep + newNames.join (sep);
    if (isDir) {
      path += sep;
    }
    
    return scheme + host + path;
  },
  
  /**
   * ディレクトリを作成する
   *
   * @param  nsIFile dir
   *         対象のディレクトリ
   */
  createDirectory : function (dir) {
    if (!dir.exists ()) {
      dir.create (Ci.nsIFile.DIRECTORY_TYPE, parseInt ("0755", 8));
    }
  },
  
  /**
   * file プロトコルをファイルに変換する
   *
   * @param  String uri
   *         file プロトコル
   * @return nsIFile
   *         ファイル、もしくは
   */
  getFileFromURLSpec : function (uri) {
    return this._fileProtocolHandler.getFileFromURLSpec (uri);
  },
  
  /**
   * ファイルを file プロトコルに変換する
   *
   * @param  nsIFile file
   *         ファイル
   * @return String
   *         file プロトコル
   */
  getURLSpecFromFile : function (file) {
    return this._fileProtocolHandler.getURLSpecFromFile (file);
  },
  
  /**
   * Applications ディレクトリを取得する
   * Darwin 専用
   *
   * @return  nsIFile
   *          Applications ディレクトリ
   */
  getApplicationsDirectory : function () {
    try {
      return Cc ["@mozilla.org/file/directory_service;1"]
      .getService (Ci.nsIProperties).get ("LocApp", Ci.nsIFile);
    }
    catch (e) {
    }
        
    return null;
  },
  
  /**
   * ユーザの Applications ディレクトリを取得する
   * Darwin 専用
   *
   * @return  nsIFile
   *          ユーザの Applications ディレクトリ
   */
  getUserApplicationsDirectory : function () {
    try {
      return Cc ["@mozilla.org/file/directory_service;1"]
      .getService (Ci.nsIProperties).get ("UsrApp", Ci.nsIFile);
    }
    catch (e) {
    }
        
    return null;
  },
  
  /**
   * Program Files ディレクトリを取得する
   * WINNT 専用
   *
   * @return  nsIFile
   *          Program Files ディレクトリ
   */
  getProgramFilesDirectory : function () {
    try {
      return Cc ["@mozilla.org/file/directory_service;1"]
      .getService (Ci.nsIProperties).get ("ProgF", Ci.nsIFile);
    }
    catch (e) {
    }
        
    return null;
  },

  /**
   * 現在のユーザの Profile ディレクトリを取得する
   *
   * @return  nsIFile
   *          現在のユーザの Profile ディレクトリ
   */
  getProfileDirectory : function () {
    try {
      return Cc ["@mozilla.org/file/directory_service;1"]
      .getService (Ci.nsIProperties).get ("ProfD", Ci.nsIFile);
    }
    catch (e) {
    }
        
    return null;
  },
  
  /**
   * 現在のユーザの Desktop ディレクトリを取得する
   *
   * @return  nsIFile
   *          現在のユーザの Desktop ディレクトリ
   */
  getDesktopDirectory : function () {
    try {
      return Cc ["@mozilla.org/file/directory_service;1"]
      .getService (Ci.nsIProperties).get ("Desk", Ci.nsIFile);
    }
    catch (e) {
    }
        
    return null;
  },
  
  /**
   * 現在のユーザの Temp ディレクトリを取得する
   *
   * @return  nsIFile
   *          現在のユーザの Temp ディレクトリ
   */
  getTempDirectory : function () {
    try {
      return Cc ["@mozilla.org/file/directory_service;1"]
      .getService (Ci.nsIProperties).get ("TmpD", Ci.nsIFile);
    }
    catch (e) {
    }
        
    return null;
  },

  /**
   * システムディレクトリを取得する
   *
   * @return  nsIFile
   *          システムディレクトリ
   */
  getSystemDirectory : function () {
    return this._systemDirectory;
  },
  
  /**
   * キャッシュディレクトリを取得する
   *
   * @return  nsIFile
   *          キャッシュディレクトリ
   */
  getCacheDirectory : function () {
    return this._cacheDirectory;
  },
  
  /**
   * キャッシュを削除する
   */
  clearCacheDirectory : function () {
    try {
      let dir = this.getCacheDirectory ().clone ();
      if (!dir.exists ()) {
        return;
      }
      
      this.getDirectoryEntries (dir).forEach (function (file) {
          file.remove (true);
        });
    }
    catch (e) {
    }
  },
  
  /* ==== private ==== */
  
  _fileProtocolHandler : null, /* nsIFileProtocolHandler  プロトコルの変換器 */
  
  _systemDirectory : null,     /* nsIFile  システムディレクトリ
                                *          Profile ディレクトリ/UnMHT */
  _cacheDirectory : null,      /* nsIFile  キャッシュディレクトリ
                                *          Profile ディレクトリ/UnMHT/Cache */
  
  /**
   * 初期化処理
   */
  init : function () {
    this._fileProtocolHandler
    = Services.io.getProtocolHandler ("file")
    .QueryInterface (Ci.nsIFileProtocolHandler);
    this._makeSystemDirectory ();
    this._makeCacheDirectory ();
  },
  
  /**
   * 終了処理
   */
  term : function () {
    this.clearCacheDirectory ();
  },
  
  /**
   * システムディレクトリを作成する
   */
  _makeSystemDirectory : function () {
    this._systemDirectory = this.getProfileDirectory ();
    this._systemDirectory.append ("UnMHT");
    this.createDirectory (this._systemDirectory);
  },
  
  /**
   * キャッシュディレクトリを作成する
   */
  _makeCacheDirectory : function () {
    this._cacheDirectory = this._systemDirectory.clone ();
    this._cacheDirectory.append ("Cache");
    this.createDirectory (this._cacheDirectory);
  }
};
UnMHTFileUtil.init ();
UnMHTShutdown.addListener (function (data, reason) {
    UnMHTFileUtil.term ();
  });
