/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileSaver.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTPrototype.jsm");
XPCOMUtils.importRelative (this, "UnMHTSave.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowHook.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTKeyboardShortcut"];

/**
 * キーボードショートカット
 *
 * @param  String pref
 *         設定名
 * @param  Function func
 *         実行する関数
 */
function UnMHTKeyboardShortcutItem (pref, func) {
  this._pref = pref;
  this._func = func;
}
UnMHTKeyboardShortcutItem.prototype = {
  /**
   * 破棄
   */
  destruct : function () {
    this._func = null;
  },
  
  /**
   * 設定を読み込む
   */
  loadPref : function () {
    this._enabled = UnMHTPref.getPref (this._pref);
    this._keycode
    = Ci.nsIDOMKeyEvent ["DOM_" + UnMHTPref.getPref (this._pref + ".keycode")];
    this._modsAlt = UnMHTPref.getPref (this._pref + ".mods.alt");
    this._modsCtrl = UnMHTPref.getPref (this._pref + ".mods.ctrl");
    this._modsMeta = UnMHTPref.getPref (this._pref + ".mods.meta");
    this._modsShift = UnMHTPref.getPref (this._pref + ".mods.shift");
  },
  
  /**
   * 有効かどうかを返す
   *
   * @return Boolean
   *         有効かどうか
   */
  isEnabled : function () {
    return this._enabled;
  },
  
  /**
   * ショートカットキーを調べる
   *
   * @param  Event event
   *         対象のイベント
   */
  check : function (event) {
    return (this._keycode == event.keyCode
            && this._modsAlt == event.altKey
            && this._modsCtrl == event.ctrlKey
            && this._modsMeta == event.metaKey
            && this._modsShift == event.shiftKey);
  },
  
  /**
   * 関数を実行する
   */
  fire : function () {
    this._func ();
  },
  
  /* ==== private ==== */
  
  _pref : "",         /* String  設定名 */
  _func : "",         /* Function  実行する関数 */
  
  _enabled : false,   /* Boolean  有効 */
  _keycode : 0,       /* Number  キーコード */
  _modsAlt : false,   /* Boolean  Alt */
  _modsCtrl : false,  /* Boolean  Ctrl */
  _modsMeta : false,  /* Boolean  Meta */
  _modsShift : false  /* Boolean  Shift */
};

/**
 * キーボードショートカットのフック
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 * @inherits   UnMHTWindowHook
 * @implements IUnMHTWindowHook
 */
function UnMHTKeyboardShortcut (window) {
  UnMHTWindowHook.call (this, window);
  
  this._items = [
    new UnMHTKeyboardShortcutItem ("shortcut.save.single", function () {
        let targetDocument = UnMHTWindowUtil.getCurrentDocument (window);
        if (!UnMHTSave.isSavable (window, targetDocument)) {
          return false;
        }
        
        UnMHTFileSaver.save (window);
        return true;
      }),
    new UnMHTKeyboardShortcutItem ("shortcut.save.multi", function () {
        UnMHTFileSaver.savemulti (window);
        return true;
      }),
    new UnMHTKeyboardShortcutItem ("shortcut.save.quick", function () {
        let targetDocument = UnMHTWindowUtil.getCurrentDocument (window);
        if (!UnMHTSave.isSavable (window, targetDocument)) {
          return false;
        }
        
        UnMHTFileSaver.quicksave (window);
        return true;
      })
    ];
  
  this.onKeyDown = this.onKeyDown.bind (this);
}
UnMHTKeyboardShortcut.prototype = {
  /* ---- UI イベント - キーボード ---- */
  
  /**
   * キーが押されたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onKeyDown : function (event) {
    this._enabledItems.some (function (item) {
        if (item.check (event)
            && item.fire ()) {
          event.preventDefault ();
          event.stopPropagation ();
          return true;
        }
        return false;
      });
  },
  
  /* ==== IUnMHTWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct : function () {
    this._enabledItems = null;
    this._items.forEach (function (item) item.destruct ());
    this._items = null;
    
    UnMHTPrototype.super (this, "destruct").call (this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param  Boolean isStartup
   *         スタートアップイベントか
   * @param  Number reason
   *         スタートアップの理由
   */
  load : function (isStartup, reason) {
    let f = function () {
      this._items.forEach (function (item) item.loadPref ());
      this._hookShortcut ();
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "shortcut.save", f, this);
    f ();
  },  
  
  /**
   * ウィンドウからアンロードする
   *
   * @param  Boolean isShutdown
   *         シャットダウンイベントか
   * @param  Number reason
   *         シャットダウンの理由
   */
  unload : function (isShutdown, reason) {
    if (this._isHooked) {
      this._window.removeEventListener ("keydown", this.onKeyDown, true);
      this._isHooked = false;
    }
    
    UnMHTPref.removeAllListeners (this);
  },
  
  /* ==== private ==== */

  _items : null,        /* Array  ショートカット
                         *   [UnMHTKeyboardShortcutItem, ...] */
  _enabledItems : null, /* Array  有効なショートカット
                         *   [UnMHTKeyboardShortcutItem, ...] */
  
  _isHooked : false,    /* Boolean  キーボードイベントをフックしたか */
  
  /**
   * 設定に合わせてキーイベントをフックする
   */
  _hookShortcut : function () {
    this._enabledItems = this._items.filter (function (item) item.isEnabled ());
    
    if (this._enabledItems.length && !this._isHooked) {
      this._window.addEventListener ("keydown", this.onKeyDown, true);
      this._isHooked = true;
    }
    if (!this._enabledItems.length && this._isHooked) {
      this._window.removeEventListener ("keydown", this.onKeyDown, true);
      this._isHooked = false;
    }
  }
};
UnMHTPrototype.createInheritance (UnMHTKeyboardShortcut, UnMHTWindowHook);
