/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileSaver.jsm");
XPCOMUtils.importRelative (this, "UnMHTPrototype.jsm");
XPCOMUtils.importRelative (this, "UnMHTSave.jsm");
XPCOMUtils.importRelative (this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowHook.jsm");

let EXPORTED_SYMBOLS = ["UnMHTMAFIntegration"];

/**
 * MAF 組み込みオブジェクト
 *
 * @param  Number saveState
 *         保存状態
 * @param  Boolean useAddURIs
 *         追加の URI を使用するか
 * @implements nsICancelable
 * @implements nsIWebBrowserPersist
 */
function UnMHTMAFSavePersist (saveState, useAddURIs) {
  this._saveState = saveState;
  this._useAddURIs = useAddURIs;
  this.currentState = Ci.nsIWebBrowserPersist.PERSIST_STATE_READY;
}
UnMHTMAFSavePersist.prototype = {
  QueryInterface : XPCOMUtils.generateQI ([Ci.nsIWebBrowserPersist,
                                           Ci.nsICancelable]),
  
  /* ==== nsICancelable インターフェース ==== */
  
  /**
   * 保存キャンセル
   *
   * @param  nsresult reason
   *         理由
   */
  cancel : function (reason) {
    this.result = reason;
  },
    
  /* ==== nsIWebBrowserPersist インターフェース ==== */
  
  currentState : 0,         /* Number  状態 */
  persistFlags : 0,         /* Number  データ取得方法フラグ */
  progressListener : null,  /* nsIWebProgressListener  進捗状況リスナ */
  result : Cr.NS_OK,        /* nsresult  理由 */
  
  /**
   * 保存キャンセルイベント
   */
  cancelSave : function () {
    this.cancel (Cr.NS_BINDING_ABORTED);
  },
  
  /**
   * 保存開始イベント
   *
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  nsISupports targetFile
   *         保存先のファイル (nsIFile)
   * @param  nsISupports targetDataDir
   *         保存先のディレクトリ? (nsIFile/nsIURI)
   * @param  String outputContentType
   *         保存先のファイルの MIME-Type
   * @param  Number encodingFlags
   *         エンコーダに渡すのフラグ
   * @param  Number wrapColumn
   *         折り返し文字数
   */
  saveDocument: function (targetDocument, targetFile,
                          targetDataDir, outputContentType,
                          encodingFlags, wrapColumn) {
    try {
      this.currentState = Ci.nsIWebBrowserPersist
      .PERSIST_STATE_SAVING;
      if (this.progressListener) {
        this.progressListener.onStateChange
          (null, null,
           Ci.nsIWebProgressListener.STATE_START
           | Ci.nsIWebProgressListener.STATE_IS_NETWORK,
           Cr.NS_OK);
      }
            
      targetFile
      = targetFile.QueryInterface (Ci.nsIFileURL).file;
            
      if (!targetFile.leafName.match (/\.mht(ml)?$/i)) {
        let path = targetFile.path + ".mht";
        file
          = Cc ["@mozilla.org/file/local;1"].createInstance (Ci.nsILocalFile);
        targetFile.initWithPath (path);
      }
            
      let callback = function () {
        this.result = Cr.NS_OK;
        this._onComplete ();
      }.bind (this);
      
      let chromeWindow = Services.wm.getMostRecentWindow ("navigator:browser");
      
      UnMHTFileSaver.saveDocumentToFile
      (chromeWindow, targetDocument,
       this._saveState, this._useAddURIs,
       targetFile, callback);
    }
    catch (e) {
      if (e instanceof Ci.nsIXPCException) {
        this.result = e.result;
      } else {
        this.result = Cr.NS_ERROR_FAILURE;
      }
            
      this._onComplete ();
    }
  },
  
  /**
   * チャネルを保存する
   *
   * @param  nsIChannel channel
   *         対象のチャネル
   * @param  nsIFile file
   *         対象のファイル
   */
  saveChannel: function (channel, file) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },
  
  /**
   * URI を保存する
   *
   * @param  nsIURI URI
   *         対象の URI
   * @param  nsISupports cacheKey
   *         キャッシュオブジェクト
   * @param  nsIURI referrer
   *         リファラ
   * @param  nsIInputStream postData
   *         POST データ
   * @param  String extraHeaders
   *         拡張ヘッダ
   * @param  nsIFile file
   *         対象のファイル
   */
  saveURI : function (URI, cacheKey, referrer, postData, extraHeaders, file) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },
  
  /**
   * プライバシーを意識して URI を保存する
   *
   * @param  nsIURI URI
   *         対象の URI
   * @param  nsISupports cacheKey
   *         キャッシュオブジェクト
   * @param  nsIURI referrer
   *         リファラ
   * @param  nsIInputStream postData
   *         POST データ
   * @param  String extraHeaders
   *         拡張ヘッダ
   * @param  nsIFile file
   *         対象のファイル
   * @param  Boolean isPrivate
   *         プライベートモードか
   */
  savePrivacyAwareURI : function (URI, cacheKey, referrer, postData,
                                  extraHeaders, file, isPrivate) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },
  
  /* ==== private ==== */
  
  _saveState : UnMHTSave.SAVESTATE_UNKNOWN, /* Number  保存状態 */
  
  _useAddURIs : false, /* Boolean  追加ファイルを使用するか */
  
  /**
   * 保存完了イベント
   */
  _onComplete : function () {
    if (this.currentState == Ci.nsIWebBrowserPersist.PERSIST_STATE_FINISHED) {
      return;
    }
        
    this.currentState
    = Ci.nsIWebBrowserPersist.PERSIST_STATE_FINISHED;
        
    if (this.progressListener) {
      this.progressListener.onStateChange
        (null, null,
         Ci.nsIWebProgressListener.STATE_STOP |
         Ci.nsIWebProgressListener.STATE_IS_NETWORK,
         this.result);
    }
  }
};

/**
 * MAF 組み込み
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 * @inherits   UnMHTWindowHook
 * @implements IUnMHTWindowHook
 */
function UnMHTMAFIntegration (window) {
  UnMHTWindowHook.call (this, window);
  
  this._behaviors = [];
}
UnMHTMAFIntegration.prototype = {
  /* ==== IUnMHTWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct : function () {
    this._behaviors = null;
    
    UnMHTPrototype.super (this, "destruct").call (this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param  Boolean isStartup
   *         スタートアップイベントか
   * @param  Number reason
   *         スタートアップの理由
   */
  load : function (isStartup, reason) {
    if (!("InternalSaveBehavior" in this._window
          && "gInternalSaveBehaviors" in this._window)) {
      return;
    }
    
    let stateList = UnMHTSave.getSaveStateList (UnMHTSave.SAVESTATE_MASK_ALL);
    
    let types
    = stateList.map (function (state) {
        return {
          label : UnMHTSave.makeFiletype (state, false),
          ext : "*.mht",
          saveState : state,
          useAddURIs : false
        };
      })
    .concat (stateList.map (function (state) {
          return {
            label : UnMHTSave.makeFiletype (state, true),
            ext : "*.mht",
            saveState : state,
            useAddURIs : true
          };
        }));
    
    this._behaviors
    = types.map (function (type) {
        let newSaveBehavior = new this._window.InternalSaveBehavior ();
          
        newSaveBehavior.isComplete = true;
        newSaveBehavior.mandatoryExtension = true;
        newSaveBehavior.isValidForSaveMode = function (saveMode) {
          return saveMode & 0x01;
        };
        
        newSaveBehavior.getFileFilter
        = function (contentType, fileExtension) {
          let result = {
            title: type.label,
            extensionstring: type.ext
          };
          return result;
        };
        newSaveBehavior.getPersistObject
        = function (saveBrowsers) {
          let result;
          result
          = new UnMHTMAFSavePersist
          (type.saveState, type.useAddURIs);
          
          return result;
        };
          
        /* 末尾は *.* なのでその前に挿入する */
        this._window.gInternalSaveBehaviors.splice
        (this._window.gInternalSaveBehaviors.length - 1, 0, newSaveBehavior);
          
        return newSaveBehavior;
      }.bind (this));
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param  Boolean isShutdown
   *         シャットダウンイベントか
   * @param  Number reason
   *         シャットダウンの理由
   */
  unload : function (isShutdown, reason) {
    if (!("InternalSaveBehavior" in this._window
          && "gInternalSaveBehaviors" in this._window)) {
      return;
    }

    this._window.gInternalSaveBehaviors
    = this._window.gInternalSaveBehaviors.filter
    ((function (b) this._behaviors.indexOf (b) == -1).bind (this));
    
    this._behaviors = [];
  },
  
  /* ==== private ==== */
  
  _behaviors : null /* Array  追加した組み込み
                     *   [InternalSaveBehavior, ...] */
};
UnMHTPrototype.createInheritance (UnMHTMAFIntegration, UnMHTWindowHook);
