/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTArrayUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTUnicodeConverter.jsm");

let EXPORTED_SYMBOLS = ["UnMHTMIMEPart", "UnMHTMIMEUtil"];

/**
 * MIME の各パート情報
 */
function UnMHTMIMEPart () {
  this.fieldList = [];
  this.fields = {};
}
UnMHTMIMEPart.prototype = {
  fieldList : null, /* Array  フィールドの配列
                     *        **encode 時に必要**
                     *   [[String フィールド名, String フィールド値], ...] */
  fields : null,    /* Object  フィールドのマップ
                     *         解析用 (フィールド名は小文字)
                     *         **encode 時に必要**
                     *   <String フィールド名, String フィールド値> */
  
  /* Content-Type: (content[RFC2045]) */
  contentTypeMap : null, /* UnMHTMIMEParam  解析結果 */
  contentType : "",      /* String  type[RFC2045] */
  contentSubType : "",   /* String  subtype[RFC2045] */
  mimetype : "",         /* String  MIME-Type */
  charset : "",          /* String  charset[???] */
  isMultipart : false,   /* Boolean  マルチパートか
                          *          **encode 時に必要** */
  boundary : "",         /* String  boundary[RFC2046]
                          *         **encode 時に必要** */
  start : "",            /* String  "start"[RFC2387] */
  startMimetype : "",    /* String  "type"[RFC2387] */
  
  /* Content-Location: (content-location[RFC2557]) */
  contentLocation : "", /* String  URI[RFC2557] */
  
  /* Content-Disposition: (disposition[RFC2138]) */
  contentDispositionMap : null,     /* UnMHTMIMEParam  解析結果 */
  contentDispositionType : "",      /* String  disposition-type[RFC2138] */
  contentDispositionFilename : "",  /* String  filename-parm[RFC2138] */
  
  /* Content-ID: (id[RFC2045]) */
  contentID : "", /* String  msg-id[RFC2045] */
  
  /* Content-Transfer-Encoding: (encoding[RFC2045]) */
  contentTransferEncoding : "", /* String  mechanism[RFC2045]
                                 *         **encode 時に必要** */
  
  body : "", /* String  body[RFC5322]
              *         **encode 時に必要** */
  
  parts : null, /* Array  マルチパート
                 *        **encode 時に必要**
                 *   [UnMHTMIMEPart  body-part[RFC2046], ...] */
  
  eParam : null, /* UnMHTExtractParam  展開で使用するパラメータ */
  sParam : null  /* UnMHTSaveParam  保存で使用するパラメータ */
};

/**
 * MIME ユーティリティ
 */
let UnMHTMIMEUtil = {
  /**
   * quoted-printable[RFC2045] にエンコードする
   *
   * @param  String text
   *         エンコードする文字列
   * @param  Boolean wrap
   *         76 文字で折り返す
   * @return String
   *         エンコードした文字列
   */
  encodeQ : function (text, wrap) {
    text
    = text.replace
    (/[^\t\r\n \x21-\x3C\x3E-\x7E]/g,
     function (matched) {
      let hexString = "0" + matched.charCodeAt (0).toString (16).toUpperCase ();
      return "=" + hexString.slice (-2);
    });
    
    if (!wrap) {
      return text;
    }
    
    return (text + "\r\n").replace
    (/([^\r\n]{0,73})(?:(=..|[ \t]|[^\r\n=]{2}[^\t\r\n =])($|\r?\n|\r)|([^\t\r\n =]{0,2})(\r?\n?))/g,
     function (matched, aMain,
               aLastThree, aLastThreeEOL,
               aLastTwo, aLastTwoEOL) {
      let line = aMain + aLastThree + aLastTwo;
      if (aLastThreeEOL || aLastTwoEOL) {
        if (line) {
          let lastChar = line [line.length - 1];
          if (lastChar === " "
              || lastChar === "\t") {
            line
              = line.slice (0, -1)
              + (lastChar === " " ? "=20" : "=09");
          }
        }
        return line + "\r\n";
      }
      return line ? (line + "=\r\n") : "";
    }).slice (0, -2);
  },
    
  /**
   * quoted-printable[RFC2045] をデコードする
   *
   * @param  String text
   *         デコードする文字列
   * @return String
   *         デコードした文字列
   */
  decodeQ : function (text) {
    return text.replace
    (/=(?:\r?\n|\r|([A-Fa-f0-9]{2}))/g,
     function (matched, octet) {
      return (octet ? String.fromCharCode (parseInt (octet, 16)) : "");
    });
  },
  
  /**
   * encoded-word にエンコードする
   *
   * @param  String text
   *         エンコードする文字列 (Unicode)
   * @return String
   *         エンコードした文字列
   */
  encodeWord : function (charset, type, text) {
    let converter
    = Cc ["@mozilla.org/intl/scriptableunicodeconverter"]
    .getService (Ci.nsIScriptableUnicodeConverter);
    converter.charset = charset;
    text = converter.ConvertFromUnicode (text);
    
    if (type == "B") {
      text = this.encodeBase64 (text, false);
    }
    else if (type == "Q") {
      text
      = this.encodeQ (text, false)
      .replace (/\?/g, "=3F")
      .replace (/ /g, "=20");
    }
    
    return ("=?" + charset + "?" + type + "?" + text + "?=");
  },
  
  /**
   * encoded-word をデコードする
   *
   * @param  String text
   *         デコードする文字列
   * @return String
   *         デコードした文字列 (Unicode)
   */
  decodeWord : function (text) {
    return  text.replace
    (/=\?([^\?]+)\?([^\?])\?([^\?]+)\?=/g,
     function (matched, charset, type, v) {
      if (type == "Q" || type == "q") {
        v = this.decodeQ (v);
      }
      else if (type == "B" || type == "b") {
        v = this.decodeBase64 (v);
      }
      
      return UnMHTUnicodeConverter.convertToUnicode (v, charset);
    }.bind (this));
  },
  
  /**
   * boundary[RFC2046] を作成する
   *
   * @return String
   *         バウンダリ文字列
   */
  createUniqueBoundary : function () {
    let randomHex = function () {
      return Math.floor (Math.random () * 15).toString (16).toUpperCase ();
    };
    
    return ["----=_Part_"].concat
    ([randomHex() for (i in UnMHTArrayUtil.range (1, 8))],
     "_",
     [randomHex() for (i in UnMHTArrayUtil.range (1, 8))],
     ".",
     (new Date ()).getTime ()).join ("");
  },
  
  /**
   * 日時を date-time[RFC5322] 形式にフォーマットする
   *
   * @param  Date date
   *         現在の日時
   * @return String
   *         フォーマットされた日時
   */
  formatDate : function (date) {
    let weekDay = [
      "Sun",
      "Mon",
      "Tue",
      "Wed",
      "Thu",
      "Fri",
      "Sat"
      ];
    let month = [
      "Jan",
      "Feb",
      "Mar",
      "Apr",
      "May",
      "Jun",
      "Jul",
      "Aug",
      "Sep",
      "Oct",
      "Nov",
      "Dec"
      ];
    
    let pad = function (x, n) {
      return ("0000" + n).slice (-x);
    };
    
    let sign = "-";
    let zone = date.getTimezoneOffset () * 100 / 60;
    if (zone < 0) {
      zone = -zone;
      sign = "+";
    }
    
    return (weekDay [date.getDay ()] + ", "
            + pad (2, date.getDate ()) + " "
            + month [date.getMonth ()] + " "
            + (date.getFullYear ()) + " "
            + pad (2, date.getHours ()) + ":"
            + pad (2, date.getMinutes ()) + ":"
            + pad (2, date.getSeconds ()) + " "
            + sign + pad (4, zone));
  },
  
  /**
   * 拡張子に対応した MIME-Type を返す
   *
   * @param  String ext
   *         拡張子
   * @return String
   *         対応した MIME-Type
   */
  getTypeFromExtension : function (ext) {
    /* システム設定が特殊場合に供えて既知のものは先にチェック */
    
    if (ext.match (/^[sx]?html?$/i)) {
      /* XHTML も text/html として扱う */
      return "text/html";
    }
    
    let knownExts = {
      "css"   : "text/css",
      "mht"   : "message/rfc822",
      "mhtml" : "message/rfc822",
      "eml"   : "message/rfc822",
      "js"    : "text/javascript",
      "txt"   : "text/plain",
      "jpg"   : "image/jpeg",
      "jpeg"  : "image/jpeg",
      "gif"   : "image/gif",
      "png"   : "image/png",
      "bmp"   : "image/bmp",
      "ico"   : "image/x-icon",
      "svg"   : "image/svg+xml",
      "swf"   : "application/swf",
      "jar"   : "application/java-archive",
      "class" : "application/octet-stream"
    };
    
    if (ext.toLowerCase () in knownExts) {
      return knownExts [ext.toLowerCase ()];
    }
    
    if (ext) {
      /* 未知なのでシステム設定を返す */
      let mime = Cc ["@mozilla.org/mime;1"].getService (Ci.nsIMIMEService);
      
      try {
        return mime.getTypeFromExtension (ext);
      }
      catch (e) {
      }
    }
    
    return "application/octet-stream";
  },
  
  /**
   * Base64 エンコードする
   *
   * @param  String text
   *         エンコードする文字列
   * @param  Boolean wrap
   *         76 文字で折り返す
   * @return String
   *         エンコードされた文字列
   */
  encodeBase64 : function (text, wrap) {
    text = btoa (text);
    if (!wrap) {
      return text;
    }
    
    return text.replace (/.{76}/g, "$&\r\n");
  },

  /**
   * Base64 デコードする
   * 改行を含むものや破損したものにも対応
   *
   * @param  String text
   *         デコードする文字列
   * @return String
   *         デコードされた文字列
   */
  decodeBase64 : function (text) {
    text = text.replace (/[^A-Za-z0-9+\/=]+/g, "");
    return atob (text.substr (0, text.length - text.length % 4));
  },
  
  /**
   * message [RFC5322] にエンコードする
   *
   * @param  UnMHTMIMEPart part
   *         トップレベルのパート
   * @param  nsIOutputStream stream
   *         メッセージ出力用ストリーム
   *         改行コードは CR LF である
   */
  encodeMessage : function (part, stream) {
    this.encodeFields (part, stream);
    
    let data = "\r\n";
    stream.write (data, data.length);

    if (part.isMultipart) {
      part.parts.forEach (function (p) {
          let data = "--" + part.boundary + "\r\n";
          stream.write (data, data.length);
          this.encodeMessage (p, stream);
          data = "\r\n";
          stream.write (data, data.length);
        }.bind (this));
      
      let data = "--" + part.boundary + "--\r\n";
      stream.write (data, data.length);
      
      return;
    }
    
    if (part.contentTransferEncoding == "quoted-printable") {
      data = this.encodeQ (part.body, true);
    }
    else if (part.contentTransferEncoding == "base64") {
      data = this.encodeBase64 (part.body, true) + "\r\n";
    }
    else {
      data = part.body;
    }
    stream.write (data, data.length);
  },
  
  /**
   * fields[RFC5322] にエンコードする
   *
   * @param  UnMHTMIMEPart part
   *         トップレベルのパート
   * @param  nsIOutputStream stream
   *         メッセージ出力用ストリーム
   *         改行コードは CR LF である
   */
  encodeFields : function (part, stream) {
    let order = [
      "from",
      "to",
      "subject",
      "date",
      "mime-version",
      "content-location",
      "content-base",
      "content-id",
      "content-disposition",
      "content-type",
      "content-transfer-encoding"
      ];
    let PLAIN        = 1;
    let ENCODED_WORD = 2;
    let PARAMETER    = 3;
    let UNICODE      = 4;
    let types = {
      "from":                      PLAIN,
      "to":                        PLAIN,
      "subject":                   ENCODED_WORD,
      "date":                      PLAIN,
      "mime-version":              PLAIN,
      "content-type":              PARAMETER,
      "content-transfer-encoding": PLAIN,
      "content-location":          UNICODE,
      "content-base":              UNICODE,
      "content-disposition":       PARAMETER,
      "content-id":                PLAIN
    };
    
    part.fieldList.sort (function ([name1,], [name2,]) {
        let p1 = order.indexOf (name1.toLowerCase ());
        let p2 = order.indexOf (name2.toLowerCase ());
        if (p1 >= 0) {
          if (p2 >= 0) {
            return p1 - p2;
          }
          else {
            return -1;
          }
        }
        else {
          if (p2 >= 0) {
            return 1;
          }
          else {
            return (name1 < name2) ? -1 : ((name1 > name2) ? 1 : 0);
          }
        }
      })
    .map (function ([name, value]) {
        let type = PLAIN;
        let lname = name.toLowerCase ();
        if (lname in types) {
          type = types [lname];
        };
          
        if (type == UNICODE) {
          try {
            value = UnMHTUnicodeConverter.convertFromUnicode (value, "utf-8");
          }
          catch (e) {
          }
        }
        else if (type == ENCODED_WORD) {
          if (!value.match (/^[\x20-\x7E]*$/)) {
            value = this.encodeWord ("utf-8", "B", value);
          }
        }
        else if (type == PARAMETER) {
          /* Emacs がおかしくなるので一箇所だけダブルクォーテーションを
           * エスケープシーケンスで書く */
          value = value.replace
            (/(([^\t\r\n \*\'%\(\)<>@,;:\\\"\/\[\]\?=]+)(?:\*([0-9]+))?(\*)?=(?:([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]*)\'([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]*)\'([^\t\r\n \*\'\(\)<>@,;:\\\"\/\[\]\?=]+)|\"((?:\\.|[^\"])*)\"|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+))|([^\t\r\n \(\)<>@,;:\\\x22\/\[\]\?=]+)\/([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+));[ \t]*/g,
             function (matched, param) {
              return param + ";\r\n\t";
            });
        }
          
        return name + ": " + value + "\r\n";
      }.bind (this))
    .forEach (function (data) stream.write (data, data.length));
  },
  
  /**
   * message [RFC5322] をデコードする
   *
   * @param  String text
   *         メッセージ
   *         改行コードは CR LF でなければならない
   * @param  Boolean allowRemain
   *         解析できないフィールドの存在を許可するか
   * @return UnMHTMIMEPart
   *         トップレベルのパート
   *         データが不正ならば null
   */
  decodeMessage : function (text, allowRemain) {
    let fields = "";
    let body = "";
    
    let pos = text.indexOf ("\r\n\r\n");
    if (pos != -1) {
      fields = text.substr (0, pos + 2);
      body = text.substr (pos + 4);
    }
    else {
      fields = text;
    }
    
    let part = this.decodeFields (fields, allowRemain);
    if (!part) {
      return null;
    }

    if (part.isMultipart) {
      part.parts = this.decodeMultipart (body, part.boundary);
    }
    else {
      if (part.contentTransferEncoding == "quoted-printable") {
        part.body = this.decodeQ (body);
      }
      else if (part.contentTransferEncoding == "base64") {
        part.body = this.decodeBase64 (body);
      }
      else {
        part.body = body;
      }
    }
    
    return part;
  },
  
  /**
   * multipart-body[RFC2046] をデコードする
   *
   * @param  String body
   *         multipart なメッセージボディ
   *         改行コードは CR LF でなければならない
   * @param  String boundary
   *         バウンダリ文字列
   * @return Array
   *         パートの配列
   *           [UnMHTMIMEPart, ...]
   */
  decodeMultipart : function (body, boundary) {
    let b = boundary
    .replace (/([\\\[\]\(\)\^\,\.\{\}\|\?\!\-\*\+\$])/g, "\\$1");
    
    let datas = ("\r\n" + body)
    .split (new RegExp ("\r\n--" + b + "(?:--)?(?:\r\n)?"));
    return datas.slice (1, datas.length - 1)
    .map ((function (data) this.decodeMessage (data)).bind (this))
    .filter (function (part) part != null);
  },
  
  /**
   * fields[RFC5322] をデコードする
   *
   * @param  String fieldsString
   *         フィールド文字列
   *         改行コードは CR LF でなければならない
   * @param  Boolean allowRemain
   *         解析できないフィールドの存在を許可するか
   * @return UnMHTMIMEPart
   *         パート
   *         データが不正ならば null
   */
  decodeFields : function (fieldsString, allowRemain) {
    let part = new UnMHTMIMEPart ();
    
    let remain
    = fieldsString.replace
    (/([\x21-\x39\x3B-\x7E]+)[ \t]*:[ \t]*([^\r\n]*)\r\n((?:[\t ]+[^\r\n]*\r\n)*)/g,
     function (matched, name, value, restLines) {
      restLines.replace
      (/[\t ]*([^\r\n]*)\r\n/g,
       function (matched, v) {
        value += v;
          
        return "";
      });
       
      this.addField (part, name, value);
      
      return "";
    }.bind (this));
    if (remain) {
      let re = remain.match (/^(From)[ \t]+(.*)$/);
      if (re) {
        /* RFC4155: application/mbox */
        remain = "";
      }
    }
    if (!allowRemain && remain) {
      return null;
    }
    
    if ("content-type" in part.fields) {
      let value = part.fields ["content-type"];
      part.contentTypeMap = this.decodeParam (value);
      part.contentType    = part.contentTypeMap.type.toLowerCase ();
      part.contentSubType = part.contentTypeMap.subtype.toLowerCase ();
      part.mimetype       = part.contentType + "/" + part.contentSubType;
      if ("charset" in part.contentTypeMap.params) {
        part.charset  = part.contentTypeMap.params ["charset"];
      }
    }
    
    if ("content-location" in part.fields) {
      part.contentLocation
      = this.tryConvertToUnicode (part.fields ["content-location"]);
    }
    
    if ("content-disposition" in part.fields) {
      let value = part.fields ["content-disposition"];
      part.contentDispositionMap  = this.decodeParam (value);
      part.contentDispositionType
      = part.contentDispositionMap.type.toLowerCase ();
      if ("filename" in part.contentDispositionMap.params) {
        part.contentDispositionFilename
          = this.tryConvertToUnicode
          (part.contentDispositionMap.params ["filename"]);
      }
    }
    
    if ("content-id" in part.fields) {
      let value = part.fields ["content-id"];
      part.contentID = value.replace (/^<([^>]+)>$/, "$1");
    }
    
    if (part.contentType == "multipart") {
      part.isMultipart = true;
      if ("boundary" in part.contentTypeMap.params) {
        part.boundary = part.contentTypeMap.params ["boundary"];
      }
      else {
        return null;
      }
      
      if ("start" in part.contentTypeMap.params) {
        let value = part.contentTypeMap.params ["start"];
        part.start = value.replace (/^<([^>]+)>$/, "$1");
      }
      
      if ("type" in part.contentTypeMap.params) {
        part.startMimetype = part.contentTypeMap.params ["type"];
      }
    }
    else {
      if ("content-transfer-encoding" in part.fields) {
        let value = part.fields ["content-transfer-encoding"];
        part.contentTransferEncoding = value.toLowerCase ();
      }
    }
    
    return part;
  },
  
  /**
   * 開始パートを探す
   *
   * @param  UnMHTMIMEPart part
   *         走査中のパート
   * @return UnMHTMIMEPart
   *         開始パート
   *         適切なパートが無ければ null
   */
  findStartPart : function (part) {
    if (part.isMultipart) {
      if (part.contentSubType == "mixed") {
        return part;
      }
      
      if (part.start) {
        let starts
        = part.parts.filter (function (p) p.contentID == part.start);
        if (starts.length) {
          return this.findStartPart (starts [0]);
        }
      }
      
      if (part.contentSubType == "related") {
        let starts
        = part.parts.filter (function (p)
                             p.mimetype == part.startMimetype);
        if (starts.length) {
          return this.findStartPart (starts [0]);
        }
      }
      
      if (part.contentSubType == "alternative") {
        let starts = part.parts.filter (function (p)
                                        p.mimetype == "text/html");
        if (starts.length) {
          return starts [0];
        }
        
        starts
        = part.parts.filter (function (p) {
            if (p.isMultipart) {
              let startPart = this.findStartPart (p);
              if (startPart && startPart.mimetype == "text/html") {
                return true;
              }
            }
          }.bind (this));
        if (starts.length) {
          return this.findStartPart (starts [0]);
        }
      }
      
      if (part.parts.length) {
        return this.findStartPart (part.parts [0]);
      }
      
      return null;
    }
    
    return part;
  },
  
  /**
   * content[RFC2045] や disposition[RFC2183] の値を
   * parameter[RFC2231] に従ってデコードする
   * 
   * ただし disposition-type[RFC2183] や type/subtype[RFC2045] の位置が
   * 先頭でない事例が存在するのでこれらのデコードも同時に行う
   *
   * @param  String value
   *         contentRFC2045] や disposition[RFC2183] の値
   * @return UnMHTMIMEParam
   *         {
   *           type    : String  disposition-type[RFC2183]/type[RFC2045],
   *           subtype : String  subtype[RFC2045],
   *           params  : Object パラメータのマップ
   *                       <String パラメータ名, String パラメータ値>
   *         }
   */
  decodeParam : function (value) {
    let map = {};
    let params = {};
    let type = "";
    let subtype = "";
    
    value.replace
    /*
     * ( 1*attribute-char [section] ["*"] "="
     *   ( ( [token] "'" [token] "'" 1*<attribute-char and %> ) /
     *     quoted-string / token ) ) /
     * ( token "/" token ) /
     * token
     */
    (/([^\t\r\n \*\'%\(\)<>@,;:\\\"\/\[\]\?=]+)(?:\*([0-9]+))?(\*)?=(?:([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]*)\'([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]*)\'([^\t\r\n \*\'\(\)<>@,;:\\\"\/\[\]\?=]+)|\"((?:\\.|[^\"])*)\"|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+))|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)\/([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)/g,
     function (matched, name, section, extended,
               charset, lang, evalue,
               qvalue, value,
               ctype, csubtype, dtype) {
      if (dtype && !type) {
        type = dtype;
      }
      else if (ctype && !type) {
        type = ctype;
        subtype = csubtype;
      }
      else if (name) {
        if (!value) {
          value = qvalue;
        }
        
        if (!(name in map)) {
          map [name] = {
            values : {},
            charset : "",
            lang : ""
          };
        }
        
        if (!section) {
          section = "0";
        }
        
        if (extended && evalue) {
          map [name].values [section] = { v: evalue, e: true };
          map [name].charset = charset;
          map [name].lang = lang;
        }
        else {
          /* 実際にはここに encoded-word は使えないが使用事例がある */
          value = this.decodeWord (value);
          
          map [name].values [section] = { v: value, e: extended != "" };
        }
      }
    }.bind (this));
    
    for (let [name, param] in Iterator (map)) {
      params [name]
        = [x for (x in Iterator (param.values))]
        .sort (function (a, b) parseInt (a [0], 10) - parseInt (b [0], 10))
        .map (function ([n, v]) {
            if (v.e) {
              if (param.charset) {
                return UnMHTUnicodeConverter.convertToUnicode
                (unescape (v.v), param.charset);
              }
              else {
                return unescape (v.v);
              }
            }
            else {
              return v.v;
            }
          })
        .join ("");
    }
    
    return {
      type: type,
      subtype: subtype,
      params: params
    };
  },

  /**
   * パートにフィールドを追加する
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   * @param  String name
   *         フィールド名
   * @param  String value
   *         フィールド値
   */
  addField : function (part, name, value) {
    part.fieldList.push ([name, value]);
    part.fields [name.toLowerCase ()] = value;
  },
  
  /**
   * Unicode への変換を試みる
   *
   * @param  String value
   *         対象の文字列
   * @return String
   *         成功したら変換した文字列
   *         失敗したら元の文字列
   */
  tryConvertToUnicode : function (value) {
    try {
      return UnMHTUnicodeConverter.convertToUnicode (value, "utf-8");
    }
    catch (e) {
    }
    try {
      return UnMHTUnicodeConverter.convertToUnicode (value, "shift_jis");
    }
    catch (e) {
    }
    return value;
  },
  
  /**
   * パートのデータを複製する
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   * @param  Object findPart
   *         検索するパート
   *         {
   *           from: UnMHTMIMEPart
   *           to: UnMHTMIMEPart
   *         }
   * @return UnMHTMIMEPart
   *         複製したパート
   */
  clonePart : function (part, findPart) {
    let newPart = new UnMHTMIMEPart ();
    
    if (findPart && part == findPart.from) {
      findPart.to = newPart;
    }
    
    for (let [name, value] in Iterator (part)) {
      if (name == "fieldList") {
        newPart.fieldList = value.map (function ([name, value]) [name, value]);
      }
      else if (name == "fields") {
        for (let [fName, fValue] in Iterator (value)) {
          newPart.fields [fName] = fValue;
        }
      }
      else if (name == "contentTypeMap"
               || name == "contentDispositionMap") {
        let newParams = {};
        for (let [pName, pValue] in Iterator (value.params)) {
          newParams [pName] = pValue;
        }
        
        newPart [name] = {
          type : value.type,
          subtype : value.subtype,
          params : newParams
        };
      }
      else if (name == "parts") {
        newPart.parts
          = value.map ((function (p) this.clonePart (p, findPart)).bind (this));
      }
      else if (name == "eParam"
               || name == "sParam") {
        /* 複製しない */
      }
      else {
        newPart [name] = value;
      }
    }
    
    return newPart;
  },
  
  /**
   * パートのデータを削除する
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   */
  destructPart : function (part) {
    if (part.parts) {
      part.parts.forEach (function (p) {
          this.destructPart (p);
        }.bind (this));
      part.parts = null;
    }
    part.fieldList = null;
    part.fields = null;
    part.contentTypeMap = null;
    part.contentDispositionMap = null;
    part.body = "";
  },
  
  /**
   * fieldList から fields を作成する
   *
   * @param  UnMHTMIMEPart part
   *         対象のパート
   */
  createFieldsFromList : function (part) {
    part.fields = {};
    part.fieldList.forEach (function ([name, value]) {
        part.fields [name.toLowerCase ()] = value;
      });
  }
};
