/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTAppInfo.jsm");
XPCOMUtils.importRelative (this, "UnMHTArrayUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTUnicodeConverter.jsm");
XPCOMUtils.importRelative (this, "UnMHTWinRegUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTMail"];

/**
 * メール送信
 */
let UnMHTMail = {
  /* アプリケーションの引数 */
  thunderbirdArg : UnMHTPref.thunderbirdArg,
  seamonkeyArg   : UnMHTPref.seamonkeyArg,
  macmailArg     : UnMHTPref.macmailArg,
  
  /**
   * メールを送信する
   *
   * @param  nsIFile targetFile
   *         添付ファイル
   * @param  String title
   *         ページのタイトル
   * @param  Function debugFunc
   *         デバッグ情報を受け取る関数
   * @return Boolean
   *         成功したか
   */
  send : function (window, targetFile, title) {
    if (UnMHTPref.getPref ("save.mail.zip")) {
      targetFile = this._createZipFile (targetFile);
    }
    
    if (UnMHTPref.getPref ("save.mail.app.type") == "seamonkey"
        && UnMHTAppInfo.isSeaMonkey ()) {
      return this._openComposeWindow (targetFile, title);
    }
    
    let [app, arg] = this._getApplicationAndArgument ();
    arg = this._replaceVars (arg, targetFile, title);
    let args = this._argToArray (arg);
    
    if (UnMHTPref.getPref ("save.mail.debug")) {
      window.alert
        ("[Debug]\n"
         + "Command line for mail\n"
         + "\n"
         + (app ? app.path : "???") + " " + arg + "\n"
         + "\n"
         + "Arguments will be passed as following array\n"
         + "\n"
         + args.join ("\n"));
    }
    
    try {
      let process
      = Cc ["@mozilla.org/process/util;1"].createInstance (Ci.nsIProcess);
      process.init (app);
      process.runw (false, args, args.length);
    }
    catch (e) {
      return false;
    }
    
    return true;
  },
  
  /* ==== private ==== */
  
  /**
   * スペース区切りの引数を引数配列に変換する
   *
   * @param  String argString
   *         スペース区切りの引数
   * @return Array
   *         引数配列
   */
  _argToArray : function (argString) {
    let STATE_NORMAL          = 0;
    let STATE_ESCAPED         = 1;
    let STATE_DQUOTED         = 10;
    let STATE_DQUOTED_ESCAPED = 11;
    let STATE_QUOTED          = 20;
    let STATE_QUOTED_ESCAPED  = 21;
    
    let args = [];
    let currentArg = "";
    let state = STATE_NORMAL;
    
    Array.forEach
    (argString,
     function (c) {
      switch (state) {
        case STATE_NORMAL: {
          if (c == "\\") {
            state = STATE_ESCAPED;
          }
          else if (currentArg == "" && c == "\"") {
            state = STATE_DQUOTED;
          }
          else if (currentArg == "" && c == "\'") {
            state = STATE_QUOTED;
          }
          else if (c == " ") {
            if (currentArg != "") {
              args.push (currentArg);
            }
            currentArg = "";
          }
          else {
            currentArg += c;
          }
          break;
        }
        case STATE_ESCAPED: {
          currentArg += c;
          break;
        }
        case STATE_DQUOTED: {
          if (c == "\"") {
            state = STATE_NORMAL;
          }
          else if (c == "\\") {
            state = STATE_DQUOTED_ESCAPED;
          }
          else {
            currentArg += c;
          }
          break;
        }
        case STATE_DQUOTED_ESCAPED: {
          currentArg += c;
          state = STATE_DQUOTED;
          break;
        }
        case STATE_QUOTED: {
          if (c == "\'") {
            state = STATE_NORMAL;
          }
          else if (c == "\\") {
            state = STATE_QUOTED_ESCAPED;
          }
          else {
            currentArg += c;
          }
          break;
        }
        case STATE_QUOTED_ESCAPED: {
          currentArg += c;
          state = STATE_QUOTED_ESCAPE;
          break;
        }
      }
    });
    if (currentArg != "") {
      args.push (currentArg);
    }
    
    return args;
  },
  
  /**
   * SeaMonkey でメール作成画面を開く
   *
   * @param  nsIFile targetFile
   *         添付ファイル
   * @param  String title
   *         ページのタイトル
   * @return Boolean
   *         成功したか
   */
  _openComposeWindow : function (targetFile, title) {
    let params
    = Cc ["@mozilla.org/messengercompose/composeparams;1"]
    .createInstance (Ci.nsIMsgComposeParams);
    
    params.composeFields
    = Cc ["@mozilla.org/messengercompose/composefields;1"]
    .createInstance (Ci.nsIMsgCompFields);
    params.composeFields.subject = title;
    
    let attachmentData
    = Cc ["@mozilla.org/messengercompose/attachment;1"]
    .createInstance (Ci.nsIMsgAttachment);
    attachmentData.url = UnMHTFileUtil.getURLSpecFromFile (targetFile);
    
    params.composeFields.addAttachment (attachmentData);
    
    let composeService
    = Cc ["@mozilla.org/messengercompose;1"]
    .getService (Ci.nsIMsgComposeService);
    
    try {
      params.identity = composeService.defaultIdentity;
    }
    catch (e) {
      params.identity = null;
    }
    
    composeService.OpenComposeWindowWithParams (null, params);
    
    return true;
  },
  
  /**
   * 使用するアプリケーションと引数を取得する
   *
   * @return Array
   *           [nsIFile アプリケーション
   *                    見付からない場合は null,
   *            String 引数]
   */
  _getApplicationAndArgument : function () {
    let arg;
    switch (UnMHTPref.getPref ("save.mail.arg.type")) {
      case "thunderbird": {
        arg = this.thunderbirdArg;
        break;
      }
      case "seamonkey": {
        arg = this.seamonkeyArg;
        break;
      }
      case "macmail": {
        arg = this.macmailArg;
        break;
      }
      default: {
        arg = UnMHTPref.getPref ("save.mail.arg.other");
      }
    }
    
    let app;
    switch (UnMHTPref.getPref ("save.mail.app.type")) {
      case "thunderbird": {
        app = this._getThunderbird ();
        break;
      }
      case "seamonkey": {
        app = this._getSeamonkey ();
        break;
      }
      case "macmail": {
        app = this._generateMacMailScript ();
        break;
      }
      default: {
        let path = UnMHTPref.getPref ("save.mail.app");
        if (path) {
          app = UnMHTFileUtil.initFile (path);
        }
        else {
          app = this._getThunderbird ();
          if (app) {
            arg = this.thunderbirdArg;
          }
        }
      }
    }
    
    return [app, arg];
  },
  
  /**
   * 引数中の変数を置換する
   *
   * @param  String arg
   *         引数
   * @param  nsIFile targetFile
   *         添付ファイル
   * @param  String title
   *         ページのタイトル
   * @return String
   *         置換した引数
   */
  _replaceVars : function (arg, targetFile, title) {
    let targetPath = targetFile.path;
    
    let targetURI
    = UnMHTFileUtil.getURLSpecFromFile (targetFile)
    .replace (/,/g, "%2C");
    
    let vars = {
      "FILE"       : targetPath.replace (/\\/g, "\\\\"),
      "URI"        : targetURI.replace (/\\/g, "\\\\"),
      "TITLE"      : (title
                      .replace (/\\/g, "\\\\")
                      .replace (/[\'\"]/g, "")),
      "TITLE_NOSP" : (title
                      .replace (/\\/g, "\\\\")
                      .replace (/[\'\"]/g, "")
                      .replace (/[ \t\r\n]/g, "_"))
    };
    
    let rexp = new RegExp ("%(" + Object.keys (vars).join ("|") + ")%", "g");
    return arg.replace (rexp, function (matched, name) vars [name]);
  },
  
  /**
   * Thunderbird の実行ファイルを探す
   *
   * @return nsIFile
   *         Thunderbird の実行ファイル
   *         見付からない場合は null
   */
  _getThunderbird : function () {
    let list;
    switch (Cc ["@mozilla.org/xre/app-info;1"]
            .getService (Ci.nsIXULRuntime).OS) {
      case "Darwin": {
        list = [
          [UnMHTFileUtil.getApplicationsDirectory (),
           "Thunderbird.app", "Contents", "MacOS", "thunderbird-bin"],
          [UnMHTFileUtil.getUserApplicationsDirectory (),
           "Thunderbird.app", "Contents", "MacOS", "thunderbird-bin"]
          ];
        break;
      }
      case "WINNT": {
        let app = null;
        
        let currentVersion
          = UnMHTWinRegUtil.getHKLMValue
          (["SOFTWARE", "Mozilla", "Mozilla Thunderbird"], "CurrentVersion");
        if (currentVersion) {
          let PathToExe
            = UnMHTWinRegUtil.getHKLMValue
            (["SOFTWARE", "Mozilla", "Mozilla Thunderbird", currentVersion,
              "Main"],
             "PathToExe");
          if (PathToExe) {
            app = UnMHTFileUtil.initFile (PathToExe);
          }
        }
        
        list = [
          [app],
          [UnMHTFileUtil.getProgramFilesDirectory (),
           "Mozilla Thunderbird", "thunderbird.exe"],
          [UnMHTFileUtil.getProgramFilesDirectory (),
           "thunderbird", "thunderbird.exe"]
          ];
        break;
      }
      default: {
        list = [];
      }
    }
    
    return this._getApplication (list);
  },

  /**
   * SeaMonkey の実行ファイルを探す
   *
   * @return nsIFile
   *         SeaMonkey の実行ファイル
   *         見付からない場合は null
   */
  _getSeamonkey : function () {
    let list;
    switch (Cc ["@mozilla.org/xre/app-info;1"]
            .getService (Ci.nsIXULRuntime).OS) {
      case "Darwin": {
        list = [
          [UnMHTFileUtil.getApplicationsDirectory (),
           "SeaMonkey.app", "Contents", "MacOS", "seamonkey-bin"],
          [UnMHTFileUtil.getUserApplicationsDirectory (),
           "SeaMonkey.app", "Contents", "MacOS", "seamonkey-bin"]
          ];
        break;
      }
      case "WINNT": {
        let app = null;
        
        let currentVersion
          = UnMHTWinRegUtil.getHKLMValue
          (["SOFTWARE", "Mozilla", "SeaMonkey"], "CurrentVersion");
        if (currentVersion) {
          let PathToExe
            = UnMHTWinRegUtil.getHKLMValue
            (["SOFTWARE", "Mozilla", "SeaMonkey", currentVersion, "Main"],
             "PathToExe");
          if (PathToExe) {
            app = UnMHTFileUtil.initFile (PathToExe);
          }
        }
        
        list = [
          [app],
          [UnMHTFileUtil.getProgramFilesDirectory (),
           "mozilla.org", "SeaMonkey", "seamonkey.exe"],
          [UnMHTFileUtil.getProgramFilesDirectory (),
           "SeaMonkey", "seamonkey.exe"]
          ];
        break;
      }
      default: {
        list = [];
      }
    }
    
    return this._getApplication (list);
  },
  
  /**
   * 存在するアプリケーションを返す
   *
   * @param  Array list
   *         アプリケーションのパスの候補
   *           [[nsIFile 親ディレクトリ, String サブディレクトリ, ...], ...]
   * @return nsIFile
   *         アプリケーションのファイル
   *         見付からない場合は null
   */
  _getApplication : function (list) {
    return UnMHTArrayUtil.find
    (list.map (function (paths) paths.slice (1).reduce (function (file, path) (file ? (file.append (path), file) : null), paths [0])),
     function (file) file && file.exists (),
     null);
  },
  
  /**
   * Mac で Mail を起動するスクリプトを作成する
   *
   * @return nsIFile
   *         Mail を起動するスクリプト
   */
  _generateMacMailScript : function () {
    let data
    = "#!/usr/bin/osascript" + "\n"
    + "on run argv" + "\n"
    + "  tell application \"Mail\"" + "\n"
    + "    tell content of (make new outgoing message with properties {visible:true, subject:(item 1 of argv)})" + "\n"
    + "      make new attachment with properties {file name:(item 2 of argv)} at after the last paragraph" + "\n"
    + "    end tell" + "\n"
    + "    activate" + "\n"
    + "  end tell" + "\n"
    + "end run" + "\n";
    
    let macmailFile = UnMHTFileUtil.getCacheDirectory ().clone ();
    macmailFile.append ("macmailscript");
    
    let fstream = UnMHTFileUtil.openOutputStream (macmailFile, "0755");
    fstream.write (data, data.length);
    fstream.close ();
    
    return macmailFile;
  },
  
  /**
   * ファイルを ZIP で圧縮する
   *
   * @param  nsIFile targetFile
   *         圧縮するファイル
   * @return nsIFile
   *         圧縮したファイル
   */
  _createZipFile : function (targetFile) {
    let zipFile = targetFile.parent.clone ();
    zipFile.append
    (targetFile.leafName.replace (/\.mht(ml)?$/i, "") + ".zip");
    
    let zip = UnMHTFileUtil.openZipWriter (zipFile);
    
    zip.addEntryFile
    (targetFile.leafName, Ci.nsIZipWriter.COMPRESSION_DEFAULT, targetFile,
     false);
        
    zip.close ();
            
    return zipFile;
  }
};
