/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTDOMUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTPrototype.jsm");
XPCOMUtils.importRelative (this, "UnMHTSave.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowHook.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTMenuClass"];

/**
 * メニュー管理
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 * @inherits   UnMHTWindowHook
 * @implements IUnMHTWindowHook
 */
function UnMHTMenuClass (window) {
  UnMHTWindowHook.call (this, window);
  
  this.onFileMenuShowing = this.onFileMenuShowing.bind (this);
  this.onContextMenuShowing = this.onContextMenuShowing.bind (this);
}
UnMHTMenuClass.prototype = {
  /**
   * メニューが開かれるイベント
   * メニューの項目の表示／非表示を設定する
   *
   * @param  Event event
   *         対象のイベント
   */
  onContextMenuShowing : function (event) {
    if (!this._frameMenuItems) {
      this._insertFrameMenuItem ();
    }
    
    this._setMenuItem ("context", true);
    this._setMenuItem ("frame-context", true,
                       this._window.document.popupNode.ownerDocument);
  },
  
  /**
   * ファイルメニューを設定する
   *
   * @param  Event event
   *         対象のイベント
   */
  onFileMenuShowing : function (event) {
    this._setMenuItem ("command");
  },
  
  /**
   * ツールバーのメニューを設定する
   */
  onToolbarMenuShowing : function () {
    this._setMenuItem ("toolbar");
  },

  /**
   * ツールメニューを設定する
   */
  onToolMenuShowing : function () {
    this._setMenuItem ("toolmenu");
  },
  
  /**
   * ステータスバーのメニューを設定する
   */
  onStatusbarMenuShowing : function () {
    this._setMenuItem ("statusbar");
  },

  /* ==== IUnMHTWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct : function () {
    UnMHTPrototype.super (this, "destruct").call (this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param  Boolean isStartup
   *         スタートアップイベントか
   * @param  Number reason
   *         スタートアップの理由
   */
  load : function (isStartup, reason) {
    this._window.UnMHTMenu = this;
    
    let popup;
    popup = this.el ("menu_FilePopup");
    if (popup) {
      popup.addEventListener ("popupshowing",this.onFileMenuShowing, false);
    }
    
    popup = this.el ("contentAreaContextMenu");
    if (popup) {
      popup.addEventListener ("popupshowing", this.onContextMenuShowing, false);
    }
    
    let f;
    f = function () {
      this._enableContextMenuSave
      = UnMHTPref.getPref ("menu.context.save");
      this._enableContextMenuQuickSave
      = UnMHTPref.getPref ("menu.context.save.quick");
      this._enableContextMenuSendmail
      = UnMHTPref.getPref ("menu.context.save.mail");
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "menu.context", f, this);
    f ();
    
    f = function () {
      this._enableMenuAllCache
      = UnMHTPref.getPref ("menu.all.cache");
      this._enableMenuAllExtract
      = UnMHTPref.getPref ("menu.all.extract");
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "menu.all", f, this);
    f ();

    let menu = this.el ("unmht-tools-preferences");
    if (menu) {
      f = function () {
        menu.hidden = !UnMHTPref.getPref ("menu.tool");
      };
      UnMHTPref.addListener (UnMHTPref.prefix + "menu.tool", f, this);
      f ();
    }
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param  Boolean isShutdown
   *         シャットダウンイベントか
   * @param  Number reason
   *         シャットダウンの理由
   */
  unload : function (isShutdown, reason) {
    UnMHTPref.removeAllListeners (this);
    
    if (this._frameMenuItems) {
      this._frameMenuItems.forEach (function (n) n.parentNode.removeChild (n));
      this._frameMenuItems = null;
    }
    
    let popup;
    popup = this.el ("contentAreaContextMenu");
    if (popup) {
      popup.removeEventListener ("popupshowing", this.onContextMenuShowing);
    }
    
    popup = this.el ("menu_FilePopup");
    if (popup) {
      popup.removeEventListener ("popupshowing",this.onFileMenuShowing);
    }
    
    if ("UnMHTMenu" in this._window) {
      delete this._window.UnMHTMenu;
    }
  },
  
  /* ==== private ==== */
  
  _frameMenuItems : null, /* Array  追加したフレームのメニューアイテム */
  
  
  /* メニューを表示するか */
  _enableContextMenuSave : false,      /* Boolean  MHT で保存 */
  _enableContextMenuQuickSave : false, /* Boolean  QuickSave */
  _enableContextMenuSendmail : false,  /* Boolean  メール送信 */
  _enableMenuAllCache : false,         /* Boolean  キャッシュを表示
                                        *          /キャッシュをクリア */
  _enableMenuAllExtract : false,       /* Boolean  展開して表示/ソースを表示 */
  
  /**
   * フレームのメニューを追加する
   * フレームのメニューの menupopup は id を持たないので手動で追加する
   */
  _insertFrameMenuItem : function () {
    this._frameMenuItems = [];
    let frameMenu = this.el ("frame");
    let unmhtMenu = this.el ("unmht-frame-context-menu");
    if (frameMenu && frameMenu.firstChild) {
      let framePopup = frameMenu.firstChild;
      for (let node = unmhtMenu.firstChild; node; node = node.nextSibling) {
        let newNode = node.cloneNode (true);
        framePopup.appendChild (newNode);
        this._frameMenuItems.push (newNode);
      }
    }
  },
  
  /**
   * メニューを設定する
   *
   * @param  String id
   *         メニューの種類
   * @param  Boolean isContext
   *         コンテキストメニューか
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   *         null の場合現在開いているドキュメント
   */
  _setMenuItem : function (id, isContext, targetDocument) {
    if (!targetDocument) {
      targetDocument = UnMHTWindowUtil.getCurrentDocument (this._window);
    }
    
    let savable
    = UnMHTSave.isSavable (this._window, targetDocument);
    
    if (id == "command") {
      let enableCommand = function (type, enabled) {
        let command = this.el ("unmht-" + id + "-" + type);
        if (command) {
          command.setAttribute ("disabled", !enabled);
        }
      }.bind (this);
      
      enableCommand ("save", savable);
      enableCommand ("sendmail", savable);
      return;
    }
    
    let showMenu = function (type, shown) {
      let menuitem = this.el ("unmht-" + id + "-menuitem-" + type);
      if (menuitem) {
        menuitem.hidden = !shown;
      }
      
      return shown;
    }.bind (this);
    
    showMenu ("cache-show",      this._enableMenuAllCache);
    showMenu ("cache-clear",     this._enableMenuAllCache);
    showMenu ("separator-cache", this._enableMenuAllCache);
    
    showMenu ("extract",           this._enableMenuAllExtract);
    showMenu ("unextract",         this._enableMenuAllExtract);
    showMenu ("separator-extract", this._enableMenuAllExtract);
    
    let onPage = isContext ? UnMHTDOMUtil.onPage (this._window) : true;
    let isContentSelected
    = UnMHTDOMUtil.isContentSelected (this._window, isContext);
    let onSavableLink
    = isContext ? UnMHTDOMUtil.onSavableLink (this._window) : false;
    
    let shown = [
      showMenu ("save",
                savable
                && (!isContext
                    || (onPage && this._enableContextMenuSave))),
      showMenu ("save-selection",
                savable
                && isContentSelected
                && (!isContext
                    || (onPage && this._enableContextMenuSave))),
      showMenu ("save-link",
                onSavableLink
                && (!isContext
                    || this._enableContextMenuSave)),
      showMenu ("quicksave",
                savable
                && (!isContext
                    || (onPage && this._enableContextMenuQuickSave))),
      showMenu ("sendmail",
                savable
                && (!isContext
                    || (onPage && this._enableContextMenuSendmail)))
      ];
    
    showMenu ("separator", shown.some (function (b) b));
  }
};
UnMHTPrototype.createInheritance (UnMHTMenuClass, UnMHTWindowHook);
