/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTAppInfo.jsm");
XPCOMUtils.importRelative (this, "UnMHTTimer.jsm");

let EXPORTED_SYMBOLS = ["UnMHTOverlayLoader"];

/**
 * XUL オーバーレイの適用と解除
 */
let UnMHTOverlayLoader = {
  /**
   * XUL オーバーレイをロードする
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   * @param  String xulName
   *         xul のパス
   * @param  Function callback
   *         ロード完了後のコールバック
   */
  loadInto : function (window, xulName, callback) {
    let browser = window.document.createElement ("browser");
    
    let f = function () {
      if (browser.contentDocument.location.href != xulName) {
        return;
      }
      
      browser.removeEventListener ("load", f, true);
      
      this._loadFromBrowser (window, xulName, browser);
      
      browser.parentNode.removeChild (browser);
      
      callback ();
    }.bind (this);
    
    browser.setAttribute ("src", xulName);
    browser.setAttribute ("collaped", "true");
    browser.setAttribute ("disablehistory", "true");
    browser.addEventListener ("load", f, true);
    
    window.document.documentElement.appendChild (browser);
  },
  
  /**
   * XUL オーバーレイをアンロードする
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   * @param  String xulName
   *         xul のパス
   */
  unloadFrom : function (window, xulName) {
    this._windows
    = this._windows.filter (function ([w, n, insertedNodes]) {
        if (w == window && n == xulName) {
          insertedNodes.forEach (function (node) {
              node.parentNode.removeChild (node);
            });
          return false;
        }
        return true;
      });
  },
  
  /* ==== private ==== */
  
  _windows : [], /* Array  適用したウィンドウと追加した要素
                  *   [[Window, String URI, [nsIDOMXULElement, ...]]] ...] */
  
  /**
   * XUL オーバーレイをブラウザからロードする
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   * @param  String xulName
   *         xul のパス
   * @param  nsIDOMXULElement browser
   *         XUL を読み込んだブラウザ
   */
  _loadFromBrowser : function (window, xulName, browser) {
    let insertedNodes = [];
    
    let node = browser.contentDocument.firstChild;
    while (node) {
      let nextSibling = node.nextSibling;
      if (node instanceof window.XMLStylesheetProcessingInstruction) {
        window.document.insertBefore (node, window.document.documentElement);
        insertedNodes.push (node);
      }
      
      node = nextSibling;
    }
    
    for (let node = browser.contentDocument.documentElement.firstChild;
         node; node = node.nextSibling) {
      if (node.nodeName == "script") {
        /* script 要素を使用するとメモリリークするので使用禁止 */
        Cu.reportError ("script is not supported: " + node.id);
        continue;
      }
      
      if (node.id == "BrowserToolbarPalette"
          || node.id == "MailToolbarPalette") {
        while (node.firstChild) {
          let currentNode = node.firstChild;
          
          this._loadIntoToolbar (window, node.id, currentNode);
          insertedNodes.push (currentNode);
        }
        continue;
      }
      
      let targetNode = window.document.getElementById (node.id);
      if (!targetNode
          && node.id == "menu_ToolsPopup") {
        /* SeaMonkey */
        targetNode = window.document.getElementById ("taskPopup");
      }
      if (!targetNode) {
        Cu.reportError ("not found: " + node.id);
        continue;
      }
      
      while (node.firstChild) {
        let currentNode = node.firstChild;
        
        let before = null;
        if (currentNode.hasAttribute ("insertbefore")) {
          [before]
            = currentNode.getAttribute ("insertbefore").split (",")
            .map (function (id) [id, targetNode.querySelector ("#" + id)])
            .filter (function ([id, n]) n && n.parentNode == targetNode)
            .map (function ([id, n]) n);
        }
        else if (currentNode.hasAttribute ("insertafter")) {
          let [after]
          = currentNode.getAttribute ("insertafter").split (",")
          .map (function (id) [id, targetNode.querySelector ("#" + id)])
          .filter (function ([id, n]) n && n.parentNode == targetNode)
          .map (function ([id, n]) n);
          before = after ? after.nextSibling : null;
        }
        
        targetNode.insertBefore (currentNode, before);
        insertedNodes.push (currentNode);
      }
    }
    
    this._windows.push ([window, xulName, insertedNodes]);
  },
  
  /**
   * ツールバーの項目をロードする
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   * @param  String id
   *         親要素の ID
   * @param  nsIDOMXULElement node
   *         対象の項目の要素
   */
  _loadIntoToolbar : function (window, id, node) {
    if (id == "BrowserToolbarPalette") {
      id = "navigator-toolbox";
    }
    else if (id == "MailToolbarPalette") {
      id = "mail-toolbox";
    }
    
    window.document.getElementById (id).palette.appendChild (node);
    
    let toolbars = window.document.getElementsByTagName ("toolbar");
    
    Array
    .map (toolbars,
          function (toolbar)  {
            let currentset = toolbar.getAttribute ("currentset").split (",");
            return [toolbar, currentset, currentset.indexOf (node.id)];
          })
    .filter (function ([toolbar, currentset, pos]) pos != -1)
    .slice (0, 1)
    .forEach (function ([toolbar, currentset, pos]) {
        if (pos < currentset.length - 1) {
          /* bootstrap なアドオンはツールバーの要素を動的に追加する事になるので
           * 実行の順番次第ではこの時点で次の要素が追加されていない場合がある
           * (同じアドオンの別の要素も含む)
           * ==== FIXME: いつまで待てば良いかは自明ではない ====
           * 解決できるなら早い方が良いので
           * 2**n [ms] おきに n = [0 .. 10] の間 (= 2 秒まで) */
          let f = function (n, timeout) {
            let before = null;
            
            let beforeId = currentset [pos + 1]
            if (beforeId == "separator"
                || beforeId == "spring"
                || beforeId == "spacer") {
              /* 複数出現可能な id を持たない要素は出現位置から取得する */
              let beforePos
              = currentset.slice (0, pos)
              .filter (function (n) n == beforeId)
              .length;
              
              let nodes = toolbar.getElementsByTagName ("toolbar" + beforeId);
              if (beforePos < nodes.length) {
                before = nodes [beforePos];
              }
            }
            else {
              before = toolbar.querySelector ("#" + beforeId);
            }
            
            if (!before) {
              if (n <= 10) {
                UnMHTTimer.setTimeout (function () {
                    f (n + 1, timeout * 2);
                  }, timeout);
              }
              else {
                toolbar.insertItem (node.id);
              }
              return;
            }
            
            toolbar.insertItem (node.id, before);
          };
          f (0, 1);
        }
        else {
          toolbar.insertItem (node.id);
        }
      });
  }
};
