/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTAppInfo.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTUnicodeConverter.jsm");
XPCOMUtils.importRelative (this, "UnMHTVersion.jsm");

let EXPORTED_SYMBOLS = ["UnMHTPref"];

/* import がループしないためにメールの引数はここで宣言する */
let _thunderbirdArg
= "-remote \"xfeDoCommand(composeMessage,subject=\'%TITLE%\',attachment=\'%URI%\')\"";
let _seamonkeyArg
= "-remote \"xfeDoCommand(composeMessage,subject=\'%TITLE_NOSP%\',attachment=\'%URI%\')\"";
let _macmailArg
= "\"%TITLE%\" \"%FILE%\"";

let defaultMail = UnMHTAppInfo.isSeaMonkey () ? "seamonkey" : "thunderbird";

/**
 * UTF8 との変換
 * prefwindow で unichar の設定は UTF8 で格納される
 */
let toUTF8 = function (text) {
  return UnMHTUnicodeConverter.convertFromUnicode (text, "utf-8");
};
let fromUTF8 = function (text) {
  return UnMHTUnicodeConverter.convertToUnicode (text, "utf-8");
};

/**
 * 設定操作と既定値
 */
let UnMHTPref = {
  /* nsIPrefBranch  pref サービス */
  prefBranch : (Cc ["@mozilla.org/preferences-service;1"]
                .getService (Ci.nsIPrefBranch)),
  
  thunderbirdArg : _thunderbirdArg,
  seamonkeyArg   : _seamonkeyArg,
  macmailArg     : _macmailArg,
  
  prefix : "extensions.unmht.",
  
  /**
   * 設定名一覧を返す
   */
  names : function () {
    return Object.keys (this._defPrefs);
  },
  
  /**
   * 設定値の型を返す
   *
   * @param  String name
   *         設定名
   * @return String
   *         設定値の型
   * @throws "UNSUPPORTED_PREFERENCE"
   *         未知の設定名
   */
  typeof : function (name) {
    return (typeof this.getDefPref (name));
  },
  
  /**
   * 既定値を返す
   *
   * @param  String name
   *         設定名
   * @return String
   *         設定値の型
   * @throws "UNSUPPORTED_PREFERENCE"
   *         未知の設定名
   */
  getDefPref : function (name) {
    if (!(name in this._defPrefs)) {
      throw "UNSUPPORTED_PREFERENCE: " + name;
    }
    
    return this._defPrefs [name];
  },
  
  /**
   * 設定を監視する
   *
   * @param String domain
   *        対象のドメイン
   * @param Function func
   *        設定が変更された時に実行する関数
   * @param Any owner
   *        登録者
   */
  addListener : function (domain, func, owner) {
    let observer = {
      observe : function (subject, topic, data) {
        if (topic == "nsPref:changed") {
          func ();
        }
      }
    }
    
    this.prefBranch.addObserver (domain, observer, false);
    
    this._observers.push ([domain, func, owner, observer]);
  },
  
  /**
   * 設定が変更されたイベントのリスナを削除する
   *
   * @param String domain
   *        対象のドメイン
   * @param Function func
   *        設定が変更された時に実行する関数
   */
  removeListener : function (domain, func) {
    this._observers = this._observers.filter (function ([d, f, o, ob]) {
        if (d == domain && f == func) {
          this.prefBranch.removeObserver (d, ob);
          return false;
        }
        return true;
      }.bind (this));
  },

  /**
   * 設定が変更されたイベントのリスナを削除する
   *
   * @param Any owner
   *        登録者
   */
  removeAllListeners : function (owner) {
    this._observers = this._observers.filter (function ([d, f, o, ob]) {
        if (o == owner) {
          this.prefBranch.removeObserver (d, ob);
          return false;
        }
        return true;
      }.bind (this));
  },
  
  /**
   * 古い設定からのマイグレート
   */
  migrate : function () {
    if (this.prefBranch.prefHasUserValue ("unmht.version")
        && !this.prefBranch.prefHasUserValue ("extensions.unmht.version")) {
      let prefHas = function (name) {
        return this.prefBranch.prefHasUserValue (name);
      }.bind (this);
      let getBoolPref = function (name) {
        return this.prefBranch.getBoolPref (name);
      }.bind (this);
      let getCharPref = function (name) {
        return unescape (this.prefBranch.getCharPref (name));
      }.bind (this);
      let getIntPref = function (name) {
        return this.prefBranch.getIntPref (name);
      }.bind (this);
      let setBoolPref = function (name, value) {
        this.prefBranch.setBoolPref (name, value);
      }.bind (this);
      let setCharPref = function (name, value) {
        this.prefBranch.setCharPref (name, toUTF8 (value));
      }.bind (this);
      let setIntPref = function (name, value) {
        this.prefBranch.setIntPref (name, value);
      }.bind (this);

      this.migrate_5_7_14_to_6_0_0 (prefHas,
                                    getBoolPref, getCharPref, getIntPref,
                                    setBoolPref, setCharPref, setIntPref);
      this.prefBranch.deleteBranch ("unmht");
    }
    
    if (this.prefBranch.prefHasUserValue
        ("extensions.unmht.save.single.type")) {
      let prefHas = function (name) {
        return this.prefBranch.prefHasUserValue (name);
      }.bind (this);
      let getBoolPref = function (name) {
        return this.prefBranch.getBoolPref (name);
      }.bind (this);
      let getCharPref = function (name) {
        return fromUTF8 (this.prefBranch.getCharPref (name));
      }.bind (this);
      let getIntPref = function (name) {
        return this.prefBranch.getIntPref (name);
      }.bind (this);
      let setBoolPref = function (name, value) {
        this.prefBranch.setBoolPref (name, value);
      }.bind (this);
      let setCharPref = function (name, value) {
        this.prefBranch.setCharPref (name, toUTF8 (value));
      }.bind (this);
      let setIntPref = function (name, value) {
        this.prefBranch.setIntPref (name, value);
      }.bind (this);
      
      this.migrate_6_0_0_to_6_1_0 (prefHas,
                                   getBoolPref, getCharPref, getIntPref,
                                   setBoolPref, setCharPref, setIntPref);
      this.prefBranch.clearUserPref ("extensions.unmht.save.single.type");
    }
  },
  
  /**
   * 設定値を読み込む
   *
   * @param  String name
   *         設定名
   * @return Boolean/String/Number
   *         設定値
   *         無ければ既定値を返す
   * @throws "UNSUPPORTED_PREFERENCE"
   *         未知の設定名
   */
  getPref : function (name) {
    let pname = this.prefix + name;
    switch (this.typeof (name)) {
      case "boolean": {
        return this.prefBranch.getBoolPref (pname);
      }
      case "string": {
        return fromUTF8 (this.prefBranch.getCharPref (pname));
      }
      case "number": {
        return this.prefBranch.getIntPref  (pname);
      }
    }
    
    return null;
  },
  
  /**
   * 設定値を保存する
   *
   * @param  String name
   *         設定名
   * @param  Boolean/String/Number value
   *         設定値
   * @throws "UNSUPPORTED_PREFERENCE"
   *         未知の設定名
   */
  setPref : function (name, value) {
    let pname = this.prefix + name;
    switch (this.typeof (name)) {
      case "boolean": {
        return this.prefBranch.setBoolPref (pname, value);
      }
      case "string": {
        return this.prefBranch.setCharPref (pname, toUTF8 (value));
      }
      case "number": {
        return this.prefBranch.setIntPref  (pname, value);
      }
    }
    
    return null;
  },
  
  /**
   * ファイルからの設定の読み込みを開始する
   *
   * @param  nsIFile file
   *         読み込む対象のファイル
   *         null にすると既定値を書いたファイル扱い
   * @return UnMHTPrefFileInputObject
   *         設定入力オブジェクト
   *         ファイルが不正な場合は null
   *           実体は Object
   *           <String 設定名, Boolean/String/Number 設定値>
   */
  initPrefFromFile : function (file) {
    let map = {};
    
    if (file == null) {
      return map;
    }
    
    let text = UnMHTFileUtil.readData (file).replace (/\r\n/g, "\n");
    
    let ok = false;
    
    text.replace
    (/extensions\.unmht\.([^\n]+),([^\n]+)\n?/g,
     function (matched, name, value) {
      name = unescape (name);
      value = unescape (value);
       
      if (name in this._defPrefs) {
        switch (this.typeof (name)) {
          case "boolean": map [name] = (value == "true");    break;
          case "string":  map [name] = unescape (value);     break;
          case "number":  map [name] = parseInt (value, 10); break;
        }
        ok = true;
      }
         
      return "";
    }.bind (this));
    
    if (!ok) {
      let oldMap = {};
      
      text.replace
        (/(unmht\.[^\n]+),([^\n]+)\n?/g,
         function (matched, name, value) {
          name = unescape (name);
          value = unescape (value);
           
          oldMap [name] = value;
          ok = true;
           
          return "";
        }.bind (this));
      
      if (ok) {
        let prefHas = function (name) {
          return (name in oldMap);
        }.bind (this);
        let getBoolPref = function (name) {
          return (oldMap [name] == "true");
        }.bind (this);
        let getCharPref = function (name) {
          return unescape (oldMap [name]);
        }.bind (this);
        let getIntPref = function (name) {
          return parseInt (oldMap [name], 10);
        }.bind (this);
        let setBoolPref = function (name, value) {
          name = name.replace (/^extensions\.unmht\./, "");
          map [name] = value;
        }.bind (this);
        let setCharPref = function (name, value) {
          name = name.replace (/^extensions\.unmht\./, "");
          map [name] = value;
        }.bind (this);
        let setIntPref = function (name, value) {
          name = name.replace (/^extensions\.unmht\./, "");
          map [name] = value;
        }.bind (this);
        
        this.migrate_5_7_14_to_6_0_0 (prefHas,
                                      getBoolPref, getCharPref, getIntPref,
                                      setBoolPref, setCharPref, setIntPref);
      }
    }
    
    if ("save.single.type" in map) {
      let oldMap = map;
      map = {};
      
      let prefHas = function (name) {
        return (name in oldMap);
      }.bind (this);
      let getBoolPref = function (name) {
        return oldMap [name];
      }.bind (this);
      let getCharPref = function (name) {
        return oldMap [name];
      }.bind (this);
      let getIntPref = function (name) {
        return oldMap [name];
      }.bind (this);
      let setBoolPref = function (name, value) {
        map [name] = value;
      }.bind (this);
      let setCharPref = function (name, value) {
        map [name] = value;
      }.bind (this);
      let setIntPref = function (name, value) {
        map [name] = value;
      }.bind (this);
      this.migrate_6_0_0_to_6_1_0 (hrefHas,
                                   getBoolPref, getCharPref, getIntPref,
                                   setBoolPref, setCharPref, setIntPref);
    }
    
    if (!ok) {
      return null;
    }
    
    return map;
  },
  
  /**
   * ファイルからの設定の読み込みを終了する
   *
   * @param  UnMHTPrefFileInputObject input
   *         設定入力オブジェクト
   */
  closePrefFromFile : function (input) {
    /* 特にやる事はない */
  },
  
  /**
   * ファイルから設定値を読み込む
   *
   * @param  UnMHTPrefFileInputObject input
   *         設定入力オブジェクト
   * @param  String name
   *         設定名
   * @return Boolean/String/Number
   *         設定値
   *         無ければ既定値を返す
   * @throws "UNSUPPORTED_PREFERENCE"
   *         未知の設定名
   */
  getPrefFromFile : function (input, name) {
    let defPref = this.getDefPref (name);
    
    if (name in input) {
      return input [name];
    }
    
    return defPref;
  },
  
  /**
   * ファイルへの設定の保存を開始する
   *
   * @param  nsIFile file
   *         保存対象のファイル
   * @return UnMHTPrefFileOutputObject
   *         設定出力オブジェクト
   *           実体は nsIFileOutputStream
   *           保存先のファイル出力ストリーム
   */
  initPrefToFile : function (file) {
    return UnMHTFileUtil.openOutputStream (file);
  },
  
  /**
   * ファイルへの設定の保存を終了する
   *
   * @param  UnMHTPrefFileOutputObject output
   *         設定出力オブジェクト
   */
  closePrefToFile : function (output) {
    output.close ();
  },
  
  /**
   * 設定を保存する
   * 
   * @param  UnMHTPrefFileOutputObject output
   *         設定出力オブジェクト
   * @param  String name
   *         設定名
   * @param  Boolean/String/Number value
   *         既定値
   */
  setPrefToFile : function (output, name, value) {
    if (name in this._defPrefs) {
      switch (this.typeof (name)) {
        case "string": value = escape (value); break;
      }
    }
    let line = escape (this.prefix + name) + "," + escape (value) + "\r\n";
    output.write (line, line.length);
  },
  
  /* ==== private ==== */
  
  /* Object  既定値
   *   <String 設定名, Boolean/String/Number 既定値>  */
  _defPrefs : {
    /* General */
    "version" : UnMHTVersion,
    
    "open.multiple.action" : "tabs", /* tabs/index/first */
    
    "open.ppt.skip_warning" : false,
    
    "open.offline" : false,
    
    "statusbar.icon" : false,
    
    /* Save */
    "save.contents.video" : "ask", /* save/ignore/ask */
    "save.contents.embed" : "ask", /* save/ignore/ask */

    "save.header.mail.original_field" : false,
    
    "save.insert.link" : false,
    "save.insert.link.pos" : "bottom", /* top/bottom */
    "save.insert.link.date" : true,
    "save.insert.link.version" : true,
    "save.insert.link.nolocal" : false,
    
    "save.history.register" : true,
    "save.progress.statusbar" : false,
    "save.show_errors" : true,
    "save.show_option_dialog" : false,
    
    "save.local.same_dir" : false,
    "save.local.same_dir.append" : "",
    
    "save.single.state" : "current", /* current/original */
    "save.single.add" : false,
    
    /* Multiple */
    "save.multi.state" : "current", /* current/original */
    "save.multi.format" : "separated", /* separated/single
                                        * /single_noindex */
    
    "save.multi.dir" : "",
    
    "save.multi.dup.action" : "rename", /* replace/rename/ask */
    
    "save.multi.autoclose" : false,
    "save.multi.tabs.autoclose" : false,
    "save.multi.check_none" : false,

    "save.multi.index.style" : false,
    "save.multi.index.style.content" : "",
    
    /* QuickSave */
    "save.quick.state" : "current", /* current/original */
    
    "save.quick.dir" : "",
    
    "save.quick.dup.action" : "rename", /* replace/rename/ask */
    
    "save.quick.use_selection" : false,
    
    /* Mail */
    "save.mail.state" : "current", /* current/original */
    
    "save.mail.app.type" : defaultMail, /* thunderbird/seamonkey
                                         * /macmail/other */
    "save.mail.app" : "",
    
    "save.mail.arg.type" : defaultMail, /* thunderbird/seamonkey
                                         * /macmail/other */
    "save.mail.arg.other" : toUTF8 (_thunderbirdArg),

    "save.mail.zip" : false,
    
    /* Filename */
    "save.filename" : "title", /* leafname/title/custom */
    
    "save.filename.dup.rename.digits" : 1,
    "save.filename.dup.rename.init" : 2,
    
    "save.filename.custom" : "%FILENAME%",
    "save.filename.custom.am" : "a.m.",
    "save.filename.custom.pm" : "p.m.",
    "save.filename.custom.limit" : false,
    "save.filename.custom.limit.count" : 100,
    "save.filename.custom.mail_date" : false,
    
    /* Menu */
    "menu.context.save" : true,
    "menu.context.save.quick" : true,
    "menu.context.save.mail" : true,
    "menu.tool" : true,
    
    /* Keyboard */
    "shortcut.save.single" : false,
    "shortcut.save.single.keycode" : "VK_S",
    "shortcut.save.single.mods.alt" : true,
    "shortcut.save.single.mods.ctrl" : true,
    "shortcut.save.single.mods.meta" : false,
    "shortcut.save.single.mods.shift" : false,
    
    "shortcut.save.multi" : false,
    "shortcut.save.multi.keycode" : "VK_A",
    "shortcut.save.multi.mods.alt" : true,
    "shortcut.save.multi.mods.ctrl" : true,
    "shortcut.save.multi.mods.meta" : false,
    "shortcut.save.multi.mods.shift" : false,
    
    "shortcut.save.quick" : false,
    "shortcut.save.quick.keycode" : "VK_Q",
    "shortcut.save.quick.mods.alt" : true,
    "shortcut.save.quick.mods.ctrl" : true,
    "shortcut.save.quick.mods.meta" : false,
    "shortcut.save.quick.mods.shift" : false,
    
    /* Convert */
    "convert.dir.suffix" : "_files",
    
    /* Advanced */
    "open.max_run_time" : true,
    "open.max_run_time.time" : 0,
    
    "save.max_run_time" : true,
    "save.max_run_time.time" : 0,
    
    "menu.all.extract" : false,
    "menu.all.cache" : false,
    
    "open.cache.expire" : true,
    "open.cache.expire.time" : 20,
    
    "save.insert.url" : false,
    "save.deep_import.action" : "ask", /* save/ignore/ask */
    
    "save.mail.debug" : false,
    
    "use_global_lastdir" : true,
    
    /* Other */
    
    "open.lastdir" : "",
    "save.lastdir" : ""
  },
  
  _observers : [], /* Array 設定が変更されたイベントのオブザーバ
                    *   [[String ドメイン, Function 実行する関数, Any 登録者
                    *     nsIObserver オブザーバ], ...] */
  
  /**
   * 5.7.14 から 6.0.0 へのマイグレート
   */
  migrate_5_7_14_to_6_0_0 : function (prefHas,
                                      getBoolPref, getCharPref, getIntPref,
                                      setBoolPref, setCharPref, setIntPref) {
    let oldPrefix = "unmht.";
    let newPrefix = this.prefix;
    
    let migrates = [
      ["open.multiple.action", "open_multiple"],
      ["open.ppt.skip_warning", "skip_ppt_warning"],
      ["open.offline", "filter.load.offline"],
      ["statusbar.icon", "statusbar.preferences"],
      
      ["save.insert.link", "save.insertlink"],
      ["save.insert.link.pos", "save.insertlink.pos"],
      ["save.insert.link.date", "save.insertlink.date"],
      ["save.insert.link.version", "save.insertlink.version"],
      ["save.insert.link.nolocal", "save.insertlink.nolocal"],
      ["save.insert.url", "save.inserturl"],
      
      ["save.history.register", "save.register_download"],
      ["save.progress.statusbar", "save.statusbar"],
      ["save.show_errors", "save.show_errors"],
      ["save.show_option_dialog", "save.show_option_dialog"],
      
      ["save.local.same_dir", "save.local_to_same"],
      
      ["save.single.type", "save.type"],
      
      ["save.multi.type", "savemulti.type"],
      ["save.multi.format", "savemulti.lastformat"],
      ["save.multi.dir", "savemulti.base"],
      ["save.multi.autoclose", "save.multi.autoclose"],
      ["save.multi.check_none", "save.multi.none"],
      
      ["save.quick.type", "quicksave.savetype"],
      ["save.quick.dir", "quicksave.base"],
      ["save.quick.dup.action", "quicksave.autoreplace.action"],
      ["save.quick.use_selection", "quicksave.selection"],
      
      ["save.mail.type", "sendmail.savetype"],
      ["save.mail.app", "sendmail.app"],
      ["save.mail.arg.other", "sendmail.arg"],
      
      ["save.filename", "save.filename"],
      
      ["save.filename.dup.rename.digits", "save.dup.digits"],
      ["save.filename.dup.rename.init", "save.dup.init"],
      
      ["save.filename.custom", "save.filename.custom"],
      ["save.filename.custom.am", "save.filename.custom.am"],
      ["save.filename.custom.pm", "save.filename.custom.pm"],
      ["save.filename.custom.limit", "save.filename.custom.limit"],
      ["save.filename.custom.limit.count", "save.filename.custom.limit.count"],
      
      ["menu.context.save", "save.contextmenu"],
      ["menu.context.save.quick", "quicksave.showmenu"],
      ["menu.context.save.mail", "sendmail.showmenu"],
      ["menu.tool", "toolmenu.preferences"],
      
      ["shortcut.save.single", "save.shortcut"],
      ["shortcut.save.single.keycode", "save.shortcut.keycode"],
      ["shortcut.save.single.mods.alt", "save.shortcut.modifiers.alt"],
      ["shortcut.save.single.mods.ctrl", "save.shortcut.modifiers.ctrl"],
      ["shortcut.save.single.mods.meta", "save.shortcut.modifiers.meta"],
      ["shortcut.save.single.mods.shift", "save.shortcut.modifiers.shift"],
      
      ["shortcut.save.multi", "save.shortcut2"],
      ["shortcut.save.multi.keycode", "save.shortcut2.keycode"],
      ["shortcut.save.multi.mods.alt", "save.shortcut2.modifiers.alt"],
      ["shortcut.save.multi.mods.ctrl", "save.shortcut2.modifiers.ctrl"],
      ["shortcut.save.multi.mods.meta", "save.shortcut2.modifiers.meta"],
      ["shortcut.save.multi.mods.shift", "save.shortcut2.modifiers.shift"],
      
      ["shortcut.save.quick", "quicksave.shortcut"],
      ["shortcut.save.quick.keycode", "quicksave.shortcut.keycode"],
      ["shortcut.save.quick.mods.alt", "quicksave.shortcut.modifiers.alt"],
      ["shortcut.save.quick.mods.ctrl", "quicksave.shortcut.modifiers.ctrl"],
      ["shortcut.save.quick.mods.meta", "quicksave.shortcut.modifiers.meta"],
      ["shortcut.save.quick.mods.shift", "quicksave.shortcut.modifiers.shift"],
      
      ["convert.dir.suffix", "convert.dir_suffix"],
      
      ["open.max_run_time", "nolimit"],
      ["open.max_run_time.time", "nolimit.time"],
      ["save.max_run_time", "save.nolimit"],
      ["save.max_run_time.time", "save.nolimit.time"],
      
      ["open.cache.expire", "cache_expire.enabled"],
      ["open.cache.expire.time", "cache_expire"],
      
      ["save.mail.debug", "sendmail.debug"],
      
      ["use_global_lastdir", "use_global_lastdir"],
      
      ["open.lastdir", "save.lastdir"],
      ["save.lastdir", "save.lastdir"]
      ];
    
    migrates.forEach (function ([newName, oldName]) {
        if (prefHas (oldPrefix + oldName)) {
          switch (this.typeof (newName)) {
            case "boolean": {
              let value = getBoolPref (oldPrefix + oldName);
              setBoolPref (newPrefix + newName, value);
              break;
            }
            case "string": {
              let value = getCharPref (oldPrefix + oldName);
              setCharPref (newPrefix + newName, value);
              break;
            }
            case "number": {
              let value = getIntPref (oldPrefix + oldName);
              setIntPref (newPrefix + newName, value);
              break;
            }
          }
        }
      }.bind (this));
    
    let name, value;
    
    name = "sendmail.argtype";
    if (prefHas (oldPrefix + name)) {
      value = getCharPref (oldPrefix + name);
      if (value == "index") {
        value = "other";
      }
      
      name = "save.mail.arg.type";
      setCharPref (newPrefix + name, value);
    }
    
    name = "sendmail.app";
    if (prefHas (oldPrefix + name)) {
      value = getCharPref (oldPrefix + name);
      if (value != "") {
        value = "other";
      }
      else {
        value = defailtMail;
      }
      name = "save.mail.app.type";
      
      setCharPref (newPrefix + name, value);
    }
    
    name = "save.skip_alert_on_deep";
    if (prefHas (oldPrefix + name)) {
      value = getBoolPref (oldPrefix + name);
      if (!value) {
        value = "ask";
      }
      else {
        name = "save.skip_alert_on_deep.default";
        if (prefHas (oldPrefix + name)) {
          value = getCharPref (oldPrefix + name);
        }
        else {
          value = "";
        }
      }
      
      if (value) {
        name = "save.deep_import.action";
        setCharPref (newPrefix + name, value);
      }
    }
    
    name = "save.multi.autoreplace";
    if (prefHas (oldPrefix + name)) {
      value = getBoolPref (oldPrefix + name);
      if (!value) {
        value = "ask";
      }
      else {
        name = "save.multi.autoreplace.action";
        if (prefHas (oldPrefix + name)) {
          value = getCharPref (oldPrefix + name);
        }
        else {
          value = "";
        }
      }
      
      if (value) {
        name = "save.multi.dup.action";
        setCharPref (newPrefix + name, value);
      }
    }
    
    name = "save.local_to_same.append";
    if (prefHas (oldPrefix + name)) {
      value = getBoolPref (oldPrefix + name);
      if (!value) {
        value = "";
      }
      else {
        name = "save.local_to_same.append.text"
        if (prefHas (oldPrefix + name)) {
          value = getCharPref (oldPrefix + name);
        }
        else {
          value = "";
        }
      }
      
      name = "save.local.same_dir.append";
      setCharPref (newPrefix + name, value);
    }
  },
  
  /**
   * 6.0.0 から 6.1.0 へのマイグレート
   */
  migrate_6_0_0_to_6_1_0 : function (prefHas,
                                     getBoolPref, getCharPref, getIntPref,
                                     setBoolPref, setCharPref, setIntPref) {
    let oldPrefix = this.prefix;
    let newPrefix = this.prefix;
    
    let migrates = [
      ["save.multi.state", "save.multi.type"],
      ["save.quick.state", "save.quick.type"],
      ["save.mail.state",  "save.mail.type"]
      ];
    
    migrates.forEach (function ([newName, oldName]) {
        if (prefHas (oldPrefix + oldName)) {
          switch (this.typeof (newName)) {
            case "boolean": {
              let value = getBoolPref (oldPrefix + oldName);
              setBoolPref (newPrefix + newName, value);
              break;
            }
            case "string": {
              let value = getCharPref (oldPrefix + oldName);
              setCharPref (newPrefix + newName, unescape (value));
              break;
            }
            case "number": {
              let value = getIntPref (oldPrefix + oldName);
              setIntPref (newPrefix + newName, value);
              break;
            }
          }
        }
      }.bind (this));
    
    let name, value;
    
    name = "save.single.type";
    if (prefHas (oldPrefix + name)) {
      let saveType = getCharPref (oldPrefix + name);
      
      let value = saveType.replace (/_add$/, "");
      
      name = "save.single.state";
      setCharPref (newPrefix + name, value);
      
      name = "save.single.add";
      
      if (saveType.match (/_add$/)) {
        value = true;
      }
      else {
        value = false;
      }
      setBoolPref (newPrefix + name, value);
    }
  },
  
  /**
   * デフォルト設定を作成する
   */
  _createDefaultBranch : function () {
    let prefService
    = Cc ["@mozilla.org/preferences-service;1"].getService (Ci.nsIPrefService);
    let branch = prefService.getDefaultBranch ("");
    for (let [name, value] in Iterator (this._defPrefs)) {
      if (name == "version") {
        continue;
      }
      switch (typeof value) {
        case "boolean": {
          branch.setBoolPref (this.prefix + name, value);
          break;
        }
        case "string": {
          branch.setCharPref (this.prefix + name, toUTF8 (value));
          break;
        }
        case "number": {
          branch.setIntPref (this.prefix + name, value);
          break;
        }
      }
    }
  }
};
UnMHTPref.migrate ();
UnMHTPref.setPref ("version", UnMHTVersion);
UnMHTPref._createDefaultBranch ();
