/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTAppInfo.jsm");
XPCOMUtils.importRelative (this, "UnMHTArrayUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTCustomFilename.jsm");
XPCOMUtils.importRelative (this, "UnMHTDOMUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileConverter.jsm");
XPCOMUtils.importRelative (this, "UnMHTFilePicker.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTLocaleUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTMail.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative (this, "UnMHTTimer.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTVersion.jsm");

let EXPORTED_SYMBOLS = ["UnMHTPrefWindowClass"];

/**
 * HTML ファイルから MHT への変換のイベントリスナ
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 * @implements IUnMHTFileConverterListener
 */
function UnMHTPrefWindowFileConverterListener (window) {
  this._window = window;
}
UnMHTPrefWindowFileConverterListener.prototype = {
  /**
   * 破棄
   */
  destruct : function () {
    this._window = null;
  },
  
  /* ==== IUnMHTFileConverterListener インターフェース ==== */
  
  /**
   * ファイル変換開始イベント
   *
   * @param  UnMHTFileConvertContext
   *         変換コンテキスト
   * @param  nsIFile file
   *         対象のファイル
   */
  onStartFile : function (context, file) {
    let pre = UnMHTStringBundle.get ("convert_progress_pre");
    let post = UnMHTStringBundle.get ("convert_progress_post");
    
    let message
    = (pre ? (pre + " ") : "") + file.leafName + (post ? (" " + post) : "");
    
    this.el ("convert_progress").value = message;
    this.el ("convert_progressbar").value = Math.floor (context.progress * 100);
    this.el ("convert_progressbar").style.visibility = "visible";
  },
  
  /**
   * 全ファイル変換完了イベント
   *
   * @param  UnMHTFileConvertContext
   *         変換コンテキスト
   * @param  Number count
   *         変換したファイル数
   */
  onComplete : function (context, count) {
    let pre = UnMHTStringBundle.get ("convert_result_pre");
    let post = UnMHTStringBundle.get ("convert_result_post");
    
    let message
    = (pre ? (pre + " ") : "") + count + (post ? (" " + post) : "");
    
    this.el ("convert_progress").value = message;
    this.el ("convert_progressbar").value = 100;
    this.el ("convert_progressbar").style.visibility = "hidden";
    
    this.destruct ();
  },
  
  /* ==== private ==== */

  _window : null, /* nsIDOMWindow  対象のウィンドウ */
  
  /**
   * getElementById の省略
   *
   * @param  String id
   *         要素の ID
   * @return nsIDOMElement
   *         id に対応する要素
   *         見付からなければ null
   */
  el : function (id) {
    return this._window.document.getElementById (id);
  }
};

/**
 * 設定ウィンドウ
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 */
function UnMHTPrefWindowClass (window) {
  this._window = window;
  this._dragService
    = Cc ["@mozilla.org/widget/dragservice;1"].getService (Ci.nsIDragService);
}
UnMHTPrefWindowClass.prototype = {
  /**
   * 破棄
   */
  destruct : function () {
    this._window = null;
    this._dragService = null;
  },
  
  /* ---- UI 関係イベント ---- */
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    this._disableItems ();
    
    this._updateUI ();
    
    this._window.sizeToContent ();
  },

  /**
   * アンロードイベント
   */
  onUnload : function () {
  },
  
  /**
   * 各パネルがロードされたイベント
   */
  onPaneLoad : function () {
    this._disableItems ();
    
    this._updateUI ();
  },
  
  /**
   * ヘルブボタンがクリックされたイベント
   */
  onHelp : function (event) {
    this._openURI
    ((UnMHTLocaleUtil.isJapanene ()
      ? "http://www.unmht.org/unmht/usage.html"
      : "http://www.unmht.org/unmht/en_usage.html")
     + "#" + this.el ("unmht_pref_window").currentPane.helpTopic);
  },
  
  /**
   * サイトへのリンクがクリックされたイベント
   * 現在未使用
   */
  onClickWebsiteLink : function () {
    this._openURI
    (UnMHTLocaleUtil.isJapanene ()
     ? "http://www.unmht.org/unmht/"
     : "http://www.unmht.org/unmht/en_index.html");
  },
  
  /**
   * ショートカットキー入力欄でキーを押したイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onShortcutKeyDown : function (event, id) {
    let elem = function (type) {
      return this.el ("ui-shortcut.save." + id + "." + type);
    }.bind (this);
    
    let keycode = "";
    for (let attribute in Ci.nsIDOMKeyEvent) {
      let re = attribute.match (/^DOM_(VK_.+)$/);
      if (!re) {
        continue;
      }
      let code = re [1];
      if (event.keyCode == Ci.nsIDOMKeyEvent [attribute]) {
        keycode = code;
      }
    }
    if (keycode == "VK_SHIFT"
        || keycode == "VK_CONTROL"
        || keycode == "VK_ALT"
        || keycode == "VK_META") {
      return;
    }
    
    if (keycode != "") {
      elem ("keycode").value = keycode;
      elem ("mods.alt").checked = event.altKey;
      elem ("mods.ctrl").checked = event.ctrlKey;
      elem ("mods.meta").checked = event.metaKey;
      elem ("mods.shift").checked = event.shiftKey;
        
      this._updateValue (elem ("keycode"));
      this._updateValue (elem ("mods.alt"));
      this._updateValue (elem ("mods.ctrl"));
      this._updateValue (elem ("mods.meta"));
      this._updateValue (elem ("mods.shift"));
    }
        
    event.preventDefault ();
    event.stopPropagation ();
  },
  
  /**
   * メール送信アプリケーションの変更ボタンを押したイベント
   */
  onClickSelectSendMailApp : function () {
    let node = this.el ("ui-save.mail.app");
    
    let file
    = UnMHTFilePicker.open
    (this._window,
     UnMHTStringBundle.get ("sendmail_app_title"),
     Ci.nsIFilePicker.modeOpen,
     false, UnMHTFileUtil.initFile (node.value), null, null);
    
    if (!file) {
      return;
    }
    
    node.value = file.path;
    this._updateValue (node);
  },
  
  /**
   * メール送信アプリケーションの引数を変更したイベント
   */
  onChangeSendMailArgType : function () {
    let arg = this.el ("ui-save.mail.arg.other");
    let value = "";
    
    switch (this.el ("ui-save.mail.arg.type").value) {
      case "thunderbird":
        value = UnMHTMail.thunderbirdArg;
        arg.disabled = true;
        break;
      case "seamonkey":
        value = UnMHTMail.seamonkeyArg;
        arg.disabled = true;
        break;
      case "macmail":
        value = UnMHTMail.macmailArg;
        arg.disabled = true;
        break;
      default:
        value = arg.value;
        arg.disabled = false;
    }
    
    arg.value = value;
    this._updateValue (arg);
  },
  
  /**
   * ディレクトリの変更ボタンを押したイベント
   *
   * @param  String id
   *         対象の設定
   */
  onClickSelectBase : function (id) {
    let node = this.el ("ui-save." + id + ".dir");
    
    let file
    = UnMHTFilePicker.open
    (this._window,
     UnMHTStringBundle.get ("quicksave_base_title"),
     Ci.nsIFilePicker.modeGetFolder,
     false, UnMHTFileUtil.initFile (node.value), null, null);
    
    if (!file) {
      return;
    }
    
    node.value = file.path;
    this._updateValue (node);
  },
  
  /**
   * カスタムファイル名のサンプルの更新が必要になったイベント
   */
  onNeedUpdateCustomFilename : function () {
    this.el ("save_filename_custom_sample").value
    = UnMHTCustomFilename.formatWith
    ("no title",
     "index.html",
     "http://localhost/index.html?query#fragment",
     new Date (),
     this.el ("ui-save.filename.custom").value,
     this.el ("ui-save.filename.custom.am").value,
     this.el ("ui-save.filename.custom.pm").value);
  },
  
  /**
   * HTML ファイルを MHT に変換するボタンを押したイベント
   */
  onClickConvertFiles : function () {
    let files
    = UnMHTFilePicker.open
    (this._window,
     UnMHTStringBundle.get ("convert_title"),
     Ci.nsIFilePicker.modeOpenMultiple,
     true, null, null, null);
    
    if (!files) {
      return;
    }
    
    UnMHTFileConverter.createContext
    (files.filter (function (f) f.isFile ()),
     this.el ("ui-convert.dir.suffix").value,
     new UnMHTPrefWindowFileConverterListener (this._window))
    .start ();
  },
  
  /**
   * ディレクトリ内の HTML ファイルを MHT に変換するボタンを押したイベント
   */
  onClickConvertDir : function () {
    let dir
    = UnMHTFilePicker.open
    (this._window,
     UnMHTStringBundle.get ("convert_title"),
     Ci.nsIFilePicker.modeGetFolder,
     true, null, null, null);
    
    if (!dir) {
      return;
    }
    
    UnMHTFileConverter.createContext
    (UnMHTFileUtil.getDirectoryEntries (dir).filter (function (f) f.isFile ()),
     this.el ("ui-convert.dir.suffix").value,
     new UnMHTPrefWindowFileConverterListener (this._window))
    .start ();
  },
  
  /**
   * 変換パネルにファイルがドラッグされたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onDragOverConvertFiles : function (event) {
    if (this._dragService.getCurrentSession ()
        .isDataFlavorSupported ("application/x-moz-file")) {
      event.stopPropagation ();
      event.preventDefault ();
    }
  },
  
  /**
   * 変換パネルにファイルがドロップされたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onDragDropConvertFiles : function (event) {
    let dataTransfer = event.dataTransfer;
    
    let files = [];
    for (let i in UnMHTArrayUtil.range (0, dataTransfer.mozItemCount)) {
      let types = dataTransfer.mozTypesAt (i);
      if (Array.indexOf (types, "application/x-moz-file") == -1) {
        continue;
      }
      
      let data = dataTransfer.mozGetDataAt ("application/x-moz-file", i);
      try {
        let file = data.QueryInterface (Ci.nsIFile);
        
        if (file.isDirectory ()) {
          files
            = files.concat
            (UnMHTFileUtil.getDirectoryEntries (file)
             .filter (function (f) f.isFile ()));
        }
        else if (file.isFile ()) {
          files.push (file);
        }
      }
      catch (e) {
      }
    }
    
    UnMHTFileConverter.createContext
    (files,
     this.el ("ui-convert.dir.suffix").value,
     new UnMHTPrefWindowFileConverterListener (this._window))
    .start ();
  },
  
  /**
   * ディレクトリ名の末尾を初期化するボタンを押したイベント
   */
  onClickInitConvertDirSuffix : function () {
    let node = this.el ("ui-convert.dir.suffix");
    node.value = "_files";
    this._updateValue (node);
  },
  
  /**
   * 設定のインポートボタンを押したイベント
   */
  onClickImportPrefs : function () {
    let setStatus = function (type) {
      this.el ("import_export_message").value
      = this.el ("import_" + type).value;
    }.bind (this);
    
    let modFunc = function (filePicker) {
      filePicker.appendFilter ("Text", "*.txt");
      filePicker.defaultString = "unmhtConfig.txt";
      filePicker.appendFilters (Ci.nsIFilePicker.filterAll);
    };
    let file
    = UnMHTFilePicker.open
    (this._window,
     this.el ("import_title").value,
     Ci.nsIFilePicker.modeOpen,
     false, null, modFunc, null);
    
    if (!file) {
      setStatus ("abort");
      return;
    }
    
    let input = UnMHTPref.initPrefFromFile (file);
    if (input) {
      this._showAllPanesAnd (function () {
          try {
            this._loadPrefs (input);
            UnMHTPref.closePrefFromFile (input);
            setStatus ("done");
          }
          catch (e) {
            setStatus ("fail");
          }
        }.bind (this));
    }
    else {
      setStatus ("illegal");
    }
  },
  
  /**
   * 設定のエクスポートボタンを押したイベント
   */
  onClickExportPrefs : function () {
    let setStatus = function (type) {
      this.el ("import_export_message").value
      = this.el ("export_" + type).value;
    }.bind (this);
    
    let modFunc = function (filePicker) {
      filePicker.appendFilter ("Text", "*.txt");
      filePicker.defaultString = "unmhtConfig.txt";
      filePicker.appendFilters (Ci.nsIFilePicker.filterAll);
    };
    let file
    = UnMHTFilePicker.open
    (this._window,
     this.el ("export_title").value,
     Ci.nsIFilePicker.modeSave,
     false, null, modFunc, null);
    
    if (!file) {
      setStatus ("abort");
      return;
    }
     
    let output = UnMHTPref.initPrefToFile (file);
    if (output) {
      this._showAllPanesAnd (function () {
          try {
            this._savePrefs (output);
            UnMHTPref.closePrefToFile (output);
            setStatus ("done");
          }
          catch (e) {
            setStatus ("fail");
          }
        }.bind (this));
    }
    else {
      setStatus ("fail");
    }
  },
  
  /**
   * 設定の初期化ボタンを押したイベント
   */
  onClickInitPrefs : function () {
    let setStatus = function (type) {
      this.el ("import_export_message").value
      = this.el ("reset_" + type).value;
    }.bind (this);
    
    let input = UnMHTPref.initPrefFromFile (null);
    this._showAllPanesAnd (function () {
        this._loadPrefs (input);
        UnMHTPref.closePrefFromFile (input);
        setStatus ("done");
      }.bind (this));
  },
  
  /**
   * カスタムファイル名に変数を挿入する
   *
   * @param  String text
   *         挿入するテキスト
   */
  insertCustom : function (text) {
    let textbox = this.el ("ui-save.filename.custom");
    let value = textbox.value;
    let start = textbox.selectionStart;
    let end = textbox.selectionEnd;
    
    textbox.value = value.substr (0, start) + text + value.substr (end);
    this._updateValue (textbox);
        
    textbox.focus ();
    
    textbox.selectionStart = textbox.selectionEnd = start + text.length;
    
    this.onNeedUpdateCustomFilename ();
  },
  
  /**
   * スピンボタンで数値を上げる
   *
   * @param  String id
   *         対象の要素の ID
   * @param  Number max
   *         数値の最大値
   */
  onUpSpin : function (id, max) {
    let node = this.el (id);
    
    let n = parseInt (node.value, 10) + 1;
    if (n > max) {
      n = max;
    }
    
    node.value = n;
    this._updateValue (node);
  },
  
  /**
   * スピンボタンで数値を下げる
   *
   * @param  String id
   *         対象の要素の ID
   * @param  Number max
   *         数値の最小値
   */
  onDownSpin : function (id, min) {
    let node = this.el (id);
    
    let n = parseInt (node.value, 10) - 1;
    if (n < min) {
      n = min;
    }
    
    node.value = n;
    this._updateValue (node);
  },
  
  /* ---- UI イベント - 依存関係 ---- */
  
  onChangeSaveLocalSameDir : function () {
    this._setDisabled
    (this.el ("box_save_local_same_dir"),
     !this.el ("ui-save.local.same_dir").checked);
  },
  onChangeSaveInsertLink : function () {
    this._setDisabled
    (this.el ("box_save_insert_link_1"),
     !this.el ("ui-save.insert.link").checked);
    this._setDisabled
    (this.el ("box_save_insert_link_2"),
     !this.el ("ui-save.insert.link").checked);
  },
  onChangeSaveMultiIndexStyle : function () {
    this._setDisabled
    (this.el ("ui-save.multi.index.style.content"),
     !this.el ("ui-save.multi.index.style").checked);
  },
  onChangeSaveMailAppType : function () {
    this._setDisabled
    (this.el ("box_save_mail_app"),
     this.el ("ui-save.mail.app.type").value != "other");
    
    if (this.el ("ui-save.mail.app.type").value != "other") {
      let argtype = this.el ("ui-save.mail.arg.type");
      
      argtype.value = this.el ("ui-save.mail.app.type").value;
      this._updateValue (argtype);
      
      this.onChangeSendMailArgType ();
    }
  },
  onChangeShortcut : function (id) {
    this._setDisabled
    (this.el ("box_shortcut_" + id.replace (/\./g, "_")),
     !this.el ("ui-shortcut." + id).checked);
    this._setDisabled
    (this.el ("field_shortcut_" + id.replace (/\./g, "_")),
     !this.el ("ui-shortcut." + id).checked);
  },
  onChangeOpenMaxRuntime : function () {
    this._setDisabled
    (this.el ("subbox_open_max_run_time"),
     !this.el ("ui-open.max_run_time").checked);
  },
  onChangeSaveMaxRuntime : function () {
    this._setDisabled
    (this.el ("subbox_save_max_run_time"),
     !this.el ("ui-save.max_run_time").checked);
  },
  onChangeOpenCacheExpire : function () {
    this._setDisabled
    (this.el ("subbox_open_cache_expire"),
     !this.el ("ui-open.cache.expire").checked);
  },
  
  /* ==== private ==== */
  
  _window : null, /* nsIDOMWindow  対象のウィンドウ */
  _dragService: null, /* nsIDragService  ドラッグ管理 */
  
  /**
   * 設定を読み込む
   *
   * @param  UnMHTPrefFileInputObject input
   *         設定入力オブジェクト
   */
  _loadPrefs : function (input) {
    UnMHTPref.names ().forEach
    (function (name) {
      let node = this.el ("ui-" + name);
      if (!node) {
        return;
      }
      let value = UnMHTPref.getPrefFromFile (input, name);
      
      switch (typeof value) {
        case "boolean": node.checked = value; break;
        case "string":  node.value   = value; break;
        case "number":  node.value   = value; break;
      }
      this._updateValue (node);
    }.bind (this));
    
    this._updateUI ();
  },
  
  /**
   * 全ての設定を保存する
   *
   * @param  UnMHTPrefFileOutputObject
   *         設定出力オブジェクト
   */
  _savePrefs : function (output) {
    UnMHTPref.names ().forEach
    (function (name) {
      let node = this.el ("ui-" + name);
      if (!node) {
        return;
      }
      
      let value;
      switch (UnMHTPref.typeof (name)) {
        case "boolean": value = node.checked;              break;
        case "string":  value = node.value;                break;
        case "number":  value = parseInt (node.value, 10); break;
      }
      UnMHTPref.setPrefToFile (output, name, value);
    }.bind (this));
  },
  
  /**
   * 全てのパネルを表示してから関数を実行する
   * パネルは選択されるまでロードされないので
   * インポート、エクスポート、初期化時に必要
   *
   * @param  Function func
   *         実行する関数
   */
  _showAllPanesAnd : function (func) {
    let panes = this._window.document.getElementsByTagName ("prefpane");
    let defaultPane = this.el ("pane_general");
    let prefWindow = this.el ("unmht_pref_window");
    let i = 0;
    
    let f = function () {
      if (i == panes.length) {
        prefWindow.showPane (defaultPane);
        func ();
        return;
      }
      
      if (!panes [i].loaded) {
        prefWindow.showPane (panes [i]);
      }
      else {
        i ++;
      }
      
      UnMHTTimer.setTimeout (f, 100);
    }.bind (this);
    
    f ();
  },
  
  /**
   * チェックボックスの状態等を更新する
   */
  _updateUI : function () {
    if (this.el ("pane_save").loaded) {
      this.onChangeSaveLocalSameDir ();
      this.onChangeSaveInsertLink ();
    }
    if (this.el ("pane_savemulti").loaded) {
      this.onChangeSaveMultiIndexStyle ();
    }
    if (this.el ("pane_mail").loaded) {
      this.onChangeSaveMailAppType ();
      this.onChangeSendMailArgType ();
    }
    if (this.el ("pane_filename").loaded) {
      this.onNeedUpdateCustomFilename ();
    }
    if (this.el ("pane_keyboard").loaded) {
      let runtime
      = Cc ["@mozilla.org/xre/app-info;1"].getService (Ci.nsIXULRuntime);
      if (runtime.OS == "Darwin") {
        this.el ("ui-shortcut.save.single.mods.alt").label
          = this.el ("ui-shortcut.save.multi.mods.alt").label
          = this.el ("ui-shortcut.save.quick.mods.alt").label
          = this.el ("keyboard_mods_option").value;
          
        this.el ("ui-shortcut.save.single.mods.meta").label
          = this.el ("ui-shortcut.save.multi.mods.meta").label
          = this.el ("ui-shortcut.save.quick.mods.meta").label
          = this.el ("keyboard_mods_command").value;
      }
      
      this.onChangeShortcut ("save.single");
      this.onChangeShortcut ("save.multi");
      this.onChangeShortcut ("save.quick");
    }
    if (this.el ("pane_advanced").loaded) {
      this.onChangeOpenMaxRuntime ();
      this.onChangeSaveMaxRuntime ();
      this.onChangeOpenCacheExpire ();
    }
  },
  
  /**
   * ノードの状態を変更する
   *
   * @param  nsIDOMXULElement node
   *         対象のノード
   * @param  Boolean disabled
   *         無効か
   * @param  Boolean uncheck
   *         チェックを外すか
   */
  _setDisabled : function (node, disabled, uncheck) {
    try {
      node.disabled = disabled;
    }
    catch (e) {
    }
    
    if (uncheck) {
      try {
        node.checked = false;
      }
      catch (e) {
      }
    }
    
    if ("childNodes" in node) {
      Array.forEach (node.childNodes,
                     function (node) {
                       this._setDisabled (node, disabled, uncheck);
                     }.bind (this));
    }
  },
  
  /**
   * 環境によって使用できない設定を無効にする
   */
  _disableItems : function () {
    if (UnMHTAppInfo.isThunderbird ()) {
      [
        "group_open",
        
        "box_insert",
        "ui-save.local.same_dir",
        "box_save_local_same_dir",
        
        "group_save_multi",
        "group_save_multi_index",
        "group_save_quick",
        "group_save_mail",
        
        "save_filename_leafname",
        
        "group_menu_context",
        "group_shortcut",
        
        "box_open_max_run_time",
        "group_menu_all",
        "group_cache",
        "group_advanced_mail",
        "group_other"
        ].forEach (function (id) {
            let node = this.el (id);
            if (node) {
              this._setDisabled (node, true, true);
            }
          }.bind (this));
    }
    
    if (UnMHTAppInfo.isFirefox ()) {
      [
        "ui-save.header.mail.original_field",
        "ui-save.filename.custom.mail_date"
        ].forEach (function (id) {
            let node = this.el (id);
            if (node) {
              this._setDisabled (node, true, true);
            }
          }.bind (this));
    }
    
    let runtime
    = Cc ["@mozilla.org/xre/app-info;1"].getService (Ci.nsIXULRuntime);
    if (runtime.OS != "Darwin") {
      [
        "save_mail_app_type_macmail",
        "save_mail_arg_type_macmail"
        ].forEach (function (id) {
            let node = this.el (id);
            if (node) {
              this._setDisabled (node, true, true);
            }
          }.bind (this));
    }
  },
  
  /**
   * UI の値を更新した事を通知する
   *
   * @param  nsIDOMElement node
   *         対象の要素
   */
  _updateValue : function (node) {
    UnMHTDOMUtil.findParentNode (node, "prefpane").userChangedValue (node);
  },
  
  /**
   * ページをタブか新しいウィンドウに開く
   *
   * @param  String uri
   *         アドレス
   */
  _openURI : function (uri) {
    let chromeWindow = Services.wm.getMostRecentWindow ("navigator:browser");
    if (chromeWindow) {
      let tabbrowser = UnMHTWindowUtil.getBrowser (chromeWindow);
      tabbrowser.selectedTab = tabbrowser.addTab (uri);
    }
    else {
      this._window.open (uri);
    }
  },
  
  /**
   * getElementById の省略
   *
   * @param  String id
   *         要素の ID
   * @return nsIDOMElement
   *         id に対応する要素
   *         見付からなければ null
   */
  el : function (id) {
    return this._window.document.getElementById (id);
  }
}
