/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

const Ci = Components.interfaces;
const Cc = Components.classes;
const Cr = Components.results;
const Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTAppInfo.jsm");
XPCOMUtils.importRelative (this, "UnMHTCachePage.jsm");
XPCOMUtils.importRelative (this, "UnMHTExtractor.jsm");
XPCOMUtils.importRelative (this, "UnMHTHTMLUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative (this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative (this, "UnMHTUnicodeConverter.jsm");

let EXPORTED_SYMBOLS = ["UnMHTProtocolHandler"];

/**
 * unmht プロトコルからチャネルを生成する
 *
 * @implements nsIProtocolHandler
 */
function UnMHTProtocolHandler () {
}
UnMHTProtocolHandler.prototype = {
  classID:          Components.ID ("{d605b959-0688-4846-887c-f6658725f7f1}"),
  contractID:       "@mozilla.org/network/protocol;1?name=unmht",
  classDescription: "UnMHT Protocol Handler JS Component",
  QueryInterface:   XPCOMUtils.generateQI ([Ci.nsIProtocolHandler]),
  
  /* ==== nsIProtocolHandler インターフェース ==== */
  
  scheme : "unmht", /* String  プロトコルスキーム */
  defaultPort : -1, /* Number  ポート番号
                     *         ネットワークベースではないので -1 */
    
  /* Number プロトコルの詳細
   *        file と同様のプロトコル */
  protocolFlags: (Ci.nsIProtocolHandler.URI_NOAUTH
                  | Ci.nsIProtocolHandler.URI_LOADABLE_BY_ANYONE),
  
  /**
   * ブラックリストのポートを上書きするか
   *   nsIProtocolHandler.allowPort
   *
   * @param  Number port
   *         ポート番号
   * @param  String scheme
   *         プロトコルスキーム
   * @return Boolean
   *         上書きするか
   */
  allowPort: function (port, scheme) {
    return false;
  },
    
  /**
   * URI を作成する
   *   nsIProtocolHandler.newURI
   *
   * @param  String spec
   *         対象の URI
   * @param  String charset
   *         対象の文字コード
   * @param  nsIURI baseURI
   *         読み込み元の URI
   * @return nsIURI
   *         作成した URI
   */
  newURI : function (spec, charset, baseURI) {
    let uri
    = Cc ["@mozilla.org/network/standard-url;1"].createInstance (Ci.nsIURI);
    
    if (!baseURI) {
      uri.spec = spec;
      return uri;
    }
    
    if (UnMHTScheme.isUnMHTURI (baseURI.spec)) {
      let resolved = UnMHTExtractor.resolveURI (baseURI.spec, spec);
      if (resolved) {
        uri.spec = resolved;
        return uri;
      }
    }
    
    uri.spec = baseURI.resolve (spec);
    return uri;
  },
    
  /**
   * チャネルを作成する
   *   nsIProtocolHandler.newChannel
   *
   * @param  nsIURI uri
   *         対象の URI
   * @return nsIChannel
   *         作成したチャネル
   */
  newChannel : function (uri) {
    let param = UnMHTScheme.getUnMHTURIParam (uri.spec);
    
    try {
      if (!("original" in param
            && param.original)) {
        /* 不正な URI */
        return this._createIllegalChannel (uri);
      }
      else if (param.scheme == "cache") {
        return this._createCacheChannel (uri);
      }
      else if (param.scheme == "cid"
               || param.scheme == "dummy"
               || param.scheme == "empty") {
        return this._createEmptyChannel (uri);
      }
      else if (param.scheme == "error") {
        return this._createNotFoundChannel (uri);
      }
      else {
        /* mht 内のファイル */
        return this._createFileChannel (uri, param);
      }
    }
    catch (e) {
      Cu.reportError (e + "\n" + e.stack);
    }
        
    return this._createEmptyChannel (uri);
  },
  
  /* ==== private ==== */
  
  /**
   * mht 内のファイルのチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI (umht:// 形式)
   * @param  UnMHTURIParam param
   *         対象の URI の情報
   * @return nsIChannel
   *         作成したチャネル
   */
  _createFileChannel : function (uri, param) {
    let [eFileInfo, part] = UnMHTExtractor.getFileInfoAndPart (uri.spec);
    if (!eFileInfo) {
      try {
        /* URI が不正でないかチェック */
        Services.io.newURI (param.original, null, null);
      }
      catch (e) {
        return this._createIllegalChannel (uri);
      }
      
      return this._createEmptyChannel (uri); 
    }
    
    if (!part) {
      return this._createEmptyChannel (uri);
    }
    
    let pipe = Cc ["@mozilla.org/pipe;1"].createInstance (Ci.nsIPipe);
    pipe.init (false, false, part.eParam.content.length, 1, null);
    pipe.outputStream.write
    (part.eParam.content, part.eParam.content.length);
    pipe.outputStream.close ();
    
    return this._createInputStreamChannel (uri,
                                           part.eParam.charset,
                                           part.eParam.mimetype,
                                           pipe.inputStream,
                                           part.eParam.content.length);
  },
    
  /**
   * キャッシュ情報のチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @return nsIChannel
   *         作成したチャネル
   */
  _createCacheChannel : function (uri) {
    let data = UnMHTCachePage.createCachePage ();
    
    return this._createStringInputStreamChannel (uri, "utf-8", "text/html", 
                                                 data, data.length);
  },
  
  /**
   * ファイルが無い情報のチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @return nsIChannel
   *         作成したチャネル
   */
  _createNotFoundChannel : function (uri) {
    let bundle = function (name) {
      return UnMHTHTMLUtil.escapeEntity (UnMHTStringBundle.get (name));
    };
    
    let message
    = bundle ("error_file_not_found") + "<br>"
    + UnMHTHTMLUtil.escapeEntity (UnMHTScheme.deErrorURI (uri.spec));
    return this._createErrorChannel (uri, message);
  },
  
  /**
   * 不正な URI 情報のチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @return nsIChannel
   *         作成したチャネル
   */
  _createIllegalChannel : function (uri) {
    let bundle = function (name) {
      return UnMHTHTMLUtil.escapeEntity (UnMHTStringBundle.get (name));
    };
    
    let message
    = bundle ("error_illegal_address") + "<br>"
    + UnMHTHTMLUtil.escapeEntity (uri.spec);
    return this._createErrorChannel (uri, message);
  },
  
  /**
   * エラーメッセージを表示するチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @param  String message
   *         表示するメッセージ (HTML)
   * @return nsIChannel
   *         作成したチャネル
   */
  _createErrorChannel : function (uri, message) {
    let bundle = function (name) {
      return UnMHTHTMLUtil.escapeEntity (UnMHTStringBundle.get (name));
    };
    
    let data
    = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
    + "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\""
    + "          \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"
    + "<html xmlns=\"http://www.w3.org/1999/xhtml\">"
    + "<head>"
    + "<title>"
    + bundle ("error_title")
    + "</title>"
    + "<style type=\"text/css\">"
    + "h1 {"
    + "color: #ff0000;"
    + "background-color: inherit;"
    + "}"
    + "div.message {"
    + "font-size: 1em;"
    + "color: #ff0000;"
    + "background-color: inherit;"
    + "}"
    + "</style>"
    + "</head>"
    + "<body>"
    + "<h1>"
    + bundle ("error_title")
    + "</h1>"
        
    + "<hr />"
        
    + "<div class=\"message\">"
    + message
    + "</div>"
        
    + "<hr />"
        
    + "</body>"
    + "</html>";
    
    data = UnMHTUnicodeConverter.convertFromUnicode (data, "utf-8");
    
    return this._createStringInputStreamChannel (uri, "utf-8", "text/html", 
                                                 data, data.length);
  },

  /**
   * 空のチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @return nsIChannel
   *         作成したチャネル
   */
  _createEmptyChannel : function (uri) {
    return this._createStringInputStreamChannel (uri, "utf-8", "text/html", 
                                                 "", 0);
  },
  
  /**
   * 文字列を返すチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @param  String charset
   *         文字セット
   * @param  String type
   *         MIME-Type
   * @param  String data
   *         文字列
   * @param  Number length
   *         文字列の長さ
   * @return nsIChannel
   *         作成したチャネル
   */
  _createStringInputStreamChannel : function (uri, charset, type,
                                              data, length) {
    let sstream
    = Cc ["@mozilla.org/io/string-input-stream;1"]
    .createInstance (Ci.nsIStringInputStream);
    sstream.setData (data, length);
    
    return this._createInputStreamChannel (uri, charset, type,
                                           sstream, length);
  },
  
  /**
   * ストリームの内容を返すチャネルを作成する
   *
   * @param  nsIURI uri
   *         対象の URI
   * @param  String charset
   *         文字セット
   * @param  String type
   *         MIME-Type
   * @param  nsIInputStream stream
   *         ストリーム
   * @param  Number length
   *         ストリームの長さ
   * @return nsIChannel
   *         作成したチャネル
   */
  _createInputStreamChannel : function (uri, charset, type, stream, length) {
    let inputStreamChannel
    = Cc ["@mozilla.org/network/input-stream-channel;1"]
    .createInstance (Ci.nsIInputStreamChannel)
    .QueryInterface (Ci.nsIChannel);
    
    inputStreamChannel.setURI (uri);
    inputStreamChannel.contentCharset = charset;
    inputStreamChannel.contentType    = type;
    inputStreamChannel.contentStream  = stream;
    inputStreamChannel.contentLength  = length;
    
    return inputStreamChannel;
  }
};
