/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTDialog.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTPrototype.jsm");
XPCOMUtils.importRelative (this, "UnMHTSave.jsm");
XPCOMUtils.importRelative (this, "UnMHTSaveProgress.jsm");
XPCOMUtils.importRelative (this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative (this, "UnMHTTimer.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveDialogClass"];

/**
 * 保存ダイアログ
 *
 * ウィンドウの引数
 * {
 *   saveState       : Number  保存状態 (必須),
 *
 *   scrapBookData   : Object  ScrapBook のデータ (必須),
 *   targetDocument  : nsIDOMDocument  対象のドキュメント (必須),
 *   targetSelection : nsISelection  選択範囲 (必須),
 *
 *   targetFile      : nsIFile  保存先のファイル (必須),
 *   addURIs         : Array  追加ファイルの URI (必須)
 *                       [String URI, ...],
 *
 *   callback        : Function  終了時コールバック (オプショナル),
 *
 *   window          : nsIDOMWindow  ドキュメントを持つブラウザを
 *                                   実行中のウィンドウ
 *                                   メールの添付ファイル取得に使う
 *                                   (オプショナル),
 * }
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 * @inherits   UnMHTDialog
 * @implements IUnMHTDialog
 * @implements IUnMHTSaveProgressOwner
 */
function UnMHTSaveDialogClass (window) {
  UnMHTDialog.call (this, window);
}
UnMHTSaveDialogClass.prototype = {
  /* ==== IUnMHTDialog インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct : function () {
    this._callback = null;
    this._saveContext = null;
    UnMHTPrototype.super (this, "destruct").call (this);
  },
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    let arg = this._window.arguments [0];
        
    let dialog = this.el ("unmht_save_dialog");
    dialog.setAttribute
    ("title",
     this.el (UnMHTSave.saveStateToString (arg.saveState) + "_label")
     .value);
    
    dialog.getButton ("accept").disabled = true;
    
    let [location, subject]
      = UnMHTSave.getLocationAndTitle (arg.targetDocument, arg.scrapBookData);
    
    if ("callback" in arg) {
      this._callback = arg.callback;
    }
    
    let saveContext
    = UnMHTSave.createContext
    (this._window, arg.targetFile, location, subject,
     new UnMHTSaveProgress (this._window, this));
    
    let arg2 = {
      targetDocument : arg.targetDocument,
      targetSelection : arg.targetSelection,
      scrapBookData : arg.scrapBookData,
      saveState : arg.saveState,
      location : location,
      isHTML : true,
      isStartPart : false,
      window : arg.window
    };
    if (!saveContext.addDocument (arg2)) {
      saveContext.destruct ();
      
      if (this._callback) {
        this._callback (true);
      }
      
      return;
    }
    
    if (arg.addURIs) {
      arg.addURIs.forEach (function (uri) {
          saveContext.addFile (null, uri, false);
        });
    }
    
    this._saveContext = saveContext;
    
    saveContext.start ();
  },
  
  /**
   * ダイアログをキャンセルしたイベント
   *
   * @return Boolean
   *         ダイアログを閉じるか
   */
  onDialogCancel : function (context) {
    if (!this._saveContext) {
      return true;
    }
    if (this._saveContext) {
      this._saveContext.cancel ();
      this._saveContext = null;
    
      return false;
    }
    
    if (this._callback) {
      this._callback (true);
    }
    return true;
  },
  
  /* ==== IUnMHTSaveProgressOwner インターフェース ===== */
  
  /**
   * キャンセルが完了したイベント
   *
   * @param  UnMHTSaveProgress progress
   *         進捗状況表示オブジェクト
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   */
  onCancel : function (progress, context) {
    if (this._callback) {
      this._callback (true);
    }
    this._window.close ();
  },
  
  /**
   * 保存失敗イベント
   *
   * @param  UnMHTSaveProgress progress
   *         進捗状況表示オブジェクト
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Number reason
   *         エラーの理由
   * @param  Exception e
   *         例外
   */
  onError : function (progress, context, reason, e) {
  },
  
  /**
   * 保存完了
   *
   * @param  UnMHTSaveProgress progress
   *         進捗状況表示オブジェクト
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Array ngList
   *         失敗したファイル
   *           [String ファイル名, ...]
   */
  onSave : function (progress, context, ngList) {
    let dialog = this.el ("unmht_save_dialog");
    
    dialog.getButton ("accept").disabled = false;
    dialog.getButton ("cancel").disabled = true;
    
    if (this._callback) {
      this._callback (false);
    }
    
    if (UnMHTPref.getPref ("save.show_errors") && ngList.length > 0) {
      progress.showFail (ngList);
    }
    else {
      let window = this._window;
      UnMHTTimer.setTimeout (function () {
          window.close ();
        }, 1000);
    }
  },
  
  /* ==== private ==== */
  
  _callback : null,   /* Function  終了時コールバック */
  
  _saveContext : null /* UnMHTSaveContext  保存コンテキスト */
};
UnMHTPrototype.createInheritance (UnMHTSaveDialogClass, UnMHTDialog);
