/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTHTMLUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTSave.jsm");
XPCOMUtils.importRelative (this, "UnMHTSaveProgress.jsm");
XPCOMUtils.importRelative (this, "UnMHTUnicodeConverter.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveMulti"];

/**
 * 複数のタブの保存コンテキスト
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 * @param  IUnMHTSaveMultiListener listener
 *         イベントリスナ
 * @implements IUnMHTSaveProgressOwner
 */
function UnMHTSaveMultiContext (window, listener) {
  this._window = window;
  this._listener = listener;
  
  this._targets = [];
  this._addURIs = [];
  this._ngList = [];
  
  this._dupAction = UnMHTPref.getPref ("save.multi.dup.action");
}
UnMHTSaveMultiContext.prototype = {
  /**
   * 破棄
   * start 後は自分で呼ぶ
   */
  destruct : function () {
    this._saveContext = null;
    this._ngList = null;
  },
  
  /**
   * 保存対象を追加する
   *
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  Any id
   *         保存対象の ID
   * @param  Object param
   *         パラメータ
   *           {
   *             filenameBase: String  ファイル名の拡張子以外,
   *             title: String  インデックスページのタイトル,
   *             desc: String  インデックスページの説明,
   *           }
   */
  addTarget : function (targetDocument, id, param) {
    this._targets.push ([targetDocument, id, param]);
  },
  
  /**
   * ファイルを追加する
   *
   * @param  String url
   *         ファイルのアドレス
   */
  addFile : function (url) {
    this._addURIs.push (url);
  },
  
  /* ==== IUnMHTSaveProgressOwner インターフェース ==== */
  
  /**
   * キャンセルが完了したイベント
   *
   * @param  UnMHTSaveProgress progress
   *         進捗状況表示オブジェクト
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   */
  onCancel : function (progress, context) {
    this._saveContext = null;
    if (this._isCanceled) {
      this._listener.onCancel (this);
      this.destruct ();
    }
  },
  
  /**
   * 保存失敗イベント
   *
   * @param  UnMHTSaveProgress progress
   *         進捗状況表示オブジェクト
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Number reason
   *         エラーの理由
   * @param  Exception e
   *         例外
   */
  onError : function (progress, context, reason, e) {
    this._saveContext = null;
    this._listener.onError (context, reason, e);
    this.destruct ();
  },
  
  /**
   * 保存完了
   *
   * @param  UnMHTSaveProgress progress
   *         進捗状況表示オブジェクト
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Array ngList
   *         失敗したファイル
   *           [String ファイル名, ...]
   */
  onSave : function (progress, context, ngList) {
    this._saveContext = null;
    
    this._ngList = this._ngList.concat (ngList);
    
    if (this.saveFormat == UnMHTSaveMulti.FORMAT_SEPARATED) {
      let [, id, ] = this._targets [this._currentIndex];
      this._listener.onStatusChange (this, id, UnMHTSaveMulti.STATUS_OK);
      
      this._saveNext ();
    }
    else {
      this._listener.onSave (this, this._ngList);
      this.destruct ();
    }
  },
  
  /**
   * 保存を開始する
   */
  start : function () {
    if (this.saveFormat == UnMHTSaveMulti.FORMAT_SEPARATED) {
      this._currentIndex = -1;
      this._saveNext ();
    }
    else {
      this._currentIndex = 0;
      this._saveSingle ();
    }
  },
  
  /**
   * 現在のファイルをスキップする
   */
  skip : function () {
    if (this._saveContext) {
      this._saveContext.cancel ();
      this._saveContext = null;
    }
    
    if (this.saveFormat == UnMHTSaveMulti.FORMAT_SEPARATED) {
      if (this._currentIndex < this._targets.length) {
        let [, id, ] = this._targets [this._currentIndex];
        this._listener.onStatusChange (this, id, UnMHTSaveMulti.STATUS_NG);
      }
      
      this._saveNext ();
    }
    else {
      this._listener.onAborted (this);
      this.destruct ();
    }
  },
  
  /**
   * 保存をキャンセルする
   */
  cancel : function () {
    this._isCanceled = true;
    
    if (this._saveContext) {
      this._saveContext.cancel ();
      this._saveContext = null;
    }
    else {
      this._listener.onCancel (this);
      this.destruct ();
    }
  },
  
  /* ==== private ==== */
  
  _window : null,      /* nsIDOMWindow  対象のウィンドウ */
  _listener : null,    /* IUnMHTSaveMultiListener  イベントリスナ */
  
  _targets : null,     /* Array  対象のドキュメントリスト
                        *   [[nsIDOMDocument, Any ID, Object param], ...] */
  _addURIs : null,     /* Array  追加 URI
                        *   [String, ...] */
  
  _dupAction : "",     /* Number  同名のファイルがあった場合にどうするか
                        *  "replace":  上書きする
                        *  "rename":  名前を変える
                        *  "ask":  ダイアログを表示する */
  
  _currentIndex : 0,   /* Number  現在保存中のドキュメントの位置 */
  
  _saveContext : null, /* UnMHTSaveContext  保存コンテキスト */
    
  _ngList : null,      /* Array  失敗したファイル */
  
  _isCanceled : false, /* Boolean  キャンセルされたか */
  
  /**
   * 次のタブを保存する
   * 別々のファイル用
   */
  _saveNext : function () {
    this._currentIndex ++;
    
    if (this._currentIndex >= this._targets.length) {
      this._listener.onSave (this, this._ngList);
      return;
    }
    
    let [targetDocument, id, param] = this._targets [this._currentIndex];
    
    if (!UnMHTSave.isAlive (targetDocument)) {
      this.skip ();
      return;
    }
    
    this._listener.onStatusChange (this, id, UnMHTSaveMulti.STATUS_CURRENT);
    
    let progress = new UnMHTSaveProgress (this._window, this);
    
    let [file, isAborted] = this._getFile (param.filenameBase);
    if (file == null) {
      if (isAborted) {
        this._listener.onAborted (this);
        this.destruct ();
      }
      else {
        this.skip ();
      }
      return;
    }
    
    let location;
    let title;
    
    try {
      location = targetDocument.location.href;
      title = targetDocument.title;
    }
    catch (e) {
      if (UnMHTSave.isAlive (targetDocument)) {
        Cu.reportError (e + "\n" + e.stack);
      }
      this.skip ();
      return;
    }
    
    let saveContext
    = UnMHTSave.createContext (this._window, file, location, title, progress);
    
    let arg = {
      targetDocument : targetDocument,
      saveState : this.saveState,
      location : location,
      isHTML : true,
      isStartPart : true
    };
    
    if (!saveContext.addDocument (arg)) {
      saveContext.destruct (); 
      this.skip ();
      return;
    }
    
    this._saveContext = saveContext;
    
    this._listener.onStart ();
    this._saveContext.start ();
  },
  
  /**
   * 単一のファイルとして保存する
   * 
   */
  _saveSingle : function () {
    let progress = new UnMHTSaveProgress (this._window, this);
    
    let [file,] = this._getFile (this.filenameBase);
    if (file == null) {
      this._listener.onAborted (this);
      this.destruct ();
      return;
    }
    
    let saveContext
    = UnMHTSave.createContext
    (this._window, file, this.location, this.filenameBase, progress);
    
    let locations = {};
    
    let isStartPart = (this.saveFormat == UnMHTSaveMulti.FORMAT_SINGLE_NOINDEX)
    
    let tabs
    = this._targets.map
    (function ([targetDocument, id, param]) {
      if (!UnMHTSave.isAlive (targetDocument)) {
        this._listener.onStatusChange (this, id, UnMHTSaveMulti.STATUS_NG);
        return null;
      }
      
      let savedFrom;
      let isHTML;
      try {
        savedFrom = targetDocument.location.href;
        isHTML = UnMHTSaveMulti.isHTML (targetDocument);
      }
      catch (e) {
        if (UnMHTSave.isAlive (targetDocument)) {
          Cu.reportError (e + "\n" + e.stack);
        }
        
        this._listener.onStatusChange (this, id, UnMHTSaveMulti.STATUS_NG);
        return null;
      }
      
      let locationBase = "";
      let locationExt = "";
      let re = savedFrom.match (/^(.+)(\.[^\/\.]+)$/);
      if (re) {
        [, locationBase, locationExt] = re;
      }
      else {
        locationBase = savedFrom;
      }
        
      let location = locationBase + locationExt;
      
      if (location in locations
          && (this.saveState & UnMHTSave.SAVESTATE_MASK_CURRENT)) {
        /* 同じ URI のタブがあり
         * 現在の状態を保存する場合には別々に保存する */
        let n = 1;
        while (location in locations) {
          n ++;
          location = locationBase + "-" + n + locationExt;
        }
      }
      
      let indexLocation = "";
      if (!(location in locations)) {
        locations [location] = true;
        
        let arg = {
          targetDocument : targetDocument,
          saveState : this.saveState,
          location : location,
          savedFrom : savedFrom,
          isHTML : isHTML,
          isStartPart : isStartPart
        };
        
        let part = saveContext.addDocument (arg);
        if (part) {
          indexLocation = part.sParam.location;
          isStartPart = false;
        }
        else {
          saveContext.destruct ();
          this._listener.onStatusChange (this, id, UnMHTSaveMulti.STATUS_NG);
          return null;
        }
      }
      else {
        indexLocation = location;
      }
      
      let title
      = ("title" in param)
      ? param.title
      : (targetDocument.title || targetDocument.location.href);
      
      let desc
      = ("desc" in param)
      ? param.desc
      : "";
      
      this._listener.onStatusChange (this, id, UnMHTSaveMulti.STATUS_OK);
      
      return [indexLocation, title, desc];
    }.bind (this))
    .filter (function (tab) tab);
    
    if (this.saveFormat == UnMHTSaveMulti.FORMAT_SINGLE) {
      let n = Math.floor (Math.random () * 1000000);
      let indexFilename = "unmht-index-" + n + ".html";
      let cssFilename = "unmht-index-" + n + ".css";
      
      saveContext.addDataDocument
        (UnMHTSaveMulti.createIndexDocument
         (tabs, this._addURIs, cssFilename),
         indexFilename, "UTF-8", "text/html", true);
      
      saveContext.addDataDocument
        (UnMHTSaveMulti.createIndexCSS (true),
         cssFilename, "UTF-8", "text/css", false);
        
      saveContext.setTabs (true);
    }
    
    this._addURIs.forEach (function (uri) {
        saveContext.addFile (null, uri, false);
      }.bind (this));
    
    this._saveContext = saveContext;
    this._listener.onStart ();
    this._saveContext.start ();
  },
  
  /**
   * 保存先のファイルを返す
   *
   * @param  String filenameBase
   *         ファイル名の拡張子以外
   * @return Array
   *           [nsIFile 保存先のファイル,
   *            Boolean 中断されたか]
   */
  _getFile : function (filenameBase) {
    let targetFile = this.baseDir.clone ();
    
    if (targetFile && !targetFile.exists ()) {
      try {
        UnMHTFileUtil.createDirectory (targetFile);
      }
      catch (e) {
        targetFile = null;
      }
    }
    if (!targetFile) {
      this._listener.onError (this, UnMHTSaveMulti.ERROR_ILLEGAL_DIRECTORY);
      
      return [null, true];
    }
    
    targetFile.append (filenameBase + ".mht");
        
    if (targetFile.exists ()) {
      let result;
      if (this._dupAction == "ask") {
        result
          = UnMHTSave.showOverwriteConfirm
          (this._window, targetFile, filenameBase,
           this.saveFormat == UnMHTSaveMulti.FORMAT_SEPARATED);
        
        if (result.result == "skip") {
          return [null, false];
        }
        else if (result.result == "abort" || result.result == "no") {
          return [null, true];
        }
        
        if (result.applyToAll) {
          this._dupAction = result.result;
        }
      }
      else {
        result = {
          result : this._dupAction,
          applyToAll: false,
          filename : ""
        };
        if (result.result == "rename") {
          result.filename
            = UnMHTSave.getUniqueFilename (this.baseDir, filenameBase);
        }
      }
      
      if (result.result == "rename") {
        targetFile = this.baseDir.clone ();
        targetFile.append (result.filename);
      }
    }
    
    return [targetFile, false];
  }
};

/**
 * 複数のタブを MHT で保存
 */
let UnMHTSaveMulti = {
  /* 保存対象の状況 */
  STATUS_NA      : 0x0001,
  STATUS_CURRENT : 0x0002,
  STATUS_OK      : 0x0003,
  STATUS_NG      : 0x0004,
  
  /* 保存形式 */
  FORMAT_SEPARATED      : 0x0001,
  FORMAT_SINGLE         : 0x0011,
  FORMAT_SINGLE_NOINDEX : 0x0012,
  
  /**
   * 保存コンテキストを作成する
   *
   * @param  nsIDOMWindow window
   *         対象のウィンドウ
   * @param  IUnMHTSaveMultiListener listener
   *         イベントリスナ
   * @return UnMHTSaveMultiContext
   *         保存コンテキスト
   */
  createContext : function (window, listener) {
    return new UnMHTSaveMultiContext (window, listener);
  },

  /**
   * インデックスページの内容を生成する
   *
   * @param  Array tabs
   *         タブのリスト
   *           [[String URI, String タイトル, String 説明], ...]
   * @param  Array addURIs
   *         追加 URI のリスト
   *           [String URI, ...]
   * @param  String cssFilename
   *         CSS ファイルのファイル名
   * @return String
   *         インデックスページの内容 (UTF-8)
   */
  createIndexDocument : function (tabs, addURIs, cssFilename) {
    let content
    = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
    + "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\""
    + "          \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"
    + "<html lang=\"ja\" xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"ja\">\n"
    + "<head>\n"
    + "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" />\n"
    + "<title>index</title>\n"
    + "<meta http-equiv=\"Content-Style-Type\" content=\"text/css\" />"
    + "<link rel=\"stylesheet\" type=\"text/css\" href=\""
    + cssFilename
    + "\" />"
    + "</head>\n"
    + "<body>\n"
    + "<ul class=\"unmht-tabs-list\">\n"
    + tabs.map (function ([location, title, desc]) 
                ("<li class=\"unmht-tabs-item\">"
                 + "<a class=\"unmht-tabs\" href=\""
                 + UnMHTHTMLUtil.escapeEntity (location)
                 + "\">"
                 + UnMHTHTMLUtil.escapeEntity (title)
                 + "</a>"
                 + (desc ? ("<div class=\"unmht-tabs-desc\">"
                            + UnMHTHTMLUtil.escapeEntity (desc)
                            + "</div>") : "")
                 + "</li>\n")).join ("")
    + addURIs.map (function (uri)
                   ("<li class=\"unmht-additional-item\">"
                    + "<a class=\"unmht-additional\" href=\""
                    + UnMHTHTMLUtil.escapeEntity (uri)
                    + "\">"
                    + UnMHTHTMLUtil.escapeEntity (uri)
                    + "</a></li>\n")).join ("")
    + "</ul>\n"
    + "</body>\n"
    + "</html>\n";
    
    return UnMHTUnicodeConverter.convertFromUnicode (content, "UTF-8");
  },
  
  /**
   * インデックスページの CSS を生成する
   *
   * @param  Boolean useCustom
   *         カスタマイズを使用するか
   * @return String
   *         インデックスページの CSS (UTF-8)
   */
  createIndexCSS : function (useCustom) {
    let content
    = ".unmht-tabs-desc {\n"
    + "  margin-left: 1em;\n"
    + "}\n";
    
    if (useCustom && UnMHTPref.getPref ("save.multi.index.style")) {
      content += UnMHTPref.getPref ("save.multi.index.style.content");
    }
    
    return UnMHTUnicodeConverter.convertFromUnicode (content, "UTF-8");
  },
  
  /**
   * HTML かどうかを調べる
   *
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @return Boolean
   *         HTML かどうか
   */
  isHTML : function (targetDocument) {
    return (targetDocument.contentType.match (/^text/)
            || targetDocument.contentType.match (/\/x(ht)?ml(\+.+)$/));
  }
};
