/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTArrayUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTDialog.jsm");
XPCOMUtils.importRelative (this, "UnMHTFilePicker.jsm");
XPCOMUtils.importRelative (this, "UnMHTFileUtil.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTPrototype.jsm");
XPCOMUtils.importRelative (this, "UnMHTSave.jsm");
XPCOMUtils.importRelative (this, "UnMHTSaveMulti.jsm");
XPCOMUtils.importRelative (this, "UnMHTSaveProgress.jsm");
XPCOMUtils.importRelative (this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative (this, "UnMHTTimer.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveMultiDialogClass"];

let STATUS_BOX_CONTROL  = 0;
let STATUS_BOX_PROGRESS = 1;
let STATUS_BOX_RESULT   = 2;

/**
 * 複数のタブを MHT で保存ダイアログ
 *
 * ウィンドウの引数
 * {
 *   targetWindow : nsIDOMWindow  現在のウィンドウ (必須)
 * }
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 * @inherits   UnMHTDialog
 * @implements IUnMHTDialog
 * @implements IUnMHTSaveMultiListener
 */
function UnMHTSaveMultiDialogClass (window) {
  UnMHTDialog.call (this, window);
}
UnMHTSaveMultiDialogClass.prototype = {
  /* ---- UI イベント - 一般 ---- */
  
  /**
   * テキストフィールドでキーが押されたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onTextFieldKeyPress : function (event) {
    if (event.keyCode == 13) {
      this.onDialogAccept ();
    }
  },
  
  /**
   * 対象を変更したイベント
   */
  onTargetChange : function () {
    this._updateTargetList ();
  },
  
  /**
   * すべてを選択/解除ボタンがクリックされたイベント
   */
  onCheckAllClick : function () {
    if (!this._targetDocuments.length) {
      return;
    }
    
    let save = !this._isSave (0);
    this._targetDocuments.forEach
    ((function (d, i) this._changeSave (i, save)).bind (this));
    
    this._updateCount ();
    this._updateSingleFilename ();
  },
  
  /**
   * インデックスページのタイトルと説明を変更するのチェックを変更したイベント
   */
  onChangeTitleChange : function () {
    this._updateTargetListState ();
  },
  
  /**
   * ファイルの URI を追加のチェックを変更したイベント
   */
  onAddFilesChange : function () {
    this.el ("add_uris").hidden = !this.el ("use_add_uris").checked;
  },
  
  /**
   * ファイル名のドロップダウンメニューを変更したイベント
   */
  onFilenameTypeChange : function () {
    this._updateFilenameBase ();
  },
  
  /**
   * 保存形式を変更したイベント
   */
  onFormatChange : function () {
    this.el ("changetitle").disabled
    = !(this.el ("saveformat").value == "single");
    
    this.el ("filename_sequential").disabled
    = !(this.el ("saveformat").value == "separated");
    
    this.el ("use_add_uris").disabled
    = !(this.el ("saveformat").value == "single"
        || this.el ("saveformat").value == "single_noindex");
    
    if (this.el ("saveformat").value == "separated") {
      this.el ("use_add_uris").checked = false;
      this.onAddFilesChange ();
    }
    
    this._updateFilenameBase ();
    
    this._updateTargetList ();
  },
  
  /**
   * ディレクトリ選択ボタンがクリックされたイベント
   */
  onBaseDirSelectClick : function () {
    let mode = Ci.nsIFilePicker.modeGetFolder;
    let baseDir = UnMHTFileUtil.init (this.el ("basedir").value);
    
    let dir = UnMHTFilePicker.open (this._window, "", mode, false, baseDir,
                                    null, null);
    if (dir) {
      this.el ("basedir").value = dir.path;
    }
  },
  
  /* ---- UI イベント - 保存制御 ---- */
  
  /**
   * スキップボタンをクリックしたイベント
   */
  onSkipClick : function () {
    if (this._savemultiContext) {
      this._savemultiContext.skip ();
    }
  },
  
  /**
   * 戻るボタンのイベント
   */
  onBackClick : function () {
    let dialog = this.el ("unmht_savemulti_dialog");
    dialog.getButton ("accept").disabled = false;
    dialog.getButton ("cancel").label
    = UnMHTStringBundle.get ("savemulti_cancel");
    
    this.el ("status_box").selectedIndex = STATUS_BOX_CONTROL;
  },
  
  /* ---- UI イベント - 対象リスト ---- */
  
  /**
   * マウスが出たイベント
   * 対象リストのドラッグを終了する
   *
   * @param  Event event
   *         対象のイベント
   */
  onDialogMouseOut : function (event) {
    if (event.explicitOriginalTarget == this.dialog) {
      this._lastDragNode = null;
    }
  },
    
  /**
   * マウスのボタンを開放したイベント
   * 対象リストのドラッグを終了する
   *
   * @param  Event event
   *         対象のイベント
   */
  onDialogMouseUp : function (event) {
    this._lastDragNode = null;
  },
  
  /**
   * 対象リストが変更されたイベント
   * タイトルの変更をチェックする
   *
   * @param  Event event
   *         対象のイベント
   */
  onTargetListChange : function (event) {
    if (this.el ("filename").value == "title") {
      this._updateSingleFilename ();
    }
  },
  
  /**
   * 対象リストでマウスを動かしたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onTargetListMouseMove : function (event) {
    let checkNode = this._getCheckNode (event);
    if (!checkNode) {
      return;
    }
    
    this.onTargetListCheckMouseMove (checkNode);
  },
  
  /**
   * 対象リストでマウスを押下したイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onTargetListMouseDown : function (event) {
    let checkNode = this._getCheckNode (event);
    if (!checkNode) {
      return;
    }
    
    this.onTargetListCheckMouseDown (checkNode);
  },
  
  /**
   * 対象リストで対象のチェック上でマウスを押下したイベント
   * チェックを変更する
   *
   * @param  nsIDOMXULElement checkNode
   *         対象の要素
   */
  onTargetListCheckMouseDown : function (checkNode) {
    let save = this._isSaveNode (checkNode);
    
    this._lastDragType = save;
    this._lastDragNode = checkNode;
    
    this._changeSaveNode (checkNode, !save);
    
    this._updateCount ();
    this._updateSingleFilename ();
  },
  
  /**
   * 対象リストで対象のチェック上でマウスを動かしたイベント
   * ドラッグ時にチェックを変更する
   *
   * @param  nsIDOMXULElement checkNode
   *         対象の要素
   */
  onTargetListCheckMouseMove : function (checkNode) {
    if (this._lastDragNode == null) {
      return;
    }
    
    if (this._lastDragNode == checkNode) {
      return;
    }
    
    this._changeSaveNode (checkNode, !this._lastDragType);
    
    this._updateCount ();
    this._updateSingleFilename ();
  },
  
  /* ---- UI イベント - ファイル名フィールド ---- */
  
  /**
   * ファイル名のフィールドがフォーカスを得たイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onFilenameBaseFocus : function (event) {
    this._lastFilename = this.el ("filename_base").value;
  },
  
  /**
   * ファイル名のフィールドがフォーカスを失なったイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onFilenameBaseBlur : function (event) {
    if (this._lastFilename != this.el ("filename_base").value) {
      this._isFilenameChanged = true;
    }
  },
  
  /* ==== IUnMHTDialog インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct : function () {
    this._targetWindow = null;
    
    this._targetDocuments = null;

    this._lastDragNode = null;
    
    UnMHTPrototype.super (this, "destruct").call (this);
  },
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    let arg = this._window.arguments [0];
        
    this._targetWindow = arg.targetWindow;
    
    this._targetDocuments = [];
    
    let dialog = this.el ("unmht_savemulti_dialog");
    dialog.getButton ("accept").setAttribute ("accesskey", "s");
    dialog.getButton ("accept").label
    = UnMHTStringBundle.get ("savemulti_save");
    dialog.getButton ("cancel").label
    = UnMHTStringBundle.get ("savemulti_cancel");
    
    this.el ("basedir").value
    = UnMHTPref.getPref ("save.multi.dir")
    || UnMHTFileUtil.getDesktopDirectory ().path;
    
    this.el ("target").value = "this";
    this.el ("filename").value = UnMHTPref.getPref ("save.filename");
    this.el ("saveformat").value = UnMHTPref.getPref ("save.multi.format")
    this.el ("savestate").value = UnMHTPref.getPref ("save.multi.state");
    this.el ("use_add_uris").checked = false;
    
    this.onFormatChange ();
    this.onAddFilesChange ();
    this._updateSingleFilename ();
  },
  
  /**
   * ダイアログを OK したイベント
   * 保存を開始する
   *
   * @return Boolean
   *         ダイアログを閉じるか
   */
  onDialogAccept : function () {
    this._startSave ();
    
    return false;
  },
  
  /**
   * ダイアログをキャンセルしたイベント
   * 保存中であればキャンセルする
   *
   * @return Boolean
   *         ダイアログを閉じるか
   */
  onDialogCancel : function () {
    if (this._savemultiContext) {
      this._savemultiContext.cancel ();
      
      let dialog = this.el ("unmht_savemulti_dialog");
      dialog.getButton ("accept").disabled = false;
      
      this.el ("status_box").selectedIndex = STATUS_BOX_CONTROL;
      
      return false;
    }
    
    return true;
  },

  /* ==== IUnMHTSaveMultiListener インターフェース ==== */
  
  /**
   * キャンセルが完了したイベント
   *
   * @param  UnMHTSaveMultiContext context
   *         対象の保存コンテキスト
   */
  onCancel : function (context) {
    this._savemultiContext = null;
  },
  
  /**
   * 保存失敗イベント
   *
   * @param  UnMHTSaveMultiContext context
   *         対象の保存コンテキスト
   */
  onError : function (context, reason, e) {
    switch (reason) {
      case UnMHTSaveMulti.ERROR_ILLEGAL_DIRECTORY: {
        this.failedLabel1.value = UnMHTStringBundle.get ("illegal_directory");
        this.failedLabel1.hidden = false;
        break;
      }
    }
    this.el ("skip").disabled = true;
    this._savemultiContext = null;
  },
  
  /**
   * 実際に保存が開始したイベント
   *
   * @param  UnMHTSaveMultiContext context
   *         対象の保存コンテキスト
   */
  onStart : function (context) {
    this.el ("skip").disabled = false;
  },
  
  /**
   * 保存状況が変わったイベント
   *
   * @param  UnMHTSaveMultiContext context
   *         保存コンテキスト
   * @param  Any id
   *         保存対象の ID
   * @param  Number status
   *         状況 UnMHTSaveMulti.STATUS_*
   */
  onStatusChange : function (context, id, status) {
    switch (status) {
      case UnMHTSaveMulti.STATUS_NA: {
        this._changeStatus (id, "")
        break;
      }
      case UnMHTSaveMulti.STATUS_OK: {
        this._changeStatus (id, "o")
        break;
      }
      case UnMHTSaveMulti.STATUS_CURRENT: {
        this._changeStatus (id, "current")
        break;
      }
      case UnMHTSaveMulti.STATUS_NG: {
        this._changeStatus (id, "x")
        break;
      }
    }
  },
  
  /**
   * 全ファイル保存完了イベント
   *
   * @param  UnMHTSaveMultiContext context
   *         保存コンテキスト
   * @param  Array ngList
   *         失敗したファイル
   *           [String ファイル名, ...]
   */
  onSave : function (context, ngList) {
    this.el ("saved_label").value = UnMHTStringBundle.get ("save_done");
    this._showResult (ngList, true);
    this._savemultiContext = null;
    
    if (!(ngList
          && ngList.length > 0
          && UnMHTPref.getPref ("save.show_errors"))
        && UnMHTPref.getPref ("save.multi.tabs.autoclose")) {
      this._targetDocuments.forEach (function (targetDocument, i) {
          if (!this._getStatus (i)) {
            return;
          }
          
          this._getTargetWindows ().forEach (function (window) {
              let tab
                = UnMHTWindowUtil.getTabForDocument (window, targetDocument);
              if (tab) {
                let tabbrowser = UnMHTWindowUtil.getBrowser (window);
                tabbrowser.removeTab (tab);
              }
            });
        }.bind (this));
    }
  },
  
  /**
   * 保存中断イベント
   *
   * @param  UnMHTSaveMultiContext context
   *         保存コンテキスト
   */
  onAborted : function (context) {
    this.el ("saved_label").value = UnMHTStringBundle.get ("save_aborted");
    this._showResult (null, false);
    this._savemultiContext = null;
  },
  
  /* ==== private ==== */
  
  _targetWindow : null,    /* nsIWindow  現在のウィンドウ */
  _targetDocuments : null, /* Array  対象のドキュメント
                            *   [nsIDOMDocument, ...] */
  
  _lastDragType : false, /* Boolean    ドラッグ開始した項目の選択の状態 */
  _lastDragNode : null,  /* nsIDOMXULElement ドラッグ開始した項目 */
  
  _isFilenameChanged : false, /* Boolean  ファイル名のフィールドが
                               *          変更されたか */
  _lastFilename : "",         /* String  ファイル名のフィールドの
                               *         フォーカス前の値 */
  
  savemultiContext : null,    /* UnMHTSaveMultiContext  保存コンテキスト */
  
  
  /* ---- 保存関係メソッド ---- */
  
  /**
   * 保存を開始する
   *
   * @return Boolean
   *         開始したか
   */
  _startSave : function () {
    if (!this._targetDocuments
        .some ((function (d, i) this._isSave (i) && UnMHTSave.isAlive (d))
               .bind (this))) {
      /* 保存できるものが 1 つも無いので中断 */
      this._targetDocuments
      .filter ((function (d) !UnMHTSave.isAlive (d)).bind (this))
      .forEach ((function (d, i) this._changeStatus (i, "x")).bind (this));
      
      return false;
    }
    
    UnMHTPref.setPref ("save.multi.dir",    this.el ("basedir").value);
    UnMHTPref.setPref ("save.multi.format", this.el ("saveformat").value);
    UnMHTPref.setPref ("save.multi.state",  this.el ("savestate").value);
    
    this.el ("status_box").selectedIndex = STATUS_BOX_PROGRESS;
    
    if (this.el ("saveformat").value == "single"
        || this.el ("saveformat").value == "single_noindex") {
      if (this.el ("filename_base").value == "") {
        this.el ("status_box").selectedIndex = STATUS_BOX_CONTROL;
        this.el ("filename_base").focus ();
      
        return false;
      }
    }
    
    let dialog = this.el ("unmht_savemulti_dialog");
    dialog.getButton ("accept").disabled = true;
    
    let filenameBase
    = UnMHTFileUtil.validateFileName (this.el ("filename_base").value);
    
    let savemultiContext
    = UnMHTSaveMulti.createContext (this._window, this);
    
    let isSingle = false;
    switch (this.el ("saveformat").value) {
      case "separated": {
        savemultiContext.saveFormat = UnMHTSaveMulti.FORMAT_SEPARATED;
        break;
      }
      case "single": {
        savemultiContext.saveFormat = UnMHTSaveMulti.FORMAT_SINGLE;
        isSingle = true;
        break;
      }
      case "single_noindex": {
        savemultiContext.saveFormat = UnMHTSaveMulti.FORMAT_SINGLE_NOINDEX;
        isSingle = true;
        break;
      }
    }
    
    savemultiContext.baseDir
    = UnMHTFileUtil.initFile (this.el ("basedir").value);
    savemultiContext.saveState
    = UnMHTSave.stringToSaveState (this.el ("savestate").value);
    savemultiContext.filenameBase = filenameBase;
    
    if (isSingle) {
      UnMHTSave.parseAddURIs (this.el ("add_uris").value)
        .forEach (function (uri) {
            savemultiContext.addFile (uri);
          });
    }
    
    let isChangeTitle
    = (savemultiContext.saveFormat == UnMHTSaveMulti.FORMAT_SINGLE
       && this.el ("changetitle").checked);
    
    let first = true;
    this._targetDocuments.forEach (function (targetDocument, i) {
        if (!this._isSave (i)) {
          return;
        }
        
        if (!UnMHTSave.isAlive (targetDocument)) {
          return;
        }
        
        let param = {};
        
        if (isSingle) {
          if (first) {
            first = false;
            savemultiContext.location = targetDocument.location.href;
          }
        }
        else {
          if (this.el ("filename").value == "leafname"
              || this.el ("filename").value == "title"
              || this.el ("filename").value == "custom") {
            param.filenameBase = this._getFilename (targetDocument);
          }
          else if (this.el ("filename").value == "sequential") {
            param.filenameBase = filenameBase + "-" + (i + 1);
          }
        }
        
        if (isChangeTitle) {
          param.title = this._getTitle (i);
          param.desc = this._getDesc (i);
        }
        
        savemultiContext.addTarget (targetDocument, i, param);
      }.bind (this));
    
    this._savemultiContext = savemultiContext;
    
    this._savemultiContext.start ();
    
    return true;
  },
  
  /**
   * 保存結果を表示する
   *
   * @param  Array ngList
   *         失敗したファイル
   *           [String ファイル名, ...]
   * @param  Boolean isClosable
   *         閉じてもよいか
   * @return
   */
  _showResult : function (ngList, isClosable) {
    let dialog = this.el ("unmht_savemulti_dialog");
    dialog.getButton ("accept").disabled = true;
    dialog.getButton ("cancel").label
    = UnMHTStringBundle.get ("savemulti_close");
    
    this.el ("status_box").selectedIndex = STATUS_BOX_RESULT;
    
    if (ngList
        && ngList.length > 0
        && UnMHTPref.getPref ("save.show_errors")) {
      new UnMHTSaveProgress (this._window, this).showFail (ngList);
    }
    else if (isClosable && UnMHTPref.getPref ("save.multi.autoclose")) {
      UnMHTTimer.setTimeout (function () {
          this._window.close ();
        }.bind (this), 1000);
    }
  },
  
  /**
   * 保存可能かどうかを調べる
   *
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @param  Boolean asSingle
   *         単一のファイルとして保存するかどうか
   * @return Boolean
   *         保存可能かどうか
   */
  _isSavable : function (targetDocument, asSingle) {
    return asSingle || UnMHTSave.isSavable (null, targetDocument);
  },
  
  /**
   * 保存する際のファイル名を取得する
   *
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @return String
   *         保存する際のファイル名
   */
  _getFilename : function (targetDocument) {
    return UnMHTSave.getFilename
    (targetDocument, null, this.el ("filename").value, new Date ());
  },
  
  /* ---- 対象リスト関連メソッド ---- */
  
  /**
   * 対象リストの選択している数を更新する
   */
  _updateCount : function () {
    this.el ("count_label").value
    = this._targetDocuments
    .filter ((function (d, i) this._isSave (i)).bind (this))
    .length
    + "/" + this._targetDocuments.length;
  },
  
  /**
   * 対象のタブが保存対象かどうかを返す
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @return Boolean
   *         対象のタブが保存対象かどうか
   */
  _isSave : function (index) {
    return this._isSaveNode (this.el ("savecheck_" + index));
  },
  
  /**
   * 対象のタブが保存対象かどうかを返す
   *
   * @param  nsIDOMXULElement node
   *         対象リストのチェックボックス
   * @return Boolean
   *         対象のタブが保存対象かどうか
   */
  _isSaveNode : function (node) {
    return node.getAttribute ("value") == "save";
  },
  
  /**
   * 対象のタブが保存対象かどうかを変更する
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @param  Boolean save
   *         対象のタブが保存対象かどうか
   */
  _changeSave : function (index, save) {
    return this._changeSaveNode (this.el ("savecheck_" + index), save);
  },
  
  /**
   * 対象のタブが保存対象かどうかを変更する
   *
   * @param  nsIDOMXULElement node
   *         対象リストのチェックボックス
   * @param  Boolean save
   *         対象のタブが保存対象かどうか
   */
  _changeSaveNode : function (node, save) {
    node.setAttribute ("value", save ? "save" : "ignore");
    node.setAttribute ("src",
                       save
                       ? "resource://unmht/res/check_v.png"
                       : "resource://unmht/res/check_dot.png");
  },
  
  /**
   * 対象のタブの保存状況を取得する
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @return Boolean status
   *         対象のタブの保存に成功したか
   */
  _getStatus : function (index) {
    return this._getStatusNode (this.el ("savestatus_" + index));
  },
  
  /**
   * 対象のタブの保存状況を取得する
   *
   * @param  nsIDOMXULElement node
   *         対象リストのチェックボックス
   * @return Boolean status
   *         対象のタブの保存に成功したか
   */
  _getStatusNode : function (node) {
    return node.getAttribute ("value") == "saved";
  },
  
  /**
   * 対象のタブの保存状況を変更する
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @param  Boolean status
   *         対象のタブの保存に成功したか
   */
  _changeStatus : function (index, status) {
    return this._changeStatusNode (this.el ("savestatus_" + index), status);
  },
  
  /**
   * 対象のタブの保存状況を変更する
   *
   * @param  nsIDOMXULElement node
   *         対象リストのチェックボックス
   * @param  Boolean status
   *         対象のタブの保存に成功したか
   */
  _changeStatusNode : function (node, status) {
    node.setAttribute ("src",
                       status ?
                       ("resource://unmht/res/check_" + status + ".png")
                       : "");
    node.setAttribute ("value", status ? "saved" : "notsaved");
  },
  
  /**
   * イベントの対象のチェックボックスを取得する
   *
   * @param  Event event
   *         対象のイベント
   * @return nsIDOMXULElement
   *         イベントの対象のチェックボックス
   */
  _getCheckNode : function (event) {
    let row = {}, col = {}, child = {};
    
    this.el ("target_list").treeBoxObject.getCellAt
    (event.clientX, event.clientY, row, col, child);    
    
    if (row.value == -1) {
      return null;
    }
    
    if (col.value == null) {
      return null;
    }
    
    if (col.value.index != 0) {
      return null;
    }
    
    let listbox = this.el ("target_list_children");
    
    return listbox.childNodes [row.value].firstChild.firstChild;
  },
  
  /**
   * 対象リストを空にする
   */
  _clearTargetList : function () {
    let listbox = this.el ("target_list_children");
    while (listbox.firstChild) {
      listbox.removeChild (listbox.firstChild);
    }
  },
  
  /**
   * 対象リストのアイテムを追加する
   * 
   * @param  Number index
   *         対象のタブのインデックス
   * @param  String title
   *         タブのタイトル
   */
  _addTargetListItem : function (index, title) {
    let listbox = this.el ("target_list_children");
    
    let listitem = this._window.document.createElement ("treeitem");
    listbox.appendChild (listitem);
    
    let listrow = this._window.document.createElement ("treerow");
    listitem.appendChild (listrow);
    
    let checkNode = this._window.document.createElement ("treecell");
    listrow.appendChild (checkNode);
    checkNode.id = "savecheck_" + index;
    checkNode.className = "save-iconic";
    checkNode.setAttribute ("editable", "false");
    this._changeSaveNode (checkNode,
                          !UnMHTPref.getPref ("save.multi.check_none"));
    
    let statusNode = this._window.document.createElement ("treecell");
    listrow.appendChild (statusNode);
    statusNode.id = "savestatus_" + index;
    statusNode.setAttribute ("editable", "false");
    statusNode.className = "save-iconic";
    
    let titleNode = this._window.document.createElement ("treecell");
    listrow.appendChild (titleNode);
    titleNode.id = "savetitle_" + index;
    titleNode.setAttribute ("editable", "true");
    titleNode.setAttribute ("label", title);
    titleNode.setAttribute ("value", title);
    titleNode.flex = 1;
    
    let descNode = this._window.document.createElement ("treecell");
    listrow.appendChild (descNode);
    descNode.id = "savedesc_" + index;
    descNode.setAttribute ("editable", "true");
    descNode.flex = 1;
    
    this._updateTargetListItemState (index);
  },
  
  /**
   * 対象リストのアイテムの状態を更新する
   *
   * @param  Number index
   *         対象のアイテムのインデックス
   */
  _updateTargetListItemState : function (index) {
    let titleNode = this.el ("savetitle_" + index);
    let descNode = this.el ("savedesc_" + index);
    
    let isEditable
    = (this.el ("saveformat").value == "single"
       && this.el ("changetitle").checked);
    
    if (!isEditable) {
      titleNode.setAttribute ("label", titleNode.getAttribute ("value"));
    }
    
    titleNode.setAttribute ("editable", isEditable);
    descNode.setAttribute ("editable", isEditable);
    descNode.hidden = !isEditable;
  },

  /**
   * 対象のドキュメントリストを更新する
   */
  _updateTargetList : function () {
    let asSingle
    = (this.el ("saveformat").value == "single"
       || this.el ("saveformat").value == "single_noindex");
    
    this._targetDocuments = [];
    this._getTargetWindows ().forEach (function (window) {
        UnMHTWindowUtil.getDocuments (window).forEach
          (function (targetDocument) {
            if (this._isSavable (targetDocument, asSingle)) {
              this._targetDocuments.push (targetDocument);
            }
          }.bind (this));
      }.bind (this));
    
    this._clearTargetList ();
    this._targetDocuments.forEach (function (targetDocument, i) {
        this._addTargetListItem
          (i, targetDocument.title || targetDocument.location.href);
      }.bind (this));
    
    this._updateTargetListState ();
    this._updateCount ();
  },
  
  /**
   * 対象リストの状態を更新する
   */
  _updateTargetListState : function () {
    this._targetDocuments.forEach
    ((function (d, i) this._updateTargetListItemState (i)).bind (this));
    
    this.el ("target_list_desc").hidden
    = !(this.el ("saveformat").value == "single"
        && this.el ("changetitle").checked);
    
    this._updateSingleFilename ();
  },
  
  /**
   * 対象になるウィンドウのリストを返す
   *
   * @return Array
   *           [Window, ...]
   */
  _getTargetWindows : function () {
    if (this.el ("target").value == "this") {
      return [this._targetWindow];
    }
    else {
      return [window for (window in XPCOMUtils.IterSimpleEnumerator
                            (Services.wm.getEnumerator ("navigator:browser"),
                             Ci.nsIDOMWindow))];
    }
  },
  
  /**
   * 対象のタブのタイトルを取得する
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @return String
   *         対象のタブのタイトル
   */
  _getTitle : function (index) {
    return this.el ("savetitle_" + index).getAttribute ("label");
  },
  
  /**
   * 対象のタブのタイトルを説明する
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @return String
   *         対象のタブの説明
   */
  _getDesc : function (index) {
    return this.el ("savedesc_" + index).getAttribute ("label");
  },
  
  /* ---- ファイル名フィールド関連メソッド ---- */
  
  /**
   * 必要であれば単一のファイルのファイル名を更新する
   */
  _updateSingleFilename : function () {
    if (!(this.el ("saveformat").value == "single"
          || this.el ("saveformat").value == "single_noindex")) {
      return;
    }
    
    if (this._isFilenameChanged && this.el ("filename_base").value != "") {
      return;
    }
    
    let [targetDocument, i]
    = UnMHTArrayUtil.find
    (this._targetDocuments.map (function (d, i) [d, i]),
     (function ([d, i]) this._isSave (i) && UnMHTSave.isAlive (d)).bind (this),
     [null, -1]);
    if (!targetDocument) {
      return;
    }
    
    let filename = "";
    if (this.el ("changetitle").checked
        && this.el ("filename").value == "title") {
      filename = this._getTitle (i);
    }
    else {
      filename = this._getFilename (targetDocument);
    }
    
    this.el ("filename_base").value = filename;
    this._lastFilename = this.el ("filename_base").value;
    this._isFilenameChanged = false;
  },
  
  /**
   * ファイル名フィールドの状態を更新する
   */
  _updateFilenameBase : function () {
    this.el ("filename_base").disabled
    = !(this.el ("filename").value == "sequential"
        || this.el ("saveformat").value == "single"
        || this.el ("saveformat").value == "single_noindex");
    
    if (this.el ("filename").value == "sequential") {
      this.el ("filename_base").setAttribute
        ("placeholder",
         UnMHTStringBundle.get ("savemulti_sequential_placeholder"));
      this.el ("filename_base").value = "";
      this._lastFilename = this.el ("filename_base").value;
      this._isFilenameChanged = false;
    }
    else {
      this.el ("filename_base").setAttribute ("placeholder", "");
    }
    
    this._updateSingleFilename ();
  }
};
UnMHTPrototype.createInheritance (UnMHTSaveMultiDialogClass, UnMHTDialog);
