/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTSave.jsm");
XPCOMUtils.importRelative (this, "UnMHTStringBundle.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveProgress"];

/**
 * 保存の進捗状況表示
 *
 * @param  nsIDOMWindow window
 *         対象のウィンドウ
 * @param  UnMHTSaveProgressOwner owner
 *         オーナー
 * @implements IUnMHTSaveListener
 */
function UnMHTSaveProgress (window, owner) {
  this._window = window;
  this._owner = owner;
  
  this._progressLabel = this.el ("progress_label");
  this._progressLabel.value = UnMHTStringBundle.get ("save_begin");
  this._progressCurrent = this.el ("progress_current");
  this._progressCurrentSize = this.el ("progress_current_size");
  this._progress = this.el ("progress");
  this._progress.value = 0;
    
  this._failedLabel1 = this.el ("failed_label1");
  this._failedLabel2 = this.el ("failed_label2");
  this._failedLabel3 = this.el ("failed_label3");
  this._failedList = this.el ("failed");
  
  this._failedLabel1.hidden = true;
  this._failedLabel2.hidden = true;
  this._failedLabel3.hidden = true;
  this._failedList.hidden = true;
}
UnMHTSaveProgress.prototype = {
  /**
   * 破棄
   */
  destruct : function () {
    this._window = null;
    
    this._progressLabel = null;
    this._progressCurrent = null;
    this._progressCurrentSize = null;
    this._progress = null;
    
    this._failedLabel1 = null;
    this._failedLabel2 = null;
    this._failedLabel3 = null;
    this._failedList = null;
    
    this._owner = null;
  },
  
  /* ==== IUnMHTSaveListener インターフェース ===== */
  
  /**
   * キャンセルが完了したイベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   */
  onCancel : function (context) {
    this._owner.onCancel (this, context)
    this.destruct ();
  },
  
  /**
   * 保存失敗イベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Number reason
   *         エラーの理由
   * @param  Exception e
   *         例外
   */
  onError : function (context, reason, e) {
    switch (reason) {
      case UnMHTSave.ERROR_ADD_DOCUMENT: {
        this._progressLabel.value = UnMHTStringBundle.get ("error_save");
        break;
      }
      case UnMHTSave.ERROR_WRITE_FILE: {
        this._progressLabel.value = UnMHTStringBundle.get ("error_open");
        break;
      }
    }
    
    this._progressCurrent.value = e ? e.toString () : "";
    
    this._owner.onError (this, context, reason, e);
    this.destruct ();
  },

  /**
   * 進捗状況変化イベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  String currentFile
   *         現在取得のファイル名
   * @param  String currentSize
   *         取得したバイト数 / ファイルサイズ
   * @param  Number ok
   *         取得完了した数
   * @param  Number na
   *         取得中の数
   * @param  Number ng
   *         取得失敗した数
   */
  onProgress : function (context, currentFile, currentSize, ok, na, ng) {
    this._progressCurrent.value = currentFile;
    this._progressCurrentSize.value = currentSize;
    
    this._progressLabel.value
    = UnMHTStringBundle.get ("save_begin")
    + " " + (ok + ng) + "/" + (ok + na + ng);
    
    this._progress.value = Math.floor (context.progress * 100);
  },
  
  /**
   * 保存完了
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Array ngList
   *         失敗したファイル
   *           [String ファイル名, ...]
   */
  onSave : function (context, ngList) {
    this._progressLabel.value = UnMHTStringBundle.get ("save_done");
    
    this._owner.onSave (this, context, ngList);
    this.destruct ();
  },
  
  /* ==== private ==== */

  _window : null, /* nsIDOMWindow  対象のウィンドウ */
  
  _owner : null,  /* UnMHTSaveProgressOwner  オーナー */
  
  _progressLabel : null,       /* nsIDOMXULElement  取得状況のラベル */
  _progressCurrent : null,     /* nsIDOMXULElement  現在取得しているファイル */
  _progressCurrentSize : null, /* nsIDOMXULElement  現在取得しているファイルの
                                *                   サイズ */
  _progress : null,            /* nsIDOMXULElement  取得状況のバー */
  
  _failedLabel1 : null,        /* nsIDOMXULElement  失敗したファイルのラベル */
  _failedLabel2 : null,        /* nsIDOMXULElement  失敗したファイルのラベル2 */
  _failedLabel3 : null,        /* nsIDOMXULElement  失敗したファイルのラベル3 */
  _failedList : null,          /* nsIDOMXULElement  失敗したファイル */
  
  /**
   * 保存結果を表示する
   */
  showFail : function (ngList) {
    if (UnMHTPref.getPref ("save.show_errors") && ngList.length > 0) {
      this._failedList.value = ngList.join ("\n");
      this._failedList.hidden = false;
      
      if (this._failedLabel1 && this._failedLabel1.value != "") {
        this._failedLabel1.hidden = false;
      }
      if (this._failedLabel2 && this._failedLabel2.value != "") {
        this._failedLabel2.hidden = false;
      }
      if (this._failedLabel3 && this._failedLabel3.value != "") {
        this._failedLabel3.hidden = false;
      }
      
      this._window.sizeToContent ();
    }
  },
  
  /**
   * getElementById の省略
   *
   * @param  String id
   *         要素の ID
   * @return nsIDOMElement
   *         id に対応する要素
   *         見付からなければ null
   */
  el : function (id) {
    return this._window.document.getElementById (id);
  }
};
