/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTMap.jsm");
XPCOMUtils.importRelative (this, "UnMHTPref.jsm");
XPCOMUtils.importRelative (this, "UnMHTPrototype.jsm");
XPCOMUtils.importRelative (this, "UnMHTShutdown.jsm");
XPCOMUtils.importRelative (this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative (this, "UnMHTTimer.jsm");
XPCOMUtils.importRelative (this, "UnMHTWindowHook.jsm");

let EXPORTED_SYMBOLS = ["UnMHTStatusbarProgressMap",
                        "UnMHTStatusbarProgress"];

/**
 * window から UnMHTStatusbarProgress へのマップ
 */
let UnMHTStatusbarProgressMap = new UnMHTMap ();
UnMHTShutdown.addListener (function (data, reason) {
    UnMHTStatusbarProgressMap.clear ();
    UnMHTStatusbarProgressMap.destruct ();
  });

/**
 * 保存状況をステータスバーに表示するリスナ
 *
 * @param  nsIDOMWindow window
 *         実行中のウィンドウ
 * @inherits   UnMHTWindowHook
 * @implements IUnMHTWindowHook
 * @implements IUnMHTSaveListener
 */
function UnMHTStatusbarProgress (window) {
  UnMHTWindowHook.call (this, window);
  
  this._contexts = [];
  
  UnMHTStatusbarProgressMap.add (window, this);
}
UnMHTStatusbarProgress.prototype = {
  /**
   * 保存コンテキストを監視対象に追加
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Function callback
   *         保存完了/中断時のコールバック
   */
  addContext : function (context, callback) {
    this._contexts.push ([context, callback]);
  },
  
  /**
   * 保存コンテキストを監視対象から削除
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   */
  removeContext : function (context) {
    this._contexts = this._contexts.filter (function ([c, f]) c != context);
    context.destruct ();
  },
  
  /**
   * 保存をキャンセルする
   */
  cancelSave : function () {
    this._contexts.forEach (function ([context, callback]) context.cancel ());
    this._contexts = [];
    
    this._updateProgress ();
  },
  
  /* ==== IUnMHTWindowHook インターフェース ==== */

  /**
   * 破棄
   */
  destruct : function () {
    try {
      UnMHTStatusbarProgressMap.remove (this._window);
    }
    catch (e) {
    }
    
    if (this._contexts) {
      this._contexts
        .forEach (function ([context, callback]) {
            context.cancel ();
          });
      this._contexts = [];
    }
    
    this._prog = null;
    this._progbar = null;
    this._proglabel = null;
    
    UnMHTPrototype.super (this, "destruct").call (this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param  Boolean isStartup
   *         スタートアップイベントか
   * @param  Number reason
   *         スタートアップの理由
   */
  load : function (isStartup, reason) {
    let pref = this.el ("unmht-statusbarpanel-preferences");
    this._prog = this.el ("unmht-statusbarpanel-progress");
    this._progbar = this.el ("unmht-statusbarpanel-progressbar");
    this._proglabel = this.el ("unmht-statusbarpanel-progress-label");
    if (pref && pref.parentNode && this._prog) {
      pref.parentNode.insertBefore (this._prog, pref.nextSibling);
    }
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param  Boolean isShutdown
   *         シャットダウンイベントか
   * @param  Number reason
   *         シャットダウンの理由
   */
  unload : function (isShutdown, reason) {
    if (this._prog) {
      this._prog.hidden = true;
    }
    if (this._progbar) {
      this._progbar.hidden = true;
    }
    if (this._proglabel) {
      this._proglabel.hidden = true;
    }
  },
  
  /* ==== IUnMHTSaveListener インターフェース ==== */
  
  /**
   * キャンセルが完了したイベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   */
  onCancel : function (context) {
    this._contexts
    .filter (function ([c, callback]) c == context)
    .forEach (function ([c, callback]) {
        if (callback) {
          callback (true);
        }
      });
    
    this.removeContext (context);
  },
  
  /**
   * 保存失敗イベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Number reason
   *         エラーの理由
   * @param  Exception e
   *         例外
   */
  onError : function (context, reason, e) {
    let mes = "";
    switch (reason) { 
      case UnMHTSave.ERROR_ADD_DOCUMENT: {
        mes = UnMHTStringBundle.get ("error_save");
        break;
      }
      case UnMHTSave.ERROR_WRITE_FILE: {
        mes = UnMHTStringBundle.get ("error_open");
        break;
      }
    }
    
    let confirmURL = "chrome://unmht/content/confirm.xul";
    let features = "chrome,dialog,modal,centerscreen";
    let arg = {
      title: UnMHTStringBundle.get ("error_title"),
      message1: mes,
      message2: e ? e.toString () : "",
      accept: UnMHTStringBundle.get ("ok"),
    };
    this._window.openDialog (confirmURL, "", features, arg);
    
    this._contexts
    .filter (function ([c, callback]) c == context)
    .forEach (function ([c, callback]) {
        if (callback) {
          callback (true);
        }
      });
    
    this.removeContext (context);
  },
  
  /**
   * 進捗状況変化イベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  String currentFile
   *         現在取得のファイル名
   * @param  String currentSize
   *         取得したバイト数 / ファイルサイズ
   * @param  Number ok
   *         取得完了した数
   * @param  Number na
   *         取得中の数
   * @param  Number ng
   *         取得失敗した数
   */
  onProgress : function (context, currentFile, currentSize, ok, na, ng) {
    this._updateProgress ();
  },
  
  /**
   * 保存完了
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Array ngList
   *         失敗したファイル
   *           [String ファイル名, ...]
   */
  onSave : function (context, ngList) {
    if (UnMHTPref.getPref ("save.show_errors") && ngList.length > 0) {
      let nglistURL = "chrome://unmht/content/nglist.xul";
      let features = "chrome,titlebar,toolbar,centerscreen,resizable";
      let arg = {
        ngList : ngList
      };
      this._window.openDialog (nglistURL, "", features, arg);
    }
    
    this._contexts
    .filter (function ([c, callback]) c == context)
    .forEach (function ([c, callback]) {
        if (callback) {
          callback (false);
        }
      });
    
    this._updateProgress ();
    this.removeContext (context);
  },
  
  /* ==== private ==== */
  
  _contexts : null,  /* Array 保存コンテキストとコールバックのマップ
                      *   [[UnMHTSaveContext, Function], ...] */
  
  _window : null,    /* nsIDOMWindow  実行中のウィンドウ */
  _prog : null,      /* nsIDOMXULElement  進捗状況全体のボックス */
  _progbar : null,   /* nsIDOMXULElement  進捗状況のプログレスバー */
  _proglabel : null, /* nsIDOMXULElement  進捗状況のラベル */
  
  /**
   * ステータスバーの進捗状況を更新する
   */
  _updateProgress : function () {
    let na = 0;
    let ok = 0;
    let progress = 0;
    
    this._contexts.forEach (function ([context, callback]) {
        if (context.progress == 1) {
          ok ++;
        }
        else {
          na ++;
        }
        progress += context.progress;
      });
    progress /= this._contexts.length;
    
    if (na == 0) {
      if (ok > 0) {
        /* 保存完了 */
        this._prog.hidden = false;
        this._progbar.hidden = true;
        this._proglabel.hidden = false;
        this._proglabel.value = UnMHTStringBundle.get ("save_done");
        
        UnMHTTimer.setTimeout
          (function () {
            this._prog.hidden = true;
          }.bind (this), 3000);
      }
      else {
        /* 現在なにも無い */
        this._prog.hidden = true;
      }
    }
    else {
      /* 保存中 */
      
      this._prog.hidden = false;
      this._progbar.hidden = false;
      this._progbar.setAttribute ("value", Math.floor (progress * 100));
      this._proglabel.hidden = true;
    }
  }
};
UnMHTPrototype.createInheritance (UnMHTStatusbarProgress, UnMHTWindowHook);
