/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTOverlayLoader.jsm");

let EXPORTED_SYMBOLS = ["UnMHTWindowHookLoader"];

/**
 * ウィンドウイベントリスナ
 *
 * @implements nsIWindowMediatorListener
 */
function UnMHTWindowListener (loader) {
  this._loader = loader;
}
UnMHTWindowListener.prototype = {
  QueryInterface:    XPCOMUtils.generateQI ([Ci.nsIWindowMediatorListener]),
  
  /**
   * 破棄
   */
  destruct : function () {
    this._loader = null;
  },
  
  /* ==== nsIWindowMediatorListener インターフェース ==== */
  
  /**
   * ウィンドウを開いたイベント
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   */
  onOpenWindow: function (window) {
    window
    = window.QueryInterface (Ci.nsIInterfaceRequestor)
    .getInterface (Ci.nsIDOMWindow);
    
    let loadFunc = function () {
      window.removeEventListener ("load", loadFunc, false);
      
      this._loader.onLoad (window);
    }.bind (this);
    
    window.addEventListener ("load", loadFunc, false);
  },
  
  /**
   * ウィンドウを閉じたイベント
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   */
  onCloseWindow: function (window) {
    /* onOpenWindow と window が一致しないので使用しない */
  },
  
  /**
   * ウィンドウを開いたイベント
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   * @param  String title
   *         タイトル
   */
  onWindowTitleChange: function (window, title) {
  },
  
  /* ==== private ==== */
  
  _loader : null, /* UnMHTWindowHookLoader  ウィンドウフックローダー */
};

/**
 * ウィンドウフックローダー
 *
 * @param  Object overlays
 *         オーバーレイ定義
 *           <<String  documentURI,
 *             {
 *               uri   : String オーバーレイの URI
 *               props : Object  ウィンドウに付与するプロパティ
 *               hooks : Array  フック
 *                         [Function{IUnMHTWindowHook}, ...]
 *             }>>
 * @param  Array dialogs
 *         ダイアログの URI
 *           [String, ...]
 * @implements IUnMHTStartupListener
 */
function UnMHTWindowHookLoader (overlays, dialogs) {
  this._overlays = overlays;
  this._dialogs = dialogs;
  
  this._listener = new UnMHTWindowListener (this);
  
  this.onUnload = this.onUnload.bind (this);
}
UnMHTWindowHookLoader.prototype = {
  /**
   * 破棄
   */
  destruct : function () {
    if (this._listener) {
      this._listener.destruct ();
      this._listener = null;
    }
    
    this._overlays = null;
    this._dialogs = null;
    this._listener = null;
  },
  
  /**
   * ウィンドウがロードされたイベント
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   */
  onLoad : function (window) {
    this._loadInto  (window, false, -1);
  },
  
  /**
   * ウィンドウに登録するアンロードイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onUnload : function (event) {
    this._unloadFrom (event.target.defaultView, false, -1);
  },
  
  /**
   * スタートアップ時のイベント
   *
   * @param  Object data
   *         Bootstrap データ
   * @param  Number reason
   *         理由
   */
  onStartup : function (data, reason) {
    for (let window in XPCOMUtils.IterSimpleEnumerator
           (Services.wm.getEnumerator (null), Ci.nsIDOMWindow)) {
      this._loadInto (window, true, reason);
    }
    
    Services.wm.addListener (this._listener);
  },
  
  
  /**
   * シャットダウン時のイベント
   *
   * @param  Object data
   *         Bootstrap データ
   * @param  Number reason
   *         理由
   */
  onShutdown : function (data, reason) {
    Services.wm.removeListener (this._listener);
    
    for (let window in XPCOMUtils.IterSimpleEnumerator
           (Services.wm.getEnumerator (null), Ci.nsIDOMWindow)) {
      this._unloadFrom (window, true, reason);
      
      if (this._dialogs.indexOf (window.document.documentURI) != -1) {
        window.close ();
      }
    }
  },
  
  /* ==== private ==== */
  
  _registered : [], /* 組み込みリスト
                     *   [[nsIXULWindow, Object], ...] */
  _listener : null, /* UnMHTWindowListener  ウィンドウイベントリスナ */
  
  _overlays: null,  /* Object  オーバーレイ定義 */
  _dialogs: null,   /* Array  ダイアログの URI */
  
  /**
   * ウィンドウにフックをロードする
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   * @param  Boolean isStartup
   *         スタートアップイベントか
   * @param  Number reason
   *         スタートアップの理由
   */
  _loadInto : function (window, isStartup, reason) {
    let location = window.document.documentURI;
    if (!(location in this._overlays)) {
      return;
    }
    
    let overlay = this._overlays [location];
    
    let callback = function () {
      for (let [name, value] in Iterator (overlay.props)) {
        window [name] = value;
      }
      
      let list = overlay.hooks.map (function (n) new n (window));
      
      list.forEach (function (item) item.load (isStartup, reason));

      this._registered.push ([window, list]);
      
      window.addEventListener ("unload", this.onUnload, false);
    }.bind (this);
    
    if (overlay.uri) {
      UnMHTOverlayLoader.loadInto (window, overlay.uri, callback);
    }
    else {
      callback ();
    }
  },
  
  /**
   * ウィンドウからフックをアンロードする
   *
   * @param  nsIXULWindow window
   *         対象のウィンドウ
   * @param  Boolean isShutdown
   *         シャットダウンイベントか
   * @param  Number reason
   *         シャットダウンの理由
   */
  _unloadFrom : function (window, isShutdown, reason) {
    if (window.document) {
      let location = window.document.documentURI;
      if (location in this._overlays) {
        let overlay = this._overlays [location];
        
        for (let [name, value] in Iterator (overlay.props)) {
          delete window [name];
        }
        
        UnMHTOverlayLoader.unloadFrom (window, overlay.uri);
      }
    }
    
    window.removeEventListener ("unload", this.onUnload, false);
    
    this._registered
    = this._registered.filter (function ([w, list]) {
        if (w == window) {
          list.forEach (function (item) {
              item.unload (isShutdown, reason);
              item.destruct ();
            });
          return false;
        }
        return true;
      });
  }
};
