/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative (this, "UnMHTAppInfo.jsm");
XPCOMUtils.importRelative (this, "UnMHTArrayUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTWindowUtil"];

/**
 * Window 操作ユーティリティ
 */
let UnMHTWindowUtil = {
  /**
   * ブラウザを取得する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @return nsIDOMXULElement
   *         ブラウザ
   */
  getBrowser : function (window) {
    if (this.isMailWindow (window)) {
      return window.document.getElementById ("messagepane");
    }
    
    return window.document.getElementById ("content");
  },
  
  /**
   * 現在開いているドキュメントを取得する
   *
   * @param  nsIDOMWindow window
   *         実行中のウィンドウ
   * @return nsIDOMDocument
   *         現在開いているドキュメント
   */
  getCurrentDocument : function (window) {
    return this.getBrowser (window).contentDocument;
  },
  
  /**
   * ウィンドウのタブの配列を取得する
   *
   * @param  nsIDOMWindow window
   *         対象のウィンドウ
   * @return Array
   *         タブの配列
   *           [nsIDOMXULElement, ...]
   */
  getTabs : function (window) {
    let tabbrowser = window.document.getElementById ("content");
    if (!tabbrowser) {
      return [];
    }
    
    return ("visibleTabs" in tabbrowser)
    ? tabbrowser.visibleTabs
    : tabbrowser.tabs;
  },

  /**
   * ウィンドウのブラウザの配列を取得する
   *
   * @param  nsIDOMWindow window
   *         対象のウィンドウ
   * @return Array
   *         ブラウザの配列
   *           [nsIDOMXULElement, ...]
   */
  getBrowsers : function (window) {
    try {
      window.document;
    }
    catch (e) {
      Cu.reportError (e + "\n" + e.stack);
    }
    let tabbrowser = window.document.getElementById ("content");
    if (!tabbrowser) {
      return [];
    }
    
    return Array.map
    (this.getTabs (window),
     function (tab) tabbrowser.getBrowserForTab (tab));
  },

  /**
   * ウィンドウのドキュメントの配列を取得する
   *
   * @param  nsIDOMWindow window
   *         対象のウィンドウ
   * @return Array
   *         ドキュメントの配列
   *           [nsIDOMDocument, ...]
   */
  getDocuments : function (window) {
    return Array.map
    (this.getBrowsers (window),
     function (browser) browser.contentDocument);
  },
  
  /**
   * ドキュメントに対応するタブを返す
   *
   * @param  nsIDOMWindow window
   *         対象のウィンドウ
   * @param  nsIDOMDocument targetDocument
   *         対象のドキュメント
   * @return nsIDOMXULElement
   *         対応するタブ
   *         見付からなかったら null
   */
  getTabForDocument : function (window, targetDocument) {
    let tabbrowser = this.getBrowser (window);
    if (!tabbrowser) {
      return null;
    }
    
    return UnMHTArrayUtil.find
    (tabbrowser.tabs,
     function (tab) (tab.linkedBrowser
                     && tab.linkedBrowser.contentDocument == targetDocument),
     null);
  },
  
  /**
   * メッセージパネルが閉じているかどうか調べる
   *
   * @param  nsIDOMWindow window
   *         対象のウィンドウ
   * @return Boolean
   *         メッセージパネルが閉じているか
   */
  isMessagePaneCollapsed : function (window) {
    if (!this.isMailWindow (window)) {
      return false;
    }
    
    let node;
    node = window.document.getElementById ("messagepanebox");
    if (node && node.collapsed) {
      return true;
    }
    node = window.document.getElementById ("messagepaneboxwrapper");
    if (node && node.collapsed) {
      return true;
    }
    
    return false;
  },
  
  /**
   * メールのウィンドウかどうかを返す
   *
   * @param  nsIDOMWindow window
   *         対象のウィンドウ
   * @return Boolean
   *         メールのウィンドウかどうか
   */
  isMailWindow : function (window) {
    return (window.document.documentURI
            == "chrome://messenger/content/messenger.xul");
  }
};
